\ EnumSerial.fth - Enumerate serial ports several ways

((
Copyright (c) 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)

http://stackoverflow.com/questions/1388871/how-do-i-get-a-list-of-available-serial-ports-in-win32

The EnumSerialPorts v1.20 suggested by Nick D uses nine
different methods to list the serial ports! We're certainly
not short on choice, though the results seem to vary.

To save others the trouble, I'll list them here and indicate
their success in finding the com0com ports on my PC (XP Pro
SP2):

    CreateFile("COM" + 1->255) as suggested by Wael Dalloul
    ? Found com0com ports, took 234ms.

    QueryDosDevice()
    ? Found com0com ports, took 0ms.

    GetDefaultCommConfig("COM" + 1->255)
    ? Found com0com ports, took 235ms.

    "SetupAPI1" using calls to SETUPAPI.DLL
    ? Found com0com ports, also reported "friendly names", took 15ms.

    "SetupAPI2" using calls to SETUPAPI.DLL
    ? Did not find com0com ports, reported "friendly names", took 32ms.

    EnumPorts()
    ? Reported some non-COM ports, did not find com0com ports, took 15ms.

    Using WMI calls
    ? Found com0com ports, also reported "friendly names", took 47ms.

    COM Database using calls to MSPORTS.DLL
    ?/? Reported some non-COM ports, found com0com ports, took 16ms.

    Iterate over registry key HKEY_LOCAL_MACHINE\HARDWARE\DEVICEMAP\SERIALCOMM
    ? Found com0com ports, took 0ms. This is apparently what SysInternals PortMon uses.

Based on those results I think the WMI method probably suits
my requirements best as it is relatively fast and as a bonus
it also gives the friendly names (e.g. "Communications Port
(COM1)", "com0com - serial port emulator").

Thanks.


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *! enumserial
\ *T Enumerating serial ports
\ ===========
\ *P This code is really tedious.

  include %vfxpath%\Lib\GUID


\ *************************
\ *S Using QueryDosDevice()
\ *************************
\ *P This method works easily but does not give us the "friendly"
\ ** names.

0 value TargetPaths	\ -- addr
\ *G Points to buffer for enumerating COM ports.
$0F000 value /TargetPaths	\ -- len
\ *G Initial size of buffer. QueryDosDevice() will fail for
\ ** size>=65536 under XP, so we just fix it at $F000. It'll
\ ** have to do for now.

Extern: DWORD WINAPI QueryDosDevice(
  char * lpDeviceName,
  char * lpTargetPath,
  DWORD ucchMax
);

: initPortFinder	\ -- ior
  /TargetPaths allocate dup if
    nip
  else
   drop  -> TargetPaths  0
  endif
; doNotSin

: termPortFinder	\ --
\ *G Free the COM port buffer.
  TargetPaths free drop  0 -> TargetPaths
; doNotSin

: getDosDevs	\ -- ior
  begin
    0 TargetPaths /TargetPaths QueryDosDevice 0=
   while
    GetLastError ERROR_INSUFFICIENT_BUFFER <>
    if  -1 exit  endif
    /TargetPaths 2* -> /TargetPaths
    termPortFinder initPortFinder
    if  -2 exit  endif
  repeat
  0
;

: showCOMports	\ --
  TargetPaths begin
    dup c@				\ until end
   while
    dup s" COM" s= if			\ starts with COM?
      cr dup zcount type
    endif
    zcount + 1+				\ step over input
  repeat
  drop
;

: getCOMports	\ --
  initPortFinder 0= if
    getDosDevs 0=
    if  showComPorts  endif
    termPortFinder
  endif
;


\ *********************
\ *S Using setupapi.dll
\ *********************
\ *P This section is the result of more effort than is proper.
\ ** The code is not pretty.

Library: setupapi.dll

Extern: HANDLE SetupDiOpenDevRegKey(
  HANDLE DeviceInfoSet,
  void * DeviceInfoData,
  DWORD Scope,
  DWORD HwProfile,
  DWORD KeyType,
  dword samDesired
);

Extern: HANDLE SetupDiGetClassDevs(
  const void * ClassGuid,
  LPCTSTR Enumerator,
  HWND hwndParent,
  DWORD Flags
);

Extern: BOOL SetupDiGetDeviceRegistryProperty(
  _In_       HANDLE DeviceInfoSet,
  _In_       void * DeviceInfoData,
  _In_       DWORD Property,
  _Out_opt_  DWORD * PropertyRegDataType,
  _Out_opt_  void * PropertyBuffer,
  _In_       DWORD PropertyBufferSize,
  _Out_opt_  DWORD * RequiredSize
);

Extern: BOOL SetupDiDestroyDeviceInfoList(
  _In_  HANDLE DeviceInfoSet
);

Extern: BOOL SetupDiEnumDeviceInfo(
  _In_   HANDLE DeviceInfoSet,
  _In_   DWORD MemberIndex,
  _Out_  void * DeviceInfoData
);

Extern: LONG RegQueryValueEx(
  HANDLE hKey,
  LPCTSTR lpValueName,
  DWORD * lpReserved,
  DWORD * lpType,
  void * lpData,
  DWORD * lpcbData
);

Extern: LONG RegCloseKey(
  HANDLE hKey
);

struct /DEVINFO_DATA	\ -- len
\ *G Common structure for SP_DEVINFO_DATA.
  _dword field did.cbSize
  16     field did.ClassGuid
  _dword field did.DevInst
  _ptr   field did.Reserved
end-struct

UUID GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR {4D36E978-E325-11CE-BFC1-08002BE10318}
\ *G GUID used to search for COM ports - this one does USB ports
\ ** as well as real ports..
\ UUID GUID_DEVINTERFACE_COMPORT {86E0D1E0-8089-11D0-9CE4-08003E301F73}
\ old one

/DEVINFO_DATA buffer: devInfo

512 constant /PortName		\ -- len
\ *G Size of port name assembly buffer.
/PortName buffer: zPortName	\ -- addr
\ *G Port name assembly buffer.
256 constant /FriendlyName	\ -- len
\ *G Size of device description buffer.
/FriendlyName buffer: zFriendlyName	\ -- addr
\ *G Device description buffer.

: getPortName	\ hDevKey -- ior
\ *G Return non-zero on failure. The port name is in the global
\ ** buffer *\fo{zPortName}.
  {: hKey | dwType dwDataSize -- ior :}
  0 -> dwType  /PortName 1- -> dwDataSize
  0 zPortName c!
  hKey z" PortName" NULL  addr dwType  zPortName  addr dwDataSize
  RegQueryValueEx
  if  -1 exit  endif
  0 zPortName dwDataSize + c!		\ force zero termination
  dwType REG_SZ <>			\ must be a string
  if  -2 exit  endif
  0
;

: EnumSerSetupAPI1	\ --
  {: | hDevInfoSet hDevKey nIndex dwSize dwType -- ior :}

\ create a "device information set" which is required to enumerate all the ports
  GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR
  NULL NULL
  DIGCF_PRESENT ( DIGCF_DEVICEINTERFACE or )
  SetupDiGetClassDevs dup -> hDevInfoSet INVALID_HANDLE_VALUE =
  if  exit  endif

  0 -> nIndex
  begin
    \ get the device info data
    /DEVINFO_DATA devInfo did.cbSize l!
    hDevInfoSet nIndex devInfo SetupDiEnumDeviceInfo
   while
    \ Get the registry key which stores the ports settings
    hDevInfoSet devInfo DICS_FLAG_GLOBAL 0 DIREG_DEV KEY_QUERY_VALUE
    SetupDiOpenDevRegKey dup -> hDevKey
    INVALID_HANDLE_VALUE <> if		\ if a good key
      hDevKey getPortName 0= if		\ does it represent a COM port?
        zPortName zstrlen 3 > if	\ COMxxx
          zPortName s" COM" s= if	\ -- caddr len
            \ extract the "device description"
            0 zFriendlyName c!
            /FriendlyName -> dwSize
            0 -> dwType
            hDevInfoSet devInfo SPDRP_DEVICEDESC
            addr dwType  zFriendlyName  dwSize  addr dwSize
	    SetupDiGetDeviceRegistryProperty if
	      dwType REG_SZ = if
	        s"  " zPortName zappend	\ add device description to port name
	        zFriendlyName zcount zPortName zappend
	      endif
	    endif
	    cr zPortName zcount type
          endif
        endif
      endif
      hDevKey RegCloseKey drop
    endif
    nIndex 1+ -> nIndex
  repeat

  hDevInfoSet SetupDiDestroyDeviceInfoList drop
;


\ ======
\ *> ###
\ ======

decimal
