(*
 * LANGUAGE    : Forth
 * PROJECT     : Working on MPE and iForth
 * DESCRIPTION : Audio processing JACK client
 * CATEGORY    : Realtime Audio tools
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2010-05-13 15:29:26
		2009-05-16 17:15:24 turnkeying ok; new 0string conversion
		2009-07-14 13:07:26 more precise error messages, checking clear connections
		2009-11-12 19:52:27 safe client stopping
		2009-11-19 23:09:31 added time-relative access to original-audio-data
		2009-12-07 23:14:20 more concise info
		2009-12-26 13:15:18 added STREAM-EXTERN-FRAMES
		2010-01-21 19:02:36 frequencies are floats now
		2010-01-27 02:04:14 #FJACK-SAMPLE correct when no plugins are installed
		2010-01-29 03:21:23 WAV handling in starting/stopping the client moved to wavfiles.frt
		2010-02-01 23:40:33 added GRAPH-REORDER-HOOK and the reordering callback
		2010-02-02 00:00:57 added PORT-CONNECT-HOOK  and the port connect callback
		2010-02-07 00:07:07 JACK strings are initialised as 0nullstrings
		2010-05-13 15:21:27 crashing/stopping jackd resulted in hanging fjack; fixed
		2011-07-31 19:46:38 added session management, LOG-INFO message system
		                    zombie handling fixed, client name defined by application
		2011-08-09 17:31:28 connections are handled without a global flag
		2011-09-29 13:42:18 wavboard integrated
		2012-07-10 23:21:00 MIDI support activated
		2012-07-15 17:55:23 Precise to the frame timing for MIDI callback
		2013-06-07 06:19:29 Testing for invalid client-id in client-name
		2014-03-02 16:29:09 jacklog.txt is always in the sysfiles directory, append to end of file
		2015-02-09 19:49:28 multichannel sfloats are working now for playing
		2016-05-09 18:04:47 jacksession fixing for starting the client; very strange bugs because of stack corruption in the Extern:
		2017-06-02 18:21:55 fixed many small things for 3.0 version
 *)

\ *S Using the Client
JACKAUDIO DEFINITIONS
\ The callbacks tell about the JACK server status, it's kept in #jack-state
0 VALUE #jack-state
0 VALUE fJACK-MAXLAT

FORTH DEFINITIONS
0 VALUE JACK-POSITION	( -- n )
\ *G The sample position in the audio data stream shared by all clients, it's an unsigned 32bit integer. Read-Only!
0 VALUE #FJACK-SAMPLE	( -- n )
\ *G In the input->output processing chain you can use this value as the sample number.
\ ** It is derived from *\forth{JACK-POSITION}. Read-Only in processing chain!
1E0 FVALUE PLAY-WAV-LEVEL
\ *G The level of the WAV data played common for all channels. The default of 1.0 is ok in most situations, when you
\ ** have wav files that are a) 8bit-coded b) on almost full level and c) need resampling you might set this to a smaller value
\ ** like 0.9 to avoid distortions.
1E0 FVALUE READ-WAV-LEVEL
\ *G The level of the recorded=read WAV data common for all channels.
1E0 FVALUE JACK-IO-LEVEL
\ *G This level is applied to data read at the InputPort before the Input->Output processing is done,
\ ** it's common for all channels. The leveler plugin uses it as an example.
#10000 VALUE JACK-SAMPLE-RATE	( -- n )
\ *G The sample rate the JACK server is running. The fJACK client is
\ ** told about it via a callback, you can - and should - only read it.
1000e0 FVALUE JACK-SAMPLE-F-RATE

#1024 VALUE JACK-CHKSIZE	( -- n )
\ *G Number of samples in a *\bold{process frames callback} as returned by the JACK server.
0 VALUE JACK-XRUNS	( -- n )
\ *G Counter for the xrun callbacks
0 VALUE fJACK-ID	( -- n )
\ *G The client id returned by the JACK library call when registering a client is kept here.

: JACK-LOAD		( -- n )
\ *G The JACK server measures the time spent in the clients callback by using the high-resolution-timer
\ ** and calculates the 'load' as it also knows the time between callbacks. The value n is the load in percent.
	fJACK-ID jack_cpu_load 0.5E0 F+ F>S ;

: JACK-RUNNING?		( -- flag )
\ *G To be used by the user application testing both the client id and the state given by
\ ** the processing callback. *\forth{TRUE} when the client is really active.
	fJACK-ID 0<>  #jack-state JackZombie <> AND ;

: JACK-ZOMBIE?		( -- flag )
\ *G This flag is TRUE when the fjack client was zombiefied by the server.
	fJACK-ID 0<>  #jack-state JackZombie = AND ;

0 VALUE JACK-SESSION-BYE?
\ *G This is a global value that can be tested by user applications. It is set to *\bold{TRUE} when a jack session manager
\ ** like QjackCtl tells save&quit. As there is no global safe way to leave all apllications this is done via this flag.

DEFER DO-PROCESS-MIDI-IN-EVENT ( struct -- ) ' DROP IS DO-PROCESS-MIDI-IN-EVENT

DEFER GRAPH-REORDER-HOOK ( -- )	' NOOP IS GRAPH-REORDER-HOOK
\ *G Called by the reorder callback; you can define a function watching the jack environment, for example
\ ** you could make sure about connections.

DEFER PORT-CONNECT-HOOK ( a b flag -- )	' 3DROP IS 	PORT-CONNECT-HOOK
\ *G Called be the port-connect callback; you can define a function that looks for connections.
\ ** This function must take the flag and both port id's from the stack

DEFER SAVE-SESSION-DATA	( event -- )	' DROP IS SAVE-SESSION-DATA
\ *G The session management of the jack server allows saving of data, this is a hook for your user applications,
\ ** See *\forth{DEFAULT-SESSION-SAVER} as an example.


JACKAUDIO DEFINITIONS
0 VALUE CALC-LATENCIES	\ maybe an xt
0 VALUE IN-MAXLATENCY
0 VALUE OUT-MAXLATENCY
CREATE *latencies 2 CELLS ALLOT
CREATE *session-arg 0 , 0 ,

SEMAPHORE MIDISEM
: <MIDI	MIDISEM GETSEMAPHORE ;
: MIDI>	MIDISEM FREESEMAPHORE ;


1E0 FVALUE #P-SCW8
1E0 FVALUE #P-SCW16
1E0 FVALUE #P-SCW32
1E0 FVALUE #R-SCW8
1E0 FVALUE #R-SCW16
1E0 FVALUE #R-SCW32
: PREP-SCALERS
	PLAY-WAV-LEVEL $80 S>F F/	TO #P-SCW8
	PLAY-WAV-LEVEL $8000 S>F F/	TO #P-SCW16
	PLAY-WAV-LEVEL $80000000 S>F F/	TO #P-SCW32
	READ-WAV-LEVEL $80 S>F F*	TO #R-SCW8
	READ-WAV-LEVEL $8000 S>F F*	TO #R-SCW16
	READ-WAV-LEVEL $80000000 S>F F*	TO #R-SCW32 ;
: WAV8@		C@ $80 - S>F ;
: WAV16@	S16B@  S>F ;
: WAV32@	S32B@  S>F ;
: WAV8!		F>S $80 + 0 MAX #255 MIN SWAP C! ;
: WAV16!	F>S $-7FFF MAX $7FFF MIN SWAP 16B! ;
64BIT?
[IF]   : WAV32!	F>S $-7FFFFFFF MAX $7FFFFFFF MIN SWAP 32B! ;
[ELSE] : WAV32! F>S SWAP 32B! ;
[THEN]
$100000 CONSTANT #wbdscale
: C>FSTRING	( a -- a cnt )			DUP BEGIN DUP C@ WHILE 1+ REPEAT OVER - ;
: PACK0		( a len a -- a )		SWAP 2DUP 2>R MOVE 2R>  OVER + 0 SWAP C! ;
: #B-ALIGNED	( addr blocksize -- 'addr )	1- TUCK + SWAP -1 XOR AND ;


0 FIFO-VALUE >f-addr		\ address where this channels data start
  FIFO-VALUE >f-channels	\ channels available in this wav block
  FIFO-VALUE >f-todo		\ Samples (corrected tgo the sample-rate) left to be worked for this channel
  FIFO-VALUE >f-mode		\ can be 8=U8bit 16=16bit 32=32bitint -32=32bitfloat
  FIFO-VALUE >f-done		\ done already - also corrected to the sample rate
  FIFO-VALUE >f-incr		\ bytes in one frame, as data in wav blocks are channel-interleaved
  FIFO-VALUE >f-fract		\ fraction of sampling rate
    CONSTANT #fifosize		\ each fifo object may have the size of 8 cells

0 VALUE #play-current		\ index of current output WAV fifo
0 VALUE #play-newest		\ index of newest output WAV fifo
0 VALUE #read-current		\ index of current for input WAV fifo
0 VALUE #read-newest		\ index of newest input WAV fifo

0 VALUE *playfifos		\ pointer to playfifo structs
0 VALUE *recordfifos		\ pointer to readfifo structs
0 VALUE *io-plug-slots		\ pointer to plugin slot memory
0 VALUE jack_position_t		\ pointer to position callback save area
0 VALUE *historybuffers		\ large history buffers for plugins and in/out
0 VALUE *jack-high-mem		\ end of dynamic memory

0 VALUE wboard-ticket		\ unique id for any wavboard job
0 VALUE *wboard-handles
0 VALUE *wboard-jobs
0 WBD-VALUE wbd-start-frame
  WBD-VALUE wbd-channel
  WBD-VALUE wbd-address
  WBD-VALUE wbd-#one
  WBD-VALUE wbd-wavmode
  WBD-VALUE wbd-repeats
  WBD-VALUE wbd-samples
  WBD-VALUE wbd-todo
  WBD-VALUE wbd-id
  WBD-VALUE wbd-scale
  WBD-VALUE wbd-fast
  WBD-VALUE wbd-newscale
  WBD-VALUE wbd-orgscale
  WBD-VALUE wbd-stop-frame
  WBD-VALUE wbd-truesamples
  WBD-VALUE wbd-fastok
  WBD-VALUE wbd-newpos
  WBD-VALUE wbd-frames?
  CONSTANT #wboard-handle


\ ****************************************************************************************************
\ The Client struct
HERE	#10 CELLS 2 DFLOATS + JACK-CHANNELS * $400 + ALLOT
	$0200 #B-ALIGNED
  CHAN-VALUE >inport				\ port ids as returned be the libjack function
  CHAN-VALUE >outport				\ each channel has JACK-CHANNELS input and output ports
  CHAN-VALUE >midiport				\ One input and one output MIDI port
  CHAN-VALUE >inaudio				\ the pointers to the currently processed audio data per channel.
  CHAN-VALUE >outaudio				\ Both are in the history
  CHAN-VALUE >inhistory				\ points to the 'parent' which may be an input port or a plugin object
  CHAN-VALUE >outhistory			\ the latest plugin-object in the channels queue
  CHAN-VALUE >latency				\ ioslots detected latency
  CHAN-FVALUE >lastwriteval			\ The last fvalue is the last used chunk of data - resampling!
  CHAN-FVALUE >outport-volume			\ output volume special for this channel
  DROP
\ ****************************************************************************************************

CREATE ACTIVE-PLUGINS #PLUGINS 1+ CELLS ALLOT	\ holds the list of activated plugins

\ ****************************************************************************************************
\ MIDI event handling stuff
$800 CONSTANT #midi-events
#8 CELLS CONSTANT #midi-event-size
#midi-events #midi-event-size * CONSTANT #midi-event-buffer
0 VALUE midi_in_queue
0 VALUE midi_out_queue
0 VALUE #midi_in
0 VALUE #midi_in_read
0 VALUE #midi_out
0 VALUE #midi_out_read
\ ****************************************************************************************************

0	DSP-VALUE plug_xt		\ is always defined by the plugin modules and MUST be the first VALUE
	DSP-VALUE plug_link		\ links all defined plugins
	DSP-VALUE plug_xt_tune		\ the xt of the function specific to this object to tune the frequency
	DSP-VALUE plug_xt_quality	\ the xt function specific for this object to change the quality
	DSP-VALUE plug_xt_initfilter	\ the xt of a function to initialize the object
	DSP-VALUE plug_channel		\ is set when it's plugged in
	DSP-VALUE plug_slot		\ the used slot for this object
 	DSP-VALUE plug_type		\ is a unique pointer to a counted string describing the plugin
	DSP-VALUE plug_filtype		\ can be one of the above list, user extensions should be negative
	DSP-VALUE plug_#used		\ already plugged in somewhere, value is xt-idx
	DSP-VALUE plug_input		\ points to the input history buffer owned by inport or another plugin
	DUP CONSTANT d_plug_output	\ helper constant as offset
	DSP-VALUE plug_output		\ points to the output history buffer
	DSP-VALUE plug_dynmem		\ pointer to the dynamic memory of the plugin
	DSP-VALUE plug_dynmemsize	\ size of the automatically allocated dynamic memory
	DSP-VALUE plug_bypassed
	DSP-VALUE plug_taps		\ used by many filters to define the size
	DSP-VALUE plug_static		\ some FIR filters might need static data to be shared
	DSP-VALUE plug_timeshift
	DSP-VALUE plug_timeshift_min	\ this is the minimum added delay
	DSP-FVALUE plug_level		\ every plugin has a level value
	DSP-FVALUE plug_frequency	\ is used by almost all plugins so its public
	DSP-FVALUE plug_frequency2	\ is used by some plugins so its public
CONSTANT #PRIVATE-DSP

\ Some library calc need forth-wrappers for string conversions
: (client_open) ( name len opt -- id err? )		{: | buff[ $100 ] error[ $10 ] :}
							JackSessionID -1 XOR AND >R	\ currently suppress the SessionID
							jack_client_name_size $10 - MIN buff[ PACK0 R> error[ *session-arg jack_client_open error[ 32B@ ;
: (client_name)	( id -- name len )			DUP 0= IF 0 EXIT THEN jack_get_client_name C>FSTRING ;
: (port_register) ( id name len f -- pt_id|0 )		{: | buf0[ $100 ] buf1[ $100 ] :} >R buf0[ PACK0  S" 32 bit float mono audio" buf1[ PACK0 R> 0 jack_port_register ;
: (midi_register) ( id name len f -- pt_id|0 )		{: | buf0[ $100 ] buf1[ $100 ] :} >R buf0[ PACK0  S" 8 bit raw midi" buf1[ PACK0 R> 0 jack_port_register ;
: (jack_port_name) ( port_id -- name len )		jack_port_name C>FSTRING ;
: (jack_connect) ( id src len dest len -- ?error )	{: | buf0[ $100 ] buf1[ $100 ] :} buf0[ PACK0 -ROT buf1[ PACK0 SWAP jack_connect ;
: (jack_disconnect) ( id src len dest len -- ?error )	{: | buf0[ $100 ] buf1[ $100 ] :} buf0[ PACK0 -ROT buf1[ PACK0 SWAP jack_disconnect ;

\ start and stop chains
\ each node for the hooks has a) link to lower b) link to upper c) xt
CREATE start-jack-hooks 0 , 0 , 0 ,
CREATE stop-jack-hooks 0 , 0 , 0 ,
CREATE clear-jack-hooks 0 , 0 , 0 ,
: low-mem ( -- )	['] ACTIVE-PLUGINS ;
: safe?	( addr -- f )	low-mem top-mem WITHIN ;
: @nd+	( nd -- nod')	1 CELLS + @ ;
: nd?	( nd -- f )	DUP safe? 0= IF DROP FALSE EXIT THEN
			DUP @ low-mem MAX top-mem MIN @nd+ = ;
: ?ndxt	( nd -- )	2 CELLS + @ DUP safe? IF EXECUTE ELSE DROP THEN ;
: topnd	( list -- nd)	BEGIN DUP @nd+ nd? WHILE @nd+ REPEAT ;
: check ( list -- )	topnd 1 CELLS + OFF ;

: fromtop ( list -- )	DUP 1 CELLS + @ 0= IF DROP EXIT THEN DUP check
			topnd	BEGIN DUP nd? WHILE DUP ?ndxt @ REPEAT DROP ;
: frombot ( list -- )	DUP 1 CELLS + @ 0= IF DROP EXIT THEN DUP check
			@nd+	BEGIN DUP nd? WHILE DUP ?ndxt @nd+ REPEAT DROP ;
: addnd	( xt chain -- )	OVER safe? 0= ABORT" xt for thread hooks out of range"
			topnd ALIGN HERE OVER ,	 SWAP 1 CELLS + ! 0 , , ;
0 VALUE top-instrument
0 VALUE instrument-action
0 VALUE midi-action

\ When starting a client all data must be initialized - done here
: INIT-JACK-VALUES
	0 TO IN-MAXLATENCY	0 TO OUT-MAXLATENCY	0 TO fJACK-MAXLAT
	0 TO #midi_in		0 TO #midi_in_read
	0 TO #midi_out		0 TO #midi_out_read
	0 TO #play-current	0 TO #play-newest
	0 TO #read-current	0 TO #read-newest
	0 TO JACK-POSITION
	FALSE TO JACK-SESSION-BYE?
	JackTransportStopped TO #jack-state
	0 TO JACK-XRUNS
	0 TO instrument-action
	0 TO midi-action ;

\ getting the address of a FIFO struct for a channel ( ch -- a ) ;
\ the currently used fifo is addressed by *C-READ/PLAY-FIDO
\ the newest fifo by *N-READ/PLAY-FIFO
\ You may also access a specific FIFO struct by it's id
: *PLAY-FIFO	( ch id -- a )	#jack-fifos 1- AND SWAP #jack-fifos * +	#fifosize *	*playfifos + ;
: *N-PLAY-FIFO	( ch --a)	#play-newest *PLAY-FIFO ;
: *C-PLAY-FIFO	( ch --a)	#play-current *PLAY-FIFO ;
: *READ-FIFO	( ch id -- a )	#jack-fifos 1- AND SWAP #jack-fifos * +	#fifosize *	*recordfifos + ;
: *N-READ-FIFO	( ch --a)	#read-newest *READ-FIFO ;
: *C-READ-FIFO	( ch --a)	#read-current *READ-FIFO ;
: get-wb-ticket	( -- id )	wboard-ticket 1+ DUP 0= IF 1+ THEN DUP TO wboard-ticket ;

: INPORT-DONE?	( port -- port )	DUP 0= IF <<DEBUG ." The input port couldn't be created" CR DEBUG>> THEN ;
: OUTPORT-DONE?	( port -- port )	DUP 0= IF <<DEBUG ." The output port couldn't be created" CR DEBUG>> THEN ;
: MIDIPORT-DONE? ( port -- port )	DUP 0= IF <<DEBUG ." The midi port couldn't be created" CR DEBUG>> THEN ;

: CREATE-PORTS	 ( -- )
	{: | buf[ $100 ] -- :}
	JACK-CHANNELS 0
	DO	S" Input_" buf[ PLACE  I 1 (#.R) buf[ PLACE+
		fJACK-ID buf[ COUNT JackPortIsInput (port_register)	INPORT-DONE? I TO >inport
		S" Output_" buf[ PLACE I 1 (#.R) buf[ PLACE+
		fJACK-ID buf[ COUNT JackPortIsOutput (port_register)	OUTPORT-DONE? I TO >outport
		1E0 I TO  >outport-volume
		0 I TO >latency
	LOOP
	S" Midi_in" buf[ PLACE fJACK-ID buf[ COUNT JackPortIsInput (midi_register) MIDIPORT-DONE? 0 TO >midiport
	S" Midi_out" buf[ PLACE fJACK-ID buf[ COUNT JackPortIsOutput (midi_register) MIDIPORT-DONE? 1 TO >midiport ;

: FORGET-PORTS	( -- )
	JACK-CHANNELS 2 * 2 + 0
	DO I >inport ?DUP IF fJACK-ID SWAP jack_port_unregister DROP THEN LOOP ;

\ Sometimes - when writing the connection file - the portname and it's mode known to the
\ JACK server must be asked.
: JACK-PORTNAME	( n -- add len)	>inport (jack_port_name) ;
: JACK-PORTFLAG	( n -- mask)	>inport jack_port_flags ;

0 VALUE JACK-TEMP-MEM	\ temporary mem for the buffers
: n->history	( n -- addr )	#HISTORY-SIZE * SFLOATS *historybuffers + ;
: plughistory	( n -- addr )	JACK-CHANNELS + n->history ;
: data-offset	( -- n )	JACK-POSITION #HISTORY-SIZE 1- AND SFLOATS ;

: INIT-PLUGINS
	ACTIVE-PLUGINS #PLUGINS 1+ CELLS ERASE
	JACK-CHANNELS 0
	DO	0e0 TO I >lastwriteval
		I n->history DUP I TO >inhistory	I TO >outhistory
	LOOP ;
:NONAME \ ( -- )
	JACK-CHANNELS  #jack-fifos #fifosize * * 2*			\ fifo memory
	JACK-CHANNELS #PLUGINS + #HISTORY-SIZE * SFLOATS +		\ history buffers
 	JACK-IO-SLOTS CELLS JACK-CHANNELS * +				\ ioplugs
	#midi-event-buffer 2 * +
	#wboard-handle JACK-BOARDJOBS * +
 	$30000 + ALLOCATE ?ALLOCATE TO JACK-TEMP-MEM
 	JACK-TEMP-MEM $1000 #B-ALIGNED
	DUP TO *historybuffers		JACK-CHANNELS #PLUGINS + #HISTORY-SIZE * SFLOATS + $1000 #B-ALIGNED
	DUP TO *io-plug-slots		JACK-IO-SLOTS CELLS JACK-CHANNELS * + $1000 #B-ALIGNED
	DUP TO *playfifos		JACK-CHANNELS  #jack-fifos #fifosize * * + $1000 #B-ALIGNED
	DUP TO *recordfifos		JACK-CHANNELS  #jack-fifos #fifosize * * + $1000 #B-ALIGNED
	DUP TO jack_position_t		$100 +
	DUP TO midi_in_queue		#midi-event-buffer +
	DUP TO midi_out_queue		#midi-event-buffer + ALIGNED
	DUP TO *wboard-handles		#wboard-handle JACK-BOARDJOBS * + ALIGNED
	TO *jack-high-mem
	[ AUTO-HISTORY-CLEAR? ] [IF] *historybuffers  JACK-CHANNELS  #HISTORY-SIZE * SFLOATS ERASE [THEN]
	*io-plug-slots *jack-high-mem OVER - ERASE
	0 TO fJACK-ID 1 TO wboard-ticket
	*wboard-handles #wboard-handle JACK-BOARDJOBS * ERASE
	INIT-PLUGINS ; DUP EXECUTE AtCold


FORTH DEFINITIONS
: SET-CHANNEL-VOLUME	\ ( channel -- ) (F level -- )
\ *G Besides setting the global volume you may set a volume to a specified output channel by this function.
	TO >outport-volume ;
: GET-CHANNEL-VOLUME	\ ( channel -- ) (F -- level )
\ *G reads the channels output volume.
	>outport-volume ;

\ *S Managing the history buffers
: OBJECT-HISTORY	( plugin-object|channel -- addr )
\ *G To use *\forth{@HISTORY-DATA}, *\forth{MOVE-HISTORY-DATA}, *\forth{@IDX-HISTORY-DATA}, *\forth{RECONNECT-PLUGIN}
\ ** or *\forth{@INTERPOLATED-HISTORY-DATA}
\ ** you need the history buffers base address, this calculates it for you.
\ ** The parameter can be either the channel getting the inport history buffer or the plugin object, then
\ ** the plugins private history buffers is returned.
\ ** As the history holds sfloat data there is a very small precision loss compared to pre-0.115 versions.
\ *P There are always *\forth{JACK-CHANNELS} history buffers for the input ports and *\forth{#PLUGINS} buffers
\ ** for the active plugin objects allocated. The size of each history buffers is *\forth{#HISTORY-SIZE SFLOATS}.
\ ** See for the config.frt file to change the default values.
\ ** Each history holds several groups of *\forth{JACK-CHKSIZE SFLOATS} segments like shown in *\bold{fjack/history.png}
	DUP JACK-CHANNELS < IF >inhistory ELSE d_plug_output + @ THEN ;


: *N-HISTORY		( history idx -- addr )	
\ *G Calculates the address inthe history buffer
	#HISTORY-SIZE 1- AND SFLOATS + ;
: @IDX-HISTORY-DATA	( history idx -- ) ( F -- data )
\ *G Reads the audio data in the history buffer at position idx.
	*N-HISTORY SF@ ;
: @HISTORY-DATA		( history offset -- ) ( F -- data )
\ *G Reads the audio data in the history buffer relative to the current position in the audio stream.
	#FJACK-SAMPLE SWAP - *N-HISTORY SF@ ;

: MOVE-HISTORY-DATA	( historybuffer startoffset size target-addr -- )
\ *G Copies size samples of audio data from the history buffer to the target buffer. The offset marks the end of data.
	0 0 LOCALS| 1st 2nd target size offset buffer |
	#FJACK-SAMPLE offset - size -  #HISTORY-SIZE 1- AND TO offset
	#HISTORY-SIZE offset - size MIN TO 1st
	size 1st - TO 2nd
	1st 0> IF offset SFLOATS buffer +  target  1st SFLOATS MOVE THEN
	2nd 0> IF buffer  target 1st SFLOATS + 2nd SFLOATS MOVE THEN ;

: @INTERPOLATED-HISTORY-DATA	\ ( historybuffer -- ) (F time-offset -- data )
\ *G You may read the interpolated audio data from the history buffer by this,
\ ** the *\forth{time-offset} is measured in seconds. a zero time offset always references
\ ** to the currently processed sample in the input->output processing chain
\ *P so it is *\bold{only} valid there../waf configure --prefix=/usr --mixed --libdir=/usr/lib64 --alsa --firewire --dbus
	#FJACK-SAMPLE JACK-SAMPLE-F-RATE F*		\ offset
	FDUP FTRUNC FDUP F>S - F-			\ buffer index (F fraction )
	FDUP 1.0E-10 F<
	IF	FDROP *N-HISTORY SF@
	ELSE	1- 2DUP *N-HISTORY SF@ 1+ *N-HISTORY SF@
		FOVER FSWAP F- FROT F* F+
	THEN ;

\ *S Managing time
: JACK-TIME?		( -- d )
\ *G The time of the JACK server is always 64bit, so for portability it's a double in Forth.
\ ** The time is derived from the high resolution timer, it is measured in usec.
\ ** This is a 'real-time' when you ask for it, it is the servers time.
	jack_get_time [ 64BIT? ] [IF] S>D [THEN] ;

: JACK-PROCESS-TIME?	( -- d )
\ *G The time of the JACK server is always 64bit, so for portability it's a double in Forth.
\ ** The time is derived from the high resolution timer, it is measured in usec.
\ ** This time is not the 'real-time' but the real-time when the last processing callback for all
\ ** connected clients started.
	jack_position_t #08 + [ 64BIT? ] [IF] @ S>D [ELSE] 2@ SWAP [THEN] ;

: USEC>SAMPLES	( d-usec -- samples )
\ *G Converts the double number usec to samples.
	JACK-SAMPLE-RATE #1000000 M*/ D>S ;
: SAMPLES>USEC	( samples -- d-usec )
\ *G Converts samples to the double number usec.
	#1000000 JACK-SAMPLE-RATE M*/ D>S ;

: JACK-UNIQUE		( -- d )
\ *G This is a unique double telling about the position in JACKs processing queue. See for more details
\ ** in the libjack sources or API documentation, not much use for this in most situations.
	jack_position_t [ 64BIT? ] [IF] @ S>D [ELSE] 2@ SWAP [THEN] ;

\ *S Managing blocks of 'WAV' data
0 VALUE WAV-PLAY-MODE
\ *G When playing wav data the audio samples are inserted in the input->output procesing stream.
\ ** This can be before or after the signal processing is done. *\forth{WAV-PLAY-MODE} is used to switch
\ ** between these modes , n=0 means before processing.

1 VALUE WAV-READ-MODE
\ *G When recording wav data from the input->output processing stream the samples can be fetched from
\ ** before or after are signal processing.
\ ** *\forth{WAV-READ-MODE} is used to switch between these modes , n=0 means before processing.

\ *P So wav-playing and recording uses history buffers selected by the values of *\forth{WAV-PLAY-MODE} and *\forth{WAV-READ-MODE}.
\ ** In this graphics using a value of 0 uses the green buffers, red buffers otherwise.
\ ** The picture *\bold{fjack/waving.png} might help you to understand.

: CALC-WAV-RATE		( srate -- true-sample-rate )
\ *G The sample rate conversion when playing or recording wav data uses a rather simple linear
\ ** approximation algorithm, the used 'real' sample rates can be any *\forth{truerate = n * samplerate / JACK-CHKSIZE}.
\ ** This ensures a small distortion and no glitches are in recorded or played data.
\ ** Another point is: there are better buffersizes than others!
\ ** When you want to use resampling, *\forth{CALC-WAV-RATE} tells the true sample rate
\ ** calculated from the desired one.
\ *P This is only relevant for *\forth{PLAY-WAV RECORD-WAV}.
	JACK-CHKSIZE JACK-SAMPLE-RATE */  JACK-SAMPLE-RATE JACK-CHKSIZE */ ;

: CALC-WAV-BLOCK	( srate	channels mode -- n )
\ *G As told at *\forth{CALC-WAV-RATE} there are 'better buffersizes than others'.
\ ** *\forth{CALC-WAV-BLOCK} gives a good suggestion what the block size should be, or better
\ ** multiples of this.
	8 / * ABS SWAP  JACK-CHKSIZE JACK-SAMPLE-RATE */ * ;

: WAV-MODE?	( mode -- flag )
\ *G Flag is true when the wav mode is supported by this system.
	DUP #8 = OVER #16 = OR SWAP ABS #32 = OR ;

: PLAY-WAV	( addr size rate channels mode -- id ?error )
\ *G Plays WAV data available at an addr and a size. These data are not played immediately
\ ** but are inserted into a FIFO 'slot' and will be played after all before playing commands have
\ ** finished.
\ ** When size is negative the FIFO is cleared before inserting this block into the FIFO so playing
\ ** will start at once. More precise - it will start when the next *\forth{PROCESS-FRAMES} callback is taken.
\ ** A negative channel number results in reversed channels when listening, playing wav files with
\ ** only one channel will be played as stereo.
\ ** Thew rate can be freely choosen, the mode can be any of the 8bit/16bit/32bit/32bit-float
\ ** modes supported here. Named constants for these are defined in *\forth{JACKAUDIO}
\ ** these are defined *\forth{STEREO MONO 8BIT-AUDIO 16BIT-AUDIO 32BIT-AUDIO SFLOAT-AUDIO}.
\ *P  The 'id' can be used to test a block of wav data in the FIFO. See: *\forth{WAV-PLAYING? STOP-PLAY-WAV},
\ ** ?error tells about the status
\ *B 0 -- everything was ok and id is 'legal'
\ *B 1 -- not supported mode
\ *B 2 -- too many channels
\ *B 3 -- FIFO overflow
\ *B 4 -- JACK is not running or has stopped
	LOCALS| mode channels frate size addr |
	JACK-RUNNING? 0= IF 4 TRUE EXIT THEN
	mode WAV-MODE? 0= IF 1 TRUE EXIT THEN	\ mode must be 8,16,32,-32
	channels ABS 1- 0 JACK-CHANNELS WITHIN 0= IF 2 TRUE EXIT THEN		\ check for channels
	size 0<
	IF	#play-newest TO #play-current
	ELSE	#play-newest #play-current -  #jack-fifos 2- > IF 3 TRUE EXIT THEN
	THEN
	mode 8 / TO mode
	JACK-CHANNELS 0
	DO	I *N-PLAY-FIFO -->FIFO
		0							TO >f-todo
		0							TO >f-done
		mode							TO >f-mode
		mode channels * ABS					TO >f-incr
		channels ABS 2 MAX					TO >f-channels
		frate JACK-CHKSIZE JACK-SAMPLE-RATE */			TO >f-fract
		I channels ABS 2 MAX  <
		IF size mode / channels / ABS  JACK-CHKSIZE >f-fract */	TO >f-todo THEN
		addr I channels ABS 1- MIN mode ABS * +			TO >f-addr
	LOOP
	channels 0<
	IF	channels ABS 2/ 0
		?DO	I *N-PLAY-FIFO -->FIFO >f-addr		channels ABS 1- I - *N-PLAY-FIFO -->FIFO >f-addr
			I *N-PLAY-FIFO -->FIFO TO >f-addr	channels ABS 1- I - *N-PLAY-FIFO -->FIFO TO >f-addr
		LOOP
	THEN
	#play-newest DUP 1+ TO #play-newest
	FALSE ;

: READ-WAV	( addr size rate channels mode -- id ?error )
\ *G *\forth{READ-WAV} works as *\forth{PLAY-WAV} except it reads the data from the audio stream and
\ ** puts them into the buffer at addr. More info at *\forth{PLAY-WAV}
	LOCALS| mode channels frate size addr |
	JACK-RUNNING? 0= IF 4 TRUE EXIT THEN
	mode WAV-MODE? 0= IF 1 TRUE EXIT THEN	\ mode must be 8,16,32,-32
	channels 1- 0 JACK-CHANNELS WITHIN 0= IF 2 TRUE EXIT THEN		\ check for channels
	size 0<
	IF	#read-newest TO #read-current
	ELSE	#read-newest #read-current -  #jack-fifos 2- > IF 3 TRUE EXIT THEN
	THEN 
	mode 8 / TO mode
	JACK-CHANNELS 0
	DO	I *N-READ-FIFO -->FIFO
		0							TO >f-todo
		0							TO >f-done
		mode							TO >f-mode
		mode channels *	ABS					TO >f-incr
		channels ABS						TO >f-channels
		frate JACK-CHKSIZE JACK-SAMPLE-RATE */			TO >f-fract
		I channels ABS <

		IF size mode / channels / ABS JACK-CHKSIZE >f-fract */	TO >f-todo THEN
		addr I channels ABS 1- MIN mode ABS * +			TO >f-addr
	LOOP
	channels 0<
	IF	channels ABS 2/ 0
		?DO	I *N-PLAY-FIFO -->FIFO >f-addr		channels ABS 1- I - *N-PLAY-FIFO -->FIFO >f-addr
			I *N-PLAY-FIFO -->FIFO TO >f-addr	channels ABS 1- I - *N-PLAY-FIFO -->FIFO TO >f-addr
		LOOP
	THEN
	#read-newest DUP 1+ TO #read-newest FALSE ;

: STOP-PLAY-WAV	( ms id -- )
\ *G The data that had earlier been put into the play FIFO and had returned 'id' will stop
\ ** playing after the specified time ms (milliseconds). When the original block had a shorter duration all
\ ** data will be played.
	JACK-RUNNING? 0= IF 2DROP EXIT THEN
	DUP #play-current #play-newest WITHIN 0= IF 2DROP EXIT THEN		\
	JACK-CHANNELS 0
	DO	OVER JACK-SAMPLE-RATE #1000 */	OVER I SWAP *PLAY-FIFO -->FIFO
		>f-todo MIN 0 MAX TO >f-todo
	LOOP	2DROP ;
: STOP-READ-WAV	( ms id -- )
\ *G The data that had earlier been put into the read FIFO and had returned 'id' will stop
\ ** reading into this buffer after the specified time ms (milliseconds). When the original block had a shorter duration all
\ ** data read later won't be put into the buffer.
	JACK-RUNNING? 0= IF 2DROP EXIT THEN
	DUP #read-current #read-newest WITHIN 0= IF 2DROP EXIT THEN
	JACK-CHANNELS 0
	DO	OVER JACK-SAMPLE-RATE #1000 */	OVER I SWAP *READ-FIFO -->FIFO
		>f-todo MIN 0 MAX TO >f-todo
	LOOP 2DROP ;

\ *N Checking for blocks of wav data
: WAV-PLAYING?		( id -- val )
\ *G Checks the status of the played wav block 'id'.
\ ** 'val' is the remaining time until this block is played completely or zero when the fifo block is unused.
\ ** A typical piece of code code be
\ *C ( id ) BEGIN DUP WAV-PLAYING? DUP MS 0= UNTIL DROP
	JACK-RUNNING? 0= IF DROP 0 EXIT THEN
	DUP #play-current #play-newest WITHIN 0= IF DROP 0 EXIT THEN
	1 SWAP 1+ #play-current ?DO 0 I *PLAY-FIFO -->FIFO >f-todo #1000 JACK-SAMPLE-RATE */ 0 MAX + LOOP ;


: WAV-READING?		( id -- val )
\ *G Checks the status of the recorded wav block 'id'.
\ ** 'val' is the remaining time until this block is recorded completely or zero when the fifo block is unused. See *\forth{ WAV-PLAYING?}
	JACK-RUNNING? 0= IF DROP 0 EXIT THEN
	DUP #read-current #read-newest WITHIN 0= IF DROP 0 EXIT THEN
	1 SWAP 1+ #read-current ?DO 0 I *READ-FIFO -->FIFO >f-todo #1000 JACK-SAMPLE-RATE */ 0 MAX + LOOP ;


\ *N Using the Wavboard
\ *P The Wavboard is a new feature in fJACK 1.10. You can now play wav data directly from memory without
\ ** the other wav playing sheme.
\ *P You can play up to *\forth{JACK-BOARDJOBS} sounds at the same time - this is a config option -,
\ ** you may also auto-repeat them and define a starting time delay.

: START-WAVBOARD-JOB	( delay firstchannel repeats address size channels wavmode -- id )	( F: speedup -- )
\ *G This starts playing the wavdata from memory after *\bold{delay} milliseconds, it will be repeated *\bold{repeats} times.
\ ** When delay is negative the absolute value is the number of samples to be delayed. All negative values between -15 and -1
\ ** should not be used as these are reserved for extensions. Also when delay is negative the returned value of *\forth{WAVBOARD-PLAYING?}
\ ** will be the number of samples.
\ ** *\bold{firstchannel} is the lowest fjack channel to be used.
\ ** *\bold{address} points to the data, *\bold{size} is the size of the wavdata block in bytes.
\ ** *\bold{wavmode} is the same as for other wavdata shown in *\forth{PLAY-WAV}.
\ *P Playing such wav data can be done at any sample rate; conversion is done automatically from wav-rate to
\ ** *\forth{JACK-SAMPLE-RATE}, an extra tuning with *\bold{speedup}. Please note that
\ ** wavdata at a) the original samplerate and b) 1-4 channels have a much lower jack load
\ ** so it might be worth to do conversion before. As the whole block of data is already available
\ ** there is no restriction for wavrates
	get-wb-ticket 0 0 0 0 LOCALS| #scale startpos samples #one ticket mode channels size address repeats lowest delay |
	FDUP 0E0 F< IF FDROP #wbdscale ELSE #wbdscale S>F F* F>S THEN TO #scale
	mode WAV-MODE? JACK-RUNNING? AND channels 0<> AND 0= IF 0 EXIT THEN
	mode 8 / channels * ABS TO #one
	size #one / 1 - #one * TO size
	size #one / #wbdscale #scale */  DUP TO samples 4 < IF 0 EXIT THEN
	delay DUP 0< IF	ABS ELSE JACK-SAMPLE-RATE #1000 */ THEN
	JACK-CHKSIZE / 1+ JACK-CHKSIZE * JACK-POSITION + TO startpos
	channels ABS 0
	DO	0 JACK-BOARDJOBS 0
		DO	*wboard-handles I #wboard-handle * + -->WBD
			wbd-start-frame JACK-POSITION < IF DROP TRUE LEAVE THEN	\ already done so this slot is free
		LOOP	\ when the flag is true a free ticket had been found
		0= IF LEAVE THEN
		channels ABS 1 5 WITHIN IF channels ELSE 0 THEN TO wbd-fastok
		#scale #wbdscale = IF wbd-fastok ELSE 0 THEN TO wbd-fast
		#one TO wbd-#one
		delay 0< TO wbd-frames?
		ticket TO wbd-id
		size #one / TO wbd-truesamples
		mode TO wbd-wavmode
		repeats TO wbd-repeats
		#scale DUP TO wbd-scale DUP TO wbd-orgscale TO wbd-newscale
		startpos TO wbd-start-frame
		I channels 0< IF NEGATE THEN lowest + JACK-CHANNELS MOD TO wbd-channel
		address  mode ABS 8 / I * + TO wbd-address
		samples DUP TO wbd-todo TO wbd-samples
		0 TO wbd-stop-frame
		0 TO wbd-newpos
	LOOP ticket ;
: GET-WAVBOARD-POSITION	( id -- done total )
\ *G You can read the number of samples done already and total. Both values are corrected
\ ** using the used *\forth{JACK-SAMPLE-RATE}
	LOCALS| ticket | 0 0
	JACK-RUNNING? 0= IF EXIT THEN
	JACK-BOARDJOBS	0
	DO	*wboard-handles I #wboard-handle * + -->WBD
		ticket wbd-id =
		IF	2DROP wbd-samples DUP wbd-todo - SWAP LEAVE
		THEN
	LOOP ;
JACKAUDIO DEFINITIONS
: (remaining-time)
	wbd-stop-frame 0=
	IF	wbd-repeats S>F wbd-samples S>F F*
		wbd-start-frame JACK-POSITION - 0 MAX S>F F+
		wbd-todo S>F F+
	ELSE	wbd-stop-frame JACK-POSITION - 0 MAX S>F
	THEN
	wbd-frames? 0= IF #1000 S>F F* JACK-SAMPLE-F-RATE F/ THEN 1E0 FMAX $7FFFFFFF S>F FMIN F>S ;
FORTH DEFINITIONS
: SET-WAWBOARD-POSITION	( newpos id -- )
\ *G Sets the playing position for this wavboard job. *\bold{newpos} is the new position,
\ ** see also *\forth{GET-WAVBOARD-POSITION}
	LOCALS| ticket newpos |
	JACK-RUNNING? 0= IF EXIT THEN
	JACK-BOARDJOBS	0
	DO	*wboard-handles I #wboard-handle * + -->WBD
		ticket wbd-id = IF newpos 1+ TO wbd-newpos THEN
	LOOP ;

: SHOW-WAVBOARD	( -- )
\ *G Lists active wavboard jobs
	JACK-RUNNING? 0= IF EXIT THEN
	JACK-BOARDJOBS	0
	DO	*wboard-handles I #wboard-handle * + -->WBD
		wbd-id
		IF	CR ." id: "	wbd-id H.
			." , samples: "	wbd-samples H.
			." , todo: "	wbd-todo H.
			." , time: "	(remaining-time) DEC.
			." , speed: "	wbd-scale S>F #wbdscale S>F F/ F.
		THEN
	LOOP ;

: STOP-WAVBOARD-JOB	( delay id -- )
\ *G A wavboard job started earlier can be stopped by this function, it requires the job id and a delay time.
\ ** delay can also be negative as described in *\forth{START-WAVBOARD-JOB}
	SWAP DUP 0< IF ABS ELSE JACK-SAMPLE-RATE #1000 */ THEN JACK-CHKSIZE 1- -1 XOR AND LOCALS| rest ticket |
	JACK-RUNNING? 0= IF EXIT THEN
	JACK-BOARDJOBS	0
	DO	*wboard-handles I #wboard-handle * + -->WBD
		ticket wbd-id =
		IF JACK-POSITION rest + TO wbd-stop-frame 0 TO wbd-repeats THEN
	LOOP ;
: SPEEDUP-WAVBOARD-JOB	( id -- ) ( F: speedup -- )
\ *G A wavboard job can be given a new samplerate, the given speedup is multiplied with the speedup used
\ ** when the wavboard was started.
	LOCALS| ticket |
	JACK-RUNNING? 0= IF FDROP EXIT THEN
	JACK-BOARDJOBS	0
	DO	*wboard-handles I #wboard-handle * + -->WBD
		ticket wbd-id =
		IF	FDUP #wbdscale S>F F* F>S wbd-orgscale #wbdscale */ $1 MAX $7FFFFFFF MIN TO wbd-newscale
		THEN
	LOOP FDROP ;

: WAVBOARD-PLAYING?		( id -- time )
\ *G Checks the status of the played wavboard 'id'.
\ ** time is the remaining time until this wavboard sound is played completely or zero if it couldn't be found.
\ ** You could use this to start another wavblock right after this one.
\ ** At least 1 is returned as val, the maximum is $7FFFFFFF. time can be either the time in msec or in samples, this depends
\ ** the delay parameter when the wavboard job had started.
	LOCALS| ticket | JACK-RUNNING? 0= IF 0 EXIT THEN
	0 JACK-BOARDJOBS 0
	DO	*wboard-handles I #wboard-handle * + -->WBD
		ticket wbd-id =
		IF	DROP (remaining-time) LEAVE
		THEN
	LOOP ;

: STOP-ALL-WAVBOARD	( -- )
\ *G You may also stop all jobs right now.
	JACK-RUNNING? 0= IF EXIT THEN
	JACK-BOARDJOBS	0
	DO	*wboard-handles I #wboard-handle * + -->WBD
		0 TO wbd-start-frame   0 TO wbd-id   0 TO wbd-repeats
	LOOP ;

JACKAUDIO DEFINITIONS
\ Internal function that playes at most frames data from addr in the FIFO struct.
: WAVDATA->BUFF	( addr frames channel -- )
	DUP *C-PLAY-FIFO -->FIFO >f-incr >f-fract >f-addr LOCALS| *in #fract incr channel frames *out |
	\ Allow only the remaining data in the buffer when it's less than frames
	>f-todo frames MIN DUP TO frames 0<= IF EXIT THEN
 	#fract JACK-CHKSIZE =
 	IF	*out  *in >f-done incr * +  frames 0
 		>f-mode
 		CASE	1 OF DO	DUP WAV8@ #P-SCW8 F*		incr +	SWAP SF+!+ SWAP	LOOP ENDOF
 			2 OF DO	DUP WAV16@ #P-SCW16 F*		incr +	SWAP SF+!+ SWAP	LOOP ENDOF
 			4 OF DO	DUP WAV32@ #P-SCW32 F*		incr +	SWAP SF+!+ SWAP	LOOP ENDOF
 		       -4 OF DO	DUP SF@ PLAY-WAV-LEVEL F*	incr + 	SWAP SF+!+ SWAP	LOOP ENDOF
 		ENDCASE	2DROP 0E0 channel TO >lastwriteval
 	ELSE
		*out			\ JACK *data address
		>f-done frames BOUNDS	\ loop parameters to be executed
		>f-mode
		CASE	1 OF	DO	I #fract JACK-CHKSIZE */MOD DUP 0=
					IF	DROP channel >lastwriteval *in WAV8@
					ELSE	1- incr * *in + DUP WAV8@ incr + WAV8@ FDUP channel TO >lastwriteval
					THEN	FOVER F- S>F JACK-CHKSIZE S>F F/ F* F+ #P-SCW8 F* SF+!+
				LOOP
			ENDOF
			2 OF	DO	I #fract JACK-CHKSIZE */MOD DUP 0=
					IF	DROP channel >lastwriteval *in WAV16@
					ELSE	1- incr * *in + DUP WAV16@ incr + WAV16@ FDUP channel TO >lastwriteval
					THEN	FOVER F- S>F JACK-CHKSIZE S>F F/ F* F+ #P-SCW16 F* SF+!+
				LOOP
			ENDOF
			4 OF	DO	I #fract JACK-CHKSIZE */MOD DUP 0=
					IF	DROP channel >lastwriteval *in WAV32@
					ELSE	1- incr * *in + DUP WAV32@ incr + WAV32@ FDUP channel TO >lastwriteval
					THEN	FOVER F- S>F JACK-CHKSIZE S>F F/ F* F+ #P-SCW32 F* SF+!+
				LOOP
			ENDOF
			-4 OF	DO	I #fract JACK-CHKSIZE */MOD DUP 0=
					IF	DROP channel >lastwriteval *in SF@
					ELSE	1- incr * *in + DUP SF@ incr + SF@ FDUP channel TO >lastwriteval
					THEN	FOVER F- S>F JACK-CHKSIZE S>F F/ F* F+ PLAY-WAV-LEVEL F* SF+!+
				LOOP
			ENDOF
		ENDCASE	DROP
	THEN
	frames DUP +TO >f-done NEGATE >f-todo + 0 MAX TO >f-todo ;

: BUFF->WAVDATA	( historybuff addr frames channel -- )
	DUP *C-READ-FIFO -->FIFO >f-incr >f-fract 0 LOCALS| *out #fract incr channel frames *in buff |
	>f-addr >f-done  #fract JACK-CHKSIZE */ incr * + TO *out	\ in wav block
	>f-todo frames MIN DUP TO frames 0<= IF EXIT THEN		\ correct frames
	#fract JACK-CHKSIZE =
	IF	*out *in frames 0
	 	>f-mode
		CASE	1 OF DO	SF@+ #R-SCW8 F*		SWAP DUP WAV8!  incr + SWAP	LOOP ENDOF
			2 OF DO	SF@+ #R-SCW16 F*	SWAP DUP WAV16! incr + SWAP	LOOP ENDOF
			4 OF DO	SF@+ #R-SCW32 F*	SWAP DUP WAV32! incr + SWAP	LOOP ENDOF
		       -4 OF DO	SF@+ READ-WAV-LEVEL F*	SWAP DUP SF! incr + SWAP	LOOP ENDOF
		ENDCASE	2DROP
	ELSE	*out frames #fract JACK-CHKSIZE */ 0	( addr cnt 0 )
		>f-mode
		CASE	1 OF	DO	I JACK-CHKSIZE #fract */MOD 1- SFLOATS *in +
					DUP buff < IF DROP buff #HISTORY-SIZE 1- SFLOATS + SF@ buff SF@ ELSE SF@+ SF@ THEN
					FOVER F- S>F #fract S>F F/ F*  F+ #R-SCW8 F*  DUP WAV8! incr +
				LOOP
			ENDOF
	 		2 OF	DO	I JACK-CHKSIZE #fract */MOD 1- SFLOATS *in +
					DUP buff < IF DROP buff #HISTORY-SIZE 1- SFLOATS + SF@ buff SF@ ELSE SF@+ SF@ THEN
					FOVER F- S>F #fract S>F F/ F*  F+ #R-SCW16 F*  DUP WAV16! incr +
				LOOP
			ENDOF
	 		4 OF	DO	I JACK-CHKSIZE #fract */MOD 1- SFLOATS *in +
					DUP buff < IF DROP buff #HISTORY-SIZE 1- SFLOATS + SF@ buff SF@ ELSE SF@+ SF@ THEN
					FOVER F- S>F #fract S>F F/ F*  F+ #R-SCW32 F*  DUP WAV32! incr +
				LOOP
			ENDOF
	 		-4 OF	DO	I JACK-CHKSIZE #fract */MOD 1- SFLOATS *in +
					DUP buff < IF DROP buff #HISTORY-SIZE 1- SFLOATS + SF@ buff SF@ ELSE SF@+ SF@ THEN
					FOVER F- S>F #fract S>F F/ F*  F+ READ-WAV-LEVEL F*  DUP SF! incr +
	 			LOOP
			ENDOF
	 	ENDCASE	DROP
	THEN
	frames DUP +TO >f-done  NEGATE >f-todo + 0 MAX TO >f-todo ;	\ correct frames

\ Both next two words use the more primitive functions above,
\ they make sure that with the PROCESS-CALLBACK switching to the next FIFO is done
\ without any delay.
: WAVSTREAM->BUFF	( frames -- )
	0 LOCALS| done frames |
	BEGIN	BEGIN	#play-current #play-newest =		\ nothing to ber played
			IF	JACK-CHANNELS 0 DO 0e0 I TO >lastwriteval LOOP	EXIT
			THEN
			0 *C-PLAY-FIFO -->FIFO >f-todo 0<=		\ this fifo is empty?
		WHILE	1 +TO #play-current
		REPEAT
		0 *C-PLAY-FIFO -->FIFO  >f-todo frames MIN 0 MAX	( to-be-done )
		>f-channels JACK-CHANNELS MIN 0
		DO	I WAV-PLAY-MODE IF >outaudio ELSE >inaudio THEN done SFLOATS +
			frames done - I WAVDATA->BUFF
		LOOP	+TO done
		done frames >=
	UNTIL ;
: BUFF->WAVSTREAM	( frames -- )
	0 LOCALS| done frames |
	BEGIN	BEGIN	#read-current #read-newest =	\ no buffers in fifo to be filled any more ?
			IF EXIT THEN
			0 *C-READ-FIFO -->FIFO >f-todo 0<=
		WHILE	1 +TO #read-current
		REPEAT
		0 *C-READ-FIFO -->FIFO >f-todo frames MIN 0 MAX	( to-be-done )
		>f-channels JACK-CHANNELS MIN 0
		DO	WAV-READ-MODE IF I >outhistory I >outaudio ELSE I >inhistory I >inaudio THEN done SFLOATS +
			frames done - I BUFF->WAVDATA
		LOOP	+TO done
		done frames >=
	UNTIL ;

: SAFE-WAITING?	( -- )
\ Only safe within the io processing hook
	#FJACK-SAMPLE LOCALS| cnt |
	BEGIN	JACK-RUNNING? 0=
		#FJACK-SAMPLE cnt - ABS JACK-CHKSIZE 2 * > OR 0=
	WHILE	#10 MS
	REPEAT ;

: CORRECT-IO-LEVEL	( frames -- )
	JACK-CHANNELS 0
	DO	I >outport-volume JACK-IO-LEVEL F* FDUP 1E0 F=
		IF	FDROP
		ELSE	I >inaudio OVER 0 DO FDUP DUP SF@ F* SF!+ LOOP FDROP DROP
		THEN
	LOOP DROP ;

: DO-PROCESS-IO	( frames -- )
	>R R@ CORRECT-IO-LEVEL ACTIVE-PLUGINS
	BEGIN	@+ DUP
	WHILE	-->DSP JACK-POSITION TO #FJACK-SAMPLE
		data-offset DUP  plug_input + SWAP plug_output +	( in out -- )
		plug_bypassed
		IF	R@ SFLOATS MOVE
		ELSE	R@ 0
			DO	SWAP SF@+ SWAP plug_xt EXECUTE SF!+
				1 +TO #FJACK-SAMPLE
			LOOP 2DROP
		THEN
	REPEAT	2DROP R> DROP ;
: >wbuff	( i -- mod addr )		wbd-scale #wbdscale */MOD 	wbd-#one * wbd-address + ;
: calc-av	( mod -- ) ( F: v1 v2 -- v )	FOVER F- S>F #wbdscale S>F F/ F* F+ ;
: PROCESS-wboard-BLOCK
	wbd-todo DUP 1 < IF DROP TRUE EXIT THEN JACK-CHKSIZE MIN >R					\ how many to do now
	wbd-channel WAV-PLAY-MODE IF >outaudio ELSE >inaudio THEN	\ target
	wbd-samples wbd-todo - R@ BOUNDS
	wbd-wavmode
	CASE	#8 OF	DO I >wbuff DUP  WAV8@ wbd-#one +  WAV8@ calc-av #P-SCW8	F* SF+!+ LOOP	ENDOF
		#16 OF	DO I >wbuff DUP WAV16@ wbd-#one + WAV16@ calc-av #P-SCW16	F* SF+!+ LOOP	ENDOF
		#32 OF	DO I >wbuff DUP WAV32@ wbd-#one + WAV32@ calc-av #P-SCW32	F* SF+!+ LOOP	ENDOF
 		#-32 OF	DO I >wbuff DUP    SF@ wbd-#one +    SF@ calc-av PLAY-WAV-LEVEL	F* SF+!+ LOOP	ENDOF
 	ENDCASE	DROP
	wbd-todo R> - DUP TO wbd-todo 2 < ;				\ nothing more to do?
: fast-PROCESS-wboard-BLOCK
	wbd-todo DUP 2 < IF DROP TRUE EXIT THEN JACK-CHKSIZE MIN 2/ >R
	wbd-channel WAV-PLAY-MODE IF >outaudio ELSE >inaudio THEN	\ target
	wbd-samples wbd-todo - wbd-#one * wbd-address +			\ source
	wbd-wavmode wbd-fast 1- +
	CASE	#8 OF	R@ 0 DO DUP  WAV8@ #01 + DUP WAV8@  #01 + SWAP FSWAP #P-SCW8  F* SF+!+ #P-SCW8  F* SF+!+ SWAP LOOP ENDOF
		#9 OF	R@ 0 DO DUP WAV8@  #02 + DUP WAV8@  #02 + SWAP FSWAP #P-SCW8  F* SF+!+ #P-SCW8  F* SF+!+ SWAP LOOP ENDOF
		#10 OF	R@ 0 DO DUP WAV8@  #03 + DUP WAV8@  #03 + SWAP FSWAP #P-SCW8  F* SF+!+ #P-SCW8  F* SF+!+ SWAP LOOP ENDOF
		#11 OF	R@ 0 DO DUP WAV8@  #04 + DUP WAV8@  #04 + SWAP FSWAP #P-SCW8  F* SF+!+ #P-SCW8  F* SF+!+ SWAP LOOP ENDOF

		#16 OF	R@ 0 DO DUP WAV16@ #02 + DUP WAV16@ #02 + SWAP FSWAP #P-SCW16 F* SF+!+ #P-SCW16 F* SF+!+ SWAP LOOP ENDOF
		#17 OF	R@ 0 DO DUP WAV16@ #04 + DUP WAV16@ #04 + SWAP FSWAP #P-SCW16 F* SF+!+ #P-SCW16 F* SF+!+ SWAP LOOP ENDOF
		#18 OF	R@ 0 DO DUP WAV16@ #06 + DUP WAV16@ #06 + SWAP FSWAP #P-SCW16 F* SF+!+ #P-SCW16 F* SF+!+ SWAP LOOP ENDOF
		#19 OF	R@ 0 DO DUP WAV16@ #08 + DUP WAV16@ #08 + SWAP FSWAP #P-SCW16 F* SF+!+ #P-SCW16 F* SF+!+ SWAP LOOP ENDOF

		#32 OF	R@ 0 DO DUP WAV32@ #04 + DUP WAV32@ #04 + SWAP FSWAP #P-SCW32 F* SF+!+ #P-SCW32 F* SF+!+ SWAP LOOP ENDOF
 		#33 OF	R@ 0 DO DUP WAV32@ #08 + DUP WAV32@ #08 + SWAP FSWAP #P-SCW32 F* SF+!+ #P-SCW32 F* SF+!+ SWAP LOOP ENDOF
		#34 OF	R@ 0 DO DUP WAV32@ #12 + DUP WAV32@ #12 + SWAP FSWAP #P-SCW32 F* SF+!+ #P-SCW32 F* SF+!+ SWAP LOOP ENDOF
		#35 OF	R@ 0 DO DUP WAV32@ #16 + DUP WAV32@ #16 + SWAP FSWAP #P-SCW32 F* SF+!+ #P-SCW32 F* SF+!+ SWAP LOOP ENDOF

		#-32 OF	R@ 0 DO DUP SF@    #04 + DUP SF@    #04 + SWAP FSWAP PLAY-WAV-LEVEL F* SF+!+ PLAY-WAV-LEVEL F* SF+!+ SWAP LOOP ENDOF
 		#-31 OF	R@ 0 DO DUP SF@    #08 + DUP SF@    #08 + SWAP FSWAP PLAY-WAV-LEVEL F* SF+!+ PLAY-WAV-LEVEL F* SF+!+ SWAP LOOP ENDOF
 		#-30 OF	R@ 0 DO DUP SF@    #12 + DUP SF@    #12 + SWAP FSWAP PLAY-WAV-LEVEL F* SF+!+ PLAY-WAV-LEVEL F* SF+!+ SWAP LOOP ENDOF
 		#-29 OF	R@ 0 DO DUP SF@    #16 + DUP SF@    #16 + SWAP FSWAP PLAY-WAV-LEVEL F* SF+!+ PLAY-WAV-LEVEL F* SF+!+ SWAP LOOP ENDOF
  	ENDCASE	2DROP
	wbd-todo R> 2* - DUP TO wbd-todo 2 < ;				\ nothing more to do?

: PROCESS-WBOARD	( -- )
	JACK-POSITION TO #FJACK-SAMPLE
	*wboard-handles JACK-BOARDJOBS	0
	DO	DUP -->WBD
		wbd-newpos			\ set new position
		IF	wbd-samples wbd-newpos 1- - 0 MAX wbd-samples MIN TO wbd-todo 0 TO wbd-newpos
		THEN
		wbd-stop-frame JACK-POSITION =	\ to be stopped ?
		IF	0 TO wbd-start-frame  0 TO wbd-newscale 0 TO wbd-todo 0 TO wbd-id
		THEN
		wbd-newscale			\ setup scale
		IF	wbd-samples wbd-todo - 1+ wbd-scale wbd-newscale */        	\  new done
			wbd-truesamples #wbdscale wbd-newscale */ DUP TO wbd-samples
			SWAP - TO wbd-todo		\ true position
 			wbd-newscale TO wbd-scale 0 TO wbd-newscale
			wbd-scale #wbdscale = IF wbd-fastok ELSE 0 THEN TO wbd-fast
		THEN
		wbd-start-frame JACK-POSITION =	\ to be played now?
		IF	wbd-fast
			IF fast-PROCESS-wboard-BLOCK ELSE PROCESS-wboard-BLOCK THEN ( nothing more to do? )
			wbd-start-frame JACK-CHKSIZE + TO wbd-start-frame		\ probably again
			IF	wbd-repeats DUP 1- TO wbd-repeats 0>			\ start again
				IF	wbd-samples TO wbd-todo
				ELSE	0 TO wbd-start-frame 0 TO wbd-id
				THEN
			ELSE
			THEN
		THEN
		#wboard-handle +
	LOOP	DROP ;

: *midi		( n -- offset)	#midi-events 1- AND #midi-event-size * ;
: >midi_in	( n -- addr )	*midi midi_in_queue + ;
: >midi_out	( n -- addr )	*midi midi_out_queue + ;	\ addr to write midi data
: .midistruct	( str len struct -- )
	JACK-MIDI-DEBUG? 0= IF 3DROP EXIT THEN LOCALS| struct |
	<<DEBUG TYPE ."  Time:"
	struct 3 CELLS + @ (H.) TYPE ."  -- "
	BASE @ HEX
	struct DUP 2 CELLS + @ BOUNDS ?DO I C@ S>D <# # # [CHAR] $ HOLD BL HOLD #> TYPE LOOP CR DEBUG>>
	BASE ! ;

: PROCESS-MIDI	( frames -- )
	JACK-POSITION TO #FJACK-SAMPLE
	0 0 0 0 0 LOCALS| cnt struct event outbuff inbuff frames |
	0 >midiport frames jack_port_get_buffer TO inbuff
	1 >midiport frames jack_port_get_buffer TO outbuff
	outbuff jack_midi_clear_buffer
	#midi_out #midi_out_read - DUP TO cnt
	IF	frames 0
		DO	#midi_out #midi_out_read
			?DO	I >midi_out TO struct
				struct 3 CELLS + @  JACK-POSITION J + - 0 <=   struct 1 CELLS + @ 0= AND
				IF	S" Write MIDI struct" struct .midistruct
					outbuff J struct 2 CELLS + @ jack_midi_event_reserve
					struct SWAP OVER 2 CELLS + @ MOVE
					TRUE struct 1 CELLS + ! -1 +TO cnt
					cnt 0= IF LEAVE THEN
				THEN
			LOOP	cnt 0= IF LEAVE THEN
		LOOP
		cnt 0= IF #midi_out TO #midi_out_read THEN
	THEN
	#midi_in TO #midi_in_read
	inbuff jack_midi_get_event_count 0
	?DO	#midi_in 1+ >midi_in TO event			\ misusing a MIDI message buffer here for the event struct
		event inbuff I jack_midi_event_get 0=
		IF	#midi_in >midi_in TO struct
			event 32B@ JACK-POSITION + DUP TO #FJACK-SAMPLE struct 3 CELLS + !
			event 1 CELLS + 32B@ DUP struct 2 CELLS + ! >R
			event 2 CELLS + @  struct R@ 1 CELLS MIN MOVE
			R> 0 1 CELLS WITHIN
			IF	struct DO-PROCESS-MIDI-IN-EVENT 1 +TO #midi_in
				S" Read MIDI Event" struct .midistruct
			ELSE	S" Received an illegal MIDI event:"  struct .midistruct
			THEN
		THEN
	LOOP ;

: PROCESS-INSTRUMENTS
	top-instrument BEGIN ?DUP WHILE DUP @ SWAP instrument-action EXECUTE REPEAT ;

\ *S The Callbacks
\ *P These functions are *\bold{not} to be used in user applications. They are functions to be called by the callbacks defined for this client.
\ ** They could be used by testing tools to measure the exact performance of the io system but so far no such tools exist.
: XT-PROCESS-FRAMES	( frames parameter -- 0 )
\ *G The main processing callback; must return 0.
\ ** When the client is initialised the jack server starts a realtime thread, this thread later calls this callback and
\ ** *\bold{MUST NEVER} use any forth word that might add some sort of delay longer than a few ms. Otherwise the client might be terminated
\ ** or lots of xruns are signalled.
\ *P *\bold{So --- No semaphores! No disk-io! No terminal-io!}
\ *P The *\forth{XT-PROCESS-FRAMES} callback does *\bold{all} the signal processing defined by you in the input->output
\ ** processing slots. This practically limits things you can do in the processing chain, when writing your own plugins you
\ ** should keep this in mind. Data sampling to/from disk or network, signal analysis, graphics ... all must be done in a strictly
\ ** asynchronous way. As you can not safely use semaphores you should use shared memory for the data and thread-local data
\ ** for your 'housekeeping'.
	DROP 0 LOCALS| midievents frames |
	USE-DEBUG-I//O S" process" SET-TASKNAME
	fJACK-ID jack_position_t jack_transport_query TO #jack-state
	jack_position_t #20 + 32B@ DUP TO JACK-POSITION TO #FJACK-SAMPLE

	JACK-ZOMBIE? IF FALSE EXIT THEN			\ this really should never happen!!!
	#jack-state DUP JackTransportRolling <>  SWAP JackTransportLooping <>  AND	\ Output silence to avoid noise
	IF	JACK-CHANNELS 0 DO I >outport frames jack_port_get_buffer frames SFLOATS ERASE LOOP
		FALSE EXIT
	THEN
	PREP-SCALERS
	\ Get all pointers and store them in the *client struct
	JACK-CHANNELS 0
	DO	data-offset DUP  I >inhistory +  I TO >inaudio	I >outhistory +  I TO >outaudio
		I >inport jack_port_connected
		IF	I >inport frames jack_port_get_buffer  I >inaudio frames SFLOATS MOVE
		ELSE	I >inaudio frames SFLOATS ERASE
		THEN
	LOOP
	midi-action IF frames PROCESS-MIDI THEN
	instrument-action IF PROCESS-INSTRUMENTS THEN
	WAV-PLAY-MODE 0= JACK-POSITION 0<> AND	IF frames WAVSTREAM->BUFF PROCESS-WBOARD THEN
	frames DO-PROCESS-IO
	WAV-PLAY-MODE 0<> JACK-POSITION 0<> AND	IF frames WAVSTREAM->BUFF PROCESS-WBOARD THEN
	frames BUFF->WAVSTREAM
	JACK-CHANNELS 0 DO I >outaudio I >outport frames jack_port_get_buffer frames SFLOATS MOVE LOOP
	FALSE ;

: XT-GRAPH-REORDER	( 0 -- 0 )
\ *G Whenever the jack servers processing graph is changed this callback is executed. This happens when ports are disconnected
\ ** or a new jack application is starting. The user-defined *\forth{GRAPH-REORDER-HOOK ( -- )} is executed in this callback to allow you
\ ** to watch the surrounding.
	USE-DEBUG-I//O S" Graph" SET-TASKNAME DROP
	GRAPH-REORDER-HOOK FALSE ;
: XT-PORT-CONNECT	( a b parameter flag 0 -- 0 )
\ *G Whenever a port is connected or reconnected this callback is executed. You can find the fJACK ports with
\ ** *\forth{>inport ( idx -- id )} and *\forth{>outport ( idx -- id)}, both words are in *\forth{JACKAUDIO}.
\ ** idx is the channel, the returned id is unique and given by the jack server. I won't give any details here
\ ** how to use these id's or how to use the *\forth{PORT-CONNECT-HOOK} - go and read the jack manual, you will
\ ** need deeper informations to make use of it. *\forth{PORT-CONNECT-HOOK} is deferred. So your callback code must
\ ** work on these stack values
\ : YOUR-DEFINITION ( a b flag -- ) ....... ;
\ *G The prototype for the client supplied function that is called whenever a connection is changed.
\ ** a and b are the two ports to be connected or disconnected. The flag is non-zero if ports were connected and zero if ports were disconnected
	USE-DEBUG-I//O S" Connect" SET-TASKNAME DROP
	PORT-CONNECT-HOOK FALSE ;

: XT-NEW-FRAMES	( fr parameter -- 0 )
\ *G When some other client changes the chunksize this is told here. So far i don't know ANY client that does so
\ ** and JACK versions up to 0.109 didn't allow this to happen at all. But the jacklib API defines it and
\ ** the latest JACK development goes in this direction - jackdmp derived sources. See *\forth{JACK-CHKSIZE}
	USE-DEBUG-I//O S" Frames" SET-TASKNAME DROP
	DUP JACK-CHKSIZE <> IF ." A new chunksize has been set:" DUP DEC. CR THEN
	DUP TO JACK-CHKSIZE STOP-ALL-WAVBOARD FALSE ;

: XT-NEW-SAMPLERATE ( sr parameter -- FALSE )
\ *G The callback when other clients change the sampling rate for the jackd. See *\forth{JACK-SAMPLE-RATE}
	USE-DEBUG-I//O S" Srate" SET-TASKNAME DROP
	DUP JACK-SAMPLE-RATE <> IF ." A new rate has been set; is this ok for your audio hardware? New rate: " DUP DEC. CR THEN
	DUP TO JACK-SAMPLE-RATE S>F TO JACK-SAMPLE-F-RATE STOP-ALL-WAVBOARD FALSE ;

: XT-CLIENT-STOPPED	( parameter -- 0 )
\ *G A client that is forced by some other client to shut down tells all it's registered clients
\ ** about the new situation.
	USE-DEBUG-I//O S" Zombie" SET-TASKNAME DROP ." The client is zombified by the server" BOLD CR
	JackZombie TO #jack-state FALSE ;

: XT-COUNT-XRUNS	( parameter -- 0 )
\ *G When the JACK server gets an interrupt by the audio master clock to do new processing and
\ ** the queue of callbacks of the last interrupt has not been fully processed, this condition
\ ** is noted to all clients - it's called a 'XRUN'. See *\forth{JACK-XRUNS}
	DROP 1 +TO JACK-XRUNS FALSE ;

: XT-NEW-TRANSPORT	( s p dummy -- f )
\ *G A Repositioning callback, so far the position is not used.
	USE-DEBUG-I//O S" Transport" SET-TASKNAME DROP
	2DUP ." The server requested a repositioning to: $"
	1 ($.R) TYPE  ." , state: $" 1 ($.R) TYPE CR
	DROP TO #jack-state TRUE ;

: XT-SERVER-ERROR ( str -- str )
\ *G The callback function to receive a jack server error message.
	USE-DEBUG-I//O S" Jackd" SET-TASKNAME
	DUP ." Error: " C>FSTRING TYPE BOLD CR ;

: XT-SERVER-INFO ( str -- str )
\ *G The callback to receive a jack server information message.
	USE-DEBUG-I//O S" Jackd" SET-TASKNAME
	DUP ." Info: " C>FSTRING TYPE CR ;

: SET-SESSION-COMMAND	( str -- )
	>R $FJACKSYSFILE COUNT R@ $100 + PLACE
	JACK-CLIENT-NAME jack_client_name_size $10 - MIN R@ $100 + PLACE+
	R@ $100 + COUNT R> SWAP 2DUP + 0 SWAP C! CMOVE ;

: DEFAULT-SESSION-SAVER	( event -- )
	DUP @+ SWAP @+ SWAP @+ SWAP @ #512 + 0 LOCALS| fd buff uuid dir #evtype *event |
	dir C>FSTRING buff PLACE  S" README" buff PLACE+
	buff COUNT R/W CREATE-FILE IF DROP EXIT THEN TO fd
	S" This currently is just dummy data for testing" fd WRITE-FILE DROP  $CR COUNT fd WRITE-FILE DROP
	fd CLOSE-FILE DROP ;
	' DEFAULT-SESSION-SAVER IS SAVE-SESSION-DATA

: XT-SESSION-EVENT	( event *args -- )
\ *G The callback serving session events
	USE-DEBUG-I//O S" CB session" SET-TASKNAME DROP
	#1024 ALLOCATE IF DROP 0 THEN OVER 3 CELLS + !
	DUP @+ SWAP @+ SWAP @+ SWAP @ LOCALS| cmdline uuid dir #evtype *event |
	cmdline ?DUP IF SET-SESSION-COMMAND THEN
	." fJACK session event, command is: " cmdline COUNT TYPE
	#evtype
	CASE	JackSessionSave		OF *event SAVE-SESSION-DATA				."  JackSessionSave" ENDOF
		JackSessionSaveAndQuit	OF *event SAVE-SESSION-DATA TRUE TO JACK-SESSION-BYE?	."  JackSessionSaveAndQuit" ENDOF
		JackSessionSaveTemplate	OF *event SAVE-SESSION-DATA				."  JackSessionSaveTemplate" ENDOF
	ENDCASE	CR
	fJACK-ID *event jack_session_reply DROP
	*event jack_session_event_free ;
: check-capture-all
	0 0 LOCALS| minlat maxlat | 0 TO IN-MAXLATENCY
	JACK-CHANNELS 0
	DO	I >inport ?DUP
		IF 0 TO minlat 0 TO maxlat jack_port_get_connections ?DUP
			IF	DUP
				BEGIN	@+ DUP
				WHILE	fJACK-ID SWAP jack_port_by_name
					JackCaptureLatency *latencies jack_port_get_latency_range
					*latencies 32B@+ minlat MAX TO minlat 32B@ maxlat MAX TO maxlat
				REPEAT	2DROP
				jack_free
			THEN
			maxlat IN-MAXLATENCY MAX TO IN-MAXLATENCY
			minlat *latencies 32B!+ maxlat SWAP 32B!
			I >inport JackCaptureLatency *latencies	jack_port_set_latency_range
		THEN
	LOOP ;
: check-playback-all
	0 0 LOCALS| minlat maxlat | 0 TO OUT-MAXLATENCY
	CALC-LATENCIES ?DUP IF EXECUTE THEN
	JACK-CHANNELS 0
	DO	I >outport ?DUP
		IF 0 TO minlat 0 TO maxlat jack_port_get_connections ?DUP
			IF	DUP
				BEGIN	@+ DUP
				WHILE	fJACK-ID SWAP jack_port_by_name
					JackPlaybackLatency *latencies jack_port_get_latency_range
					*latencies 32B@+ minlat MAX TO minlat 32B@ maxlat MAX TO maxlat
				REPEAT	2DROP
				jack_free
			THEN
			maxlat OUT-MAXLATENCY MAX TO OUT-MAXLATENCY
			minlat I >latency + *latencies 32B!+ maxlat I >latency + SWAP 32B!
			I >inport JackCaptureLatency *latencies	jack_port_set_latency_range
		THEN
	LOOP ;
: XT-LATENCIES		( mode *args -- )
\ *G The latency callback
	DROP LOCALS| mode |
	mode
	CASE	JackPlaybackLatency	OF	check-playback-all	ENDOF
		JackCaptureLatency	OF	check-capture-all	ENDOF
	ENDCASE	;

IS-IFORTH? [IF]
' XT-PROCESS-FRAMES	CB( _int _int )CB-int PROCESS-FRAMES
' XT-NEW-FRAMES		CB( _int _int )CB-int NEW-FRAMES
' XT-NEW-SAMPLERATE	CB( _int _int )CB-int NEW-SAMPLERATE
' XT-CLIENT-STOPPED	CB( _int )CB-int CLIENT-STOPPED
' XT-COUNT-XRUNS 	CB( _int )CB-int COUNT-XRUNS
' XT-NEW-TRANSPORT	CB( _int _int _int )CB-int NEW-TRANSPORT
' XT-SERVER-ERROR	CB( _int )CB-int SERVER-ERROR
' XT-SERVER-INFO 	CB( _int )CB-int SERVER-INFO
' XT-GRAPH-REORDER	CB( _int )CB-int SERVER-REORDER
' XT-PORT-CONNECT	CB( _int _int _int _int )CB-int SERVER-PORT-CONNECT
' XT-SESSION-EVENT	CB( _int _int )CB-int SESSION-EVENT
' XT-LATENCIES		CB( _int _int )CB-int CHECK-LATENCY
[THEN]

IS-VFX? [IF]
fromc 2 1 CALLBACK: PROCESS-FRAMES	ASSIGN XT-PROCESS-FRAMES to-callback PROCESS-FRAMES
fromc 2 1 CALLBACK: NEW-FRAMES		ASSIGN XT-NEW-FRAMES to-callback NEW-FRAMES
fromc 2 1 CALLBACK: NEW-SAMPLERATE	ASSIGN XT-NEW-SAMPLERATE to-callback NEW-SAMPLERATE
fromc 1 1 CALLBACK: CLIENT-STOPPED	ASSIGN XT-CLIENT-STOPPED to-callback CLIENT-STOPPED
fromc 1 1 CALLBACK: COUNT-XRUNS		ASSIGN XT-COUNT-XRUNS to-callback COUNT-XRUNS
fromc 3 1 CALLBACK: NEW-TRANSPORT	ASSIGN XT-NEW-TRANSPORT to-callback NEW-TRANSPORT
fromc 1 1 CALLBACK: SERVER-ERROR	ASSIGN XT-SERVER-ERROR to-callback SERVER-ERROR
fromc 1 1 CALLBACK: SERVER-INFO		ASSIGN XT-SERVER-INFO to-callback SERVER-INFO
fromc 1 1 CALLBACK: SERVER-REORDER	ASSIGN XT-GRAPH-REORDER to-callback SERVER-REORDER
fromc 4 1 CALLBACK: SERVER-PORT-CONNECT	ASSIGN XT-PORT-CONNECT to-callback SERVER-PORT-CONNECT
fromc 2 1 CALLBACK: SESSION-EVENT	ASSIGN XT-SESSION-EVENT to-callback SESSION-EVENT
fromc 2 1 CALLBACK: CHECK-LATENCY	ASSIGN XT-LATENCIES to-callback CHECK-LATENCY
[THEN]


: ISOLATE-PORT		( addr len -- addr1 len1 )
	BEGIN	OVER C@ [CHAR] : <> WHILE 1- SWAP 1+ SWAP REPEAT ;

: START-CLIENT-ERROR	( id flag -- id flag )
	<<DEBUG CR 
	DUP JackFailure AND		IF ." Unspecified jack server failure" BOLD CR THEN
	DUP JackInvalidOption AND	IF ." Invalid JACK option" BOLD CR THEN
	DUP JackVersionError AND	IF ." Server/Client protocol mismatch, probably an old jack server in a 64/32 mixed system" BOLD CR THEN
	DUP JackShmFailure AND		IF ." Couldn't access shared memory; probably there is a crashed jack server" BOLD CR THEN
	DUP JackInitFailure AND		IF ." Unable to initialize the fjack forth client" BOLD CR THEN
	DUP JackLoadFailure AND		IF ." Unable to load internal client. This should not happen at all because fJACK is an external client" BOLD CR THEN
	DUP JackNoSuchClient AND	IF ." Requested client does not exist." BOLD CR THEN
	DUP JackServerError AND		IF ." Communication error with the jack server, probably a jack library problem" BOLD CR THEN
	DUP JackServerFailed AND	IF ." Server failure" BOLD CR THEN 
	DUP JackServerStarted AND	IF ." The jack server just started" CR	THEN
	DUP JackNameNotUnique AND	IF ." Selected another client name: " OVER (client_name) TYPE CR THEN DEBUG>> ;
: (test-ports)		( flag -- flag )
	JACK-CHANNELS 0 DO I >outport jack_port_get_connections DUP jack_free 0<> OR LOOP ;

: ?logerror	( flag addr len -- )
	ROT 0= IF 2DROP EXIT THEN <<DEBUG ." Couldn't install " TYPE CR DEBUG>> ;
: session-support?	( -- flag )
		JACK-SESSION? WINDOWS? 0= AND ;
FORTH DEFINITIONS
\ *S Connecting the ports
\ *P There are some graphical tools around to connect ports between the different JACK clients or the
\ ** playback or capture ports. You may use these nice tools to connect any ports but how can you
\ ** save and restore your configuration?
: ?CONNECT-JACKPORTS	\ ( flag -- ) name1 name2    - true means disconnect/ false connect
	{: disconn? | dest[ $100 ]  src[ $100 ] -- :}
	BL WORD COUNT  OVER 5 S" fJACK" COMPARE 0=		\ fJACK portname found
	IF	fJACK-ID (client_name) src[ PLACE
		ISOLATE-PORT src[ PLACE+
	ELSE	src[ PLACE
	THEN
	BL WORD COUNT OVER 5 S" fJACK" COMPARE 0=		\ fJACK port
	IF	fJACK-ID (client_name) dest[ PLACE		\ get the real clients name
		ISOLATE-PORT dest[ PLACE+			\ just add the :port part of the original
	ELSE	dest[ PLACE
	THEN
	<<DEBUG
	fJACK-ID  src[ COUNT dest[ COUNT disconn?
	IF	(jack_disconnect)
		IF ." Couldn't disconnect: " src[ COUNT TYPE  ." -->" dest[ COUNT TYPE THEN
	ELSE	(jack_connect) DUP 0 = SWAP #17 = OR 0=
		IF ." Couldn't connect: " src[ COUNT TYPE  ." -->" dest[ COUNT TYPE THEN
	THEN DEBUG>> ;

: SAVE-CONNECTIONS	( addr cnt -- )
\ *G writes a file specified by the name 'addr cnt' holding information about all ports connected to your fJACK client.
\ ** This file can later be used with *\forth{LOAD-CONNECTIONS} and *\forth{CLEAR-CONNECTIONS}.
\ ** Actually the file is a forth source.
\ *P Please make sure the directory fjack/sysfiles has writing enabled for you.
        JACK-RUNNING? 0= IF 2DROP <<DEBUG ." Can't save connections as client fJACK is not running" CR DEBUG>> EXIT THEN
	{: | fd buff[ $100 ] -- :}
	$FJACKSYSFILE COUNT buff[ PLACE  buff[ PLACE+ S" .frt" buff[ PLACE+
	buff[ COUNT R/W CREATE-FILE IF <<DEBUG ." The fjack/sysfiles directory must be writable!" BOLD CR DEBUG>> ABORT THEN
	TO fd
	JACK-CHANNELS 2 * 2 + 0
	DO	I >inport
		IF	I >inport jack_port_get_connections ?DUP
			IF	DUP
				BEGIN	DUP @
				WHILE	DUP @ C>FSTRING	\ ( *array *name len )
					I >inport (jack_port_name)	\ ( *array *name len *name2 len2 )
					S"      ?CONNECT-JACKPORTS " fd WRITE-FILE ?FILE
					I JACK-PORTFLAG JackPortIsInput AND IF 2SWAP THEN
					fd WRITE-FILE ?FILE S"  " fd WRITE-FILE ?FILE fd WRITE-FILE ?FILE
					$CR COUNT fd WRITE-FILE ?FILE
					CELL+
				REPEAT	DROP jack_free
			THEN
		THEN
	LOOP fd CLOSE-FILE ?FILE ;

: CONNECT-JACKPORTS	( -- )	\ source destination
\ *G Connects two ports, the portnames are read from the input stream.
	FALSE ?CONNECT-JACKPORTS ;
: UNCONNECT-JACKPORTS	( -- )	\ source destination
\ *G Disconnects two ports, the portnames are read from the input stream.
	TRUE ?CONNECT-JACKPORTS ;

: CLEAR-CONNECTIONS	( -- )
\ *G Unconnects all ports of the fJACK client.
        JACK-RUNNING? 0= IF EXIT THEN 
	JACK-CHANNELS 2 * 2 + 0
	DO	I >inport ?DUP IF fJACK-ID SWAP jack_port_disconnect DROP THEN
	LOOP ;

: ?LOAD-CONNECTIONS ( addr cnt flag -- )	\ flag=true unconnect / false connect
        JACK-RUNNING? 0= IF DROP 2DROP <<DEBUG ." Not loading connection file as fJACK client is not running" CR DEBUG>> THEN
	[CHAR] 0 SWAP IF 1+ THEN
	{: patchchar | fd buff[ $100 ] -- :}
	$FJACKSYSFILE COUNT buff[ PLACE buff[ PLACE+ S" .frt" buff[ PLACE+
	buff[ COUNT R/W OPEN-FILE
	IF	DROP <<DEBUG ." Can't load connections file: " buff[ COUNT TYPE CR DEBUG>> EXIT
	THEN	TO fd
	BEGIN	buff[ #200 fd READ-LINE DROP
	WHILE	\ the buffer has a space free for patching a flag
		patchchar buff[ C!
		buff[ SWAP EVALUATE
	REPEAT	DROP
	fd CLOSE-FILE DROP ;
: LOAD-CONNECTIONS	( addr len -- )
\ *G Connects all connections specified in the file.
	FALSE ?LOAD-CONNECTIONS ;
: UNLOAD-CONNECTIONS	( addr len -- )
\ *G Disconnects all connections specified in the file.
	TRUE ?LOAD-CONNECTIONS ;

\ *S Getting detailed information.
: SETUP-LATENCIES	( -- )
\ The current fJACK client (>1.12) takes care about setting latencies to be checked by other clients and the
\ jack server. *\forth{SETUP-LATENCIES} makes sure these data are available, you normally don't have to call it as the
\ plugin handler does so already.
	JACK-RUNNING? 0= IF EXIT THEN
	0 TO IN-MAXLATENCY
	0 TO OUT-MAXLATENCY
	JACK-LATENCY? 0= IF <<DEBUG ." SETUP-LATENCIES asked for latencies but library function is not available" BOLD CR DEBUG>> EXIT THEN
	fJACK-ID jack_recompute_total_latencies IF <<DEBUG ." Calculation of latencies was not successfull, returning defaults" CR DEBUG>> EXIT THEN ;
: JACK-LATENCIES	( -- max min )
\ *G Returns the latencies found for the fJACK client and it's ports.
	SETUP-LATENCIES #200 MS
	fJACK-MAXLAT DUP IN-MAXLATENCY OUT-MAXLATENCY + + SWAP ;

: JACKLOG	( -- )
\ *G Shows the jack servers and fjack client error messages. This works slightly different on vfx and iForth systems; on iForth we have
\ ** the thread-safe debugging device that also tells about exact timing of the messages. This might help a lot when debugging asynchronous
\ ** messages from callbacks or midi interfaces. On vfx the genio buffer device is used. For most situations this is fine.
\ *P On all systems supported by fJACK you can be sure *\bold{any} debugging info will *\bold{not} interfere with realtime behaviour. If you write
\ ** applications based on fJACK you might use *\forth{ <<DEBUG ." This is nice CR DEBUG>> } sequences.  
	SHOW-LOG ;

: .JACK-VERSION	( -- )
\ *G prints the current fJACK version and mods
	(jack-version) TYPE FFTW3-LIB? IF ." -fftw3" THEN ;

: JACK-INFO	( -- )
\ *G Prints information about the fJACK client.
	??CR fJACK-ID 0= IF ." No fJACK client" EXIT THEN
	fJACK-ID (client_name) TYPE SPACE .JACK-VERSION ." , "
	#jack-state
	CASE	JackZombie		OF ." zombiefied" EXIT	ENDOF
		JackTransportStopped	OF ." stopped, "	ENDOF
		JackTransportRolling	OF ." rolling, "	ENDOF
		JackTransportStarting	OF ." started, "	ENDOF
		JackTransportLooping	OF ." looping, "	ENDOF
		JackNetStarting		OF ." netstart, "	ENDOF
		." found in an undefined error state: " #jack-state H. EXIT
	ENDCASE
	BASE @ DECIMAL JACK-SAMPLE-RATE 1 .R ." /s (" JACK-CHKSIZE 1 .R ." ), ld:" JACK-LOAD 1 .R ." %, xr:" JACK-XRUNS 1 .R
	BASE ! ;


: ANY-PORT-CONNECTED? ( -- flag )
\ *G Checks for output port connections done.
 	FALSE JACK-RUNNING? 0= IF EXIT THEN
	#25 0 DO #20 MS (test-ports) DUP IF LEAVE THEN LOOP ;

: JACK-PORTS	( -- )
\ *G Prints information about all connections to the fJACK client.
	JACK-RUNNING? 0= IF ." no ports because client is not running" EXIT THEN
	JACK-CHANNELS 2 * 2 + 0
	DO	??CR I >inport DUP H. SPACE
		IF	I JACK-PORTNAME TYPE SPACE I JACK-PORTFLAG
			CASE
			JackPortIsOutput			OF ." --> "	ENDOF
			JackPortIsOutput JackPortIsTerminal OR	OF ." ||--> "	ENDOF
			JackPortIsInput				OF ." <-- "	ENDOF
			JackPortIsInput JackPortIsTerminal OR	OF ." ||<-- "	ENDOF
								   ." <>?<> " I JACK-PORTFLAG H.
			ENDCASE
			I >inport jack_port_get_connections ?DUP
			IF	DUP
				BEGIN	@+ DUP
				WHILE	C>FSTRING TYPE SPACE
				REPEAT	2DROP jack_free
			THEN
		THEN
	LOOP ;

\ *S Running the fJACK client
: START-JACK-TRANSPORT	( -- )
\ *G As the name says the JACK server transport is started.
	JACK-RUNNING? 0= IF EXIT THEN
	fJACK-ID jack_transport_start
	#400 0 DO #10 MS #jack-state JackTransportRolling = IF LEAVE THEN LOOP
	#jack-state JackTransportRolling <>
	IF <<DEBUG ." Client couldn't start rolling" CR DEBUG>> THEN ;

: STOP-JACK-TRANSPORT	( -- )
\ *G As the name says the JACK server transport is stopped.
	JACK-RUNNING? 0= IF EXIT THEN
	fJACK-ID jack_transport_stop
	#400 0 DO #10 MS #jack-state JackTransportStopped = IF LEAVE THEN LOOP
	#jack-state JackTransportStopped <>
	IF <<DEBUG ." Client couldn't stop" CR DEBUG>> THEN ;

: JACK-SERVER-RUNNING?	( -- flag )
\ *G flag is *\forth{TRUE} when a running jack server is available
	S" test" JackNoStartServer (client_open)
	JackNameNotUnique -1 XOR AND	( id starting problem? )
	IF	<<DEBUG ." checking ... no server running" CR DEBUG>>		DROP FALSE
	ELSE	<<DEBUG ." checking ... running server found" CR DEBUG>> jack_client_close DROP TRUE
	THEN ;

: STOP-JACK-CLIENT		( -- )
\ *G Stops the fJACK client running.
	fJACK-ID 0= IF EXIT THEN
	JACK-ZOMBIE? 0=
	IF	stop-jack-hooks fromtop
		CLEAR-CONNECTIONS FORGET-PORTS fJACK-ID jack_deactivate DROP
	THEN
	fJACK-ID jack_client_close DROP
	JACK-LIB+? IF 0 jack_set_info_function THEN
	0 jack_set_error_function
	0 TO fJACK-ID  JackTransportStopped TO #jack-state
	<<DEBUG ." fjack client stopped" CR DEBUG>> ;


: START-JACK-CLIENT	( -- )
\ *G  Starts the fJACK client - this is done automatically when fjack is compiled and also via the *\forth{AtCold} hook
\ ** so you will only need it after you have stopped the client or it has been stopped by the control application.
\ ** If the clients name is not unique - there could be
\ ** another fJACK client running already - the jack server will choose another name like client-01.
\ ** This is totally OK as the client struct in fJACK takes care about this, and knows about the connections.
\ ** There is one good reason to do so:
\ *P Different applications can all use the default client fJACK and don't have to care about other names.
\ ** So you may start several Forth terminals and start a fJACK client, all these clients
\ ** can share the connection files - the true client name is just hidden.
\ ** You might also choose another client name, see the config option *\forth{JACK-CLIENT-NAME}
\ *P When starting the jack client you will often want to connect some ports. This can be done via a session manager although
\ ** this might not be very intuitive. Another very simple way is: the forth code file *\bold{fjack/sysfiles/defaultports.frt}
\ ** is always interpreted when \*forth{START-JACK-CLIENT} is done. 
\ *P Earlier fJACK versions tried to start the jack server somehow; this has always been somewhat tricky and needed
\ ** OS dependent tricks. *\bold{All} jack versions 0.121/1.97 or later can start the jack server automatically but
\ ** autostarting the jack server is *\bold{not} the default anymore, you can change this via the *\forth{JACK-AUTOSTART} config
\ ** option.
\ *P BTW, you should really switch to modern jack control applications like QjackCtl as they support
\ ** application sessions, and many more features for a better desktop integration.
	{: | buf[ $100 ] -- :}
	JACK-LIB? 0=	ABORT" Please check you jack installation carefully, the jack library could not be found!"
	JACK-LIB+? 0=	IF ??CR ." You have a very old jack audio system installed. Please concider updating!" CR THEN
	JACK-RUNNING?	IF EXIT THEN
	$FJACKSYSFILE COUNT buf[ PLACE  S" defaultports.frt" buf[ PLACE+
	JACK-ZOMBIE?
	IF	<<DEBUG ." Zombie client found; trying to remove it before starting" CR DEBUG>>
		STOP-JACK-CLIENT JACK-ZOMBIE? ABORT" Severe problem, you have to remove the jack zombie manually or restart jackd"
	THEN
	INIT-JACK-VALUES
	<<DEBUG
	JACK-SESSION? 0=	IF ." Session handling is not supported by your jack library!" CR THEN
	JACK-LATENCY? 0=	IF ." Latency handling is not supported by your jack library!" CR THEN
	FFTW3-LIB? 0=		IF ." No fftw3 library found. You have restrictions in FIR filter design and signal analysis!" THEN
	DEBUG>>
	JACK-LIB+? IF SERVER-INFO jack_set_info_function THEN
	SERVER-ERROR jack_set_error_function
 	JACK-SERVER-RUNNING? 0=  JACK-AUTOSTARTING? 0=  AND IF KICKSTART THEN
	JACK-CLIENT-NAME JackNullOption
	session-support? IF JackSessionID OR THEN
	JACK-AUTOSTARTING? 0=	IF JackNoStartServer OR THEN
	(client_open) ( client clientflag)
	START-CLIENT-ERROR
	JackServerStarted -1 XOR AND  JackNameNotUnique -1 XOR AND
  	IF	DROP <<DEBUG ." Couldn't start a client" CR DEBUG>> CR JACKLOG EXIT
	ELSE	<<DEBUG ." A jack client has been registered: " DUP . CR DEBUG>>
	THEN
	TO fJACK-ID
	fJACK-ID jack_get_sample_rate DUP TO JACK-SAMPLE-RATE S>F TO JACK-SAMPLE-F-RATE
 	fJACK-ID jack_get_buffer_size DUP TO JACK-CHKSIZE #8 $1001 WITHIN 0=	ABORT" Chunksize problem, outside #8-$1000"
	fJACK-ID PROCESS-FRAMES	0	jack_set_process_callback		ABORT" Can't setup JACK process callback"
	fJACK-ID NEW-FRAMES	0	jack_set_buffer_size_callback		S" buffersize callback"	?logerror
	fJACK-ID NEW-SAMPLERATE	0	jack_set_sample_rate_callback		S" samplerate callback"	?logerror
	fJACK-ID NEW-TRANSPORT	$1234	jack_set_sync_callback			S" transport callback"	?logerror
	fJACK-ID COUNT-XRUNS	0	jack_set_xrun_callback			S" xruns callback"	?logerror
	fJACK-ID SERVER-REORDER 0	jack_set_graph_order_callback		S" reorder callback"	?logerror
	fJACK-ID SERVER-PORT-CONNECT 0	jack_set_port_connect_callback		S" connect callback"	?logerror
	fJACK-ID CLIENT-STOPPED	0	jack_on_shutdown 0			S" shutdown callback"	?logerror
	JACK-LATENCY? IF fJACK-ID CHECK-LATENCY 0 jack_set_latency_callback	S" latency callback"	?logerror THEN
	session-support? IF fJACK-ID SESSION-EVENT HERE jack_set_session_callback	S" session callback"	?logerror THEN
	CREATE-PORTS
	MIDISEM 0 INITSEMAPHORE
	fJACK-ID jack_activate DROP
	START-JACK-TRANSPORT #50 MS 
	\ You wouldn't expect any connections existing here, but a session manager might have
	\ done so right after the client start was detected.
	ANY-PORT-CONNECTED? 0= IF buf[ COUNT INCLUDED THEN
	start-jack-hooks frombot
	CR JACK-INFO ;

JACKAUDIO DEFINITIONS
: FINISH-JACK-RUNNING	( -- )
	STOP-JACK-CLIENT
	JACK-TEMP-MEM ?DUP IF FREE DROP THEN ; ' FINISH-JACK-RUNNING AtExit		
FORTH DEFINITIONS

: AT-START-JACK-CLIENT		( xt -- )
\ *G Your application might want to do something more when the jack client is started.
\ ** This adds this 'xt' into a chain.
	start-jack-hooks addnd ;

: AT-STOP-JACK-CLIENT		( xt -- )
\ *G Your application might want to do something more when the jack client is stopped.
\ ** This adds this 'xt' into a chain.
	stop-jack-hooks addnd ;

: AT-REMOVE-JACK-CLIENT
\ *G Your application might want to do something more when the jack system is stopped/removed.
\ ** This adds this 'xt' into a chain.
	clear-jack-hooks addnd ;

IS-IFORTH?
[IF]	:NONAME DROP FINISH-JACK-RUNNING ; IS-FORGET AT-REMOVE-JACK-CLIENT
	:ABOUT	??CR ." A jack audio client" ;
[THEN]

IS-VFX?
[IF]	prune: FINISH-JACK-RUNNING ;

[THEN]


\ *N Sessions
\ *P fJACK 1.10 brings support for session management, this feature allows control applications like
\ ** 'qjackctl' to start a bunch of jack applications and connect them all in a way you defined. fJACK
\ ** should be saved in a turnkey application and have a starting bash script. This script can be called
\ ** by the session manager. This feature is very new and not well tested by other users and probably only works
\ ** on iforth linux systems ... please let me know what you find out.

\ FORTH DEFINITIONS
