(*
 * LANGUAGE    : Forth; iForth and MPE Forth working
 * PROJECT     : Forth Tool
 * DESCRIPTION : Audio processing client using JACK
 * CATEGORY    : Realtime Audio tools
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2017-04-20 21:57:11 hanno@schwalm-bremen.de

 *)

\ *! fjack

\ *S Hallo, dear FORTH-audio friend!

\ *P This is a about FORTH programs doing real time audio processing. We don't
\ ** write hardware drivers for the AD/DA but use standard libraries to access the hardware.  
\ ** Modern sound systems use callback driven models, ALSA and Pulseaudio are examples.
\ ** But both are not portable and are not designed for professional audio environments.
\ *P One popular system targeting this is the 'jack-audio-connection-kit' JACK.
\ ** It is available for all major desktop platforms Linux, OSX and Windows. It
\ ** makes use of the hardware driver in a realtime process and can be used with hardare on the motherboard, 
\ ** connected via USB, firewire or even 'over the net'. There are many audio applications
\ ** around using the JACK interface and Forth can now connect to *\bold{all} of them.
\ *P fJACK is a client for JACK; it's written in FORTH and uses advanced techniques like threading,
\ ** access to external libraries and callbacks. So it is *\bold{not} ANS Forth.
\ ** The good news: it's available for all current iForth and VFX Forth (Intel & ARM) versions.
\ *P fJACK is a pretty large and complex system -- don't get frustrated.
\ ** Read the example sources and feel free to send me an email, hanno@schwalm-bremen.de

\ *T Installation

\ *S Installing a JACK system and get it running.
\ *P Read the HOWTO in the net
\ ** and specialized audio groups to find out how you can get the most out of it. Getting JACK running with low latency and no drops/xruns
\ ** is sometimes a bit difficult. Look out for hardware that is well supported for your environment.
\ *P It used to be necessary to use a special realtime-kernel, current mainline linux kernels with optimized settings
\ ** work sufficiently in most cases. On my Sandybridge generation desktop i find latencies of ~10ms for my Focusrite 6i6 interface or ~20ms on the motherboard audio without any xruns
\ ** using the out-of-.the box Fedora 25. 

\ *S Downloading jack
\ *P The first place to look for jack is: http://jackaudio.org/download or go for precompiled packages for your system.
\ ** All linux distributions have rpm or deb packages available.

\ *S What sort of JACK?
\ *P You should always use the newer jack2 system which is available for
\ ** all linux distributions, windows and OSX, the current version is 1.9.10.
\ *P In 2017 most computers run a 64bit operating system so the jack server is a 64bit application too. When you want to run a 32bit
\ ** jack client you need support for this *\bold{mixed mode} withing the jack server. The mixed mode had problems related to a jack2 bug
\ ** related to the way futexes were used (fixed in jack2/git).
\ *P When you observe endless xruns for your 32bit client you will probably have to recompile your jack system using the git code. 
\ ** I use the code compiled with:
\ *C ./waf configure --prefix=/usr --mixed --libdir=/usr/lib64 --alsa --firewire --dbus
\ *P You have to find out where your jack libraries are on your system and you might have to modify the source in
\ ** *\forth{fjack/jacklib.frt} accordingly.

\ *S How is fJACK made running?
\ *P *\bold{All} newer JACK versions (since 0.118 or 1.9.4) have the ability to auto-start
\ ** a jack server when a client wants to register and no running server is found.
\ ** But - this is not a good way to start a jack server at all! ( See comments on jackaudio.org about this.)
\ *P fJACK can autostart but it's much better to use controlling applications like QjackCtl - on linux i use *\bold{cadence} from KXStudio.
\ ** Both tools make sure that other sound systems like pulseaudio are easy to integrate.
\ *P So you can choose between three options:
\ *B Start the jack server before using fJACK with a tool like qjackctl. This allows a precise setup of
\ ** jack according to your system and can be done when you log-in. I think this is the best way. Also it allows session management,
\ ** starting fJACK turnkey applications and other audio application with one click and connecting them all.
\ ** Current linux distributions support starting jack using the DBUS interface.
\ *B Let fJACK start the server automatically. This needs a correct setup of JACK and you will have to set
\ ** *\forth{JACK-AUTOSTARTING?} to true in the config file.
\ *B Write your special *\forth{KICKSTART} word before loading fJACK. This method is the least preferable and
\ ** will probably be removed when JACKs Windows integration has improved.


\ *S What about the fftw3 library?
\ *P All recent fJACK versions use the fftw3 library to define the filter response of FIR filters so you should install it.
\ ** *\forth{BAND-PASS BAND-STOP} and *\forth{USER-FIR} filters with a user-specified
\ ** frequency response and the *\forth{EQUALIZER} will not work without it.
\ ** Calculating the filter parameters is not a real performance problem. But doing signal analysis on large amount of data
\ ** or with sizes <> 2^n depends on a 'calibrated' fftw3 system. Currently creating fftw3 plans uses *\forth{#FFTW_ESTIMATE} because
\ ** in fjack only 2^n sizes are used and then there not much of a difference. See fftw3lib.frt and fftw3.frt

\ *S What about performance?
\ *P fJACK 3.0 can be used on three very different platforms, also the supported Forth compilers differ a lot.
\ *P On iForth 64bit system we have a lot of ram available and a *\bold{very} good floating point compiler. Also the
\ ** timing critical FIR filters have optimized assembler code using SSE2 instructions. On CPUs with a small cache
\ ** or slow RAM access you might want to decrease the sie of the history buffers.
\ *P In intel vfx platforms ths situation is slightly different. We also have the SSE2 filter and 
\ ** the Ndp387.fth is even slightly faster than the iForth generated code
\ ** but is restricted to a FP stack depth of 8. The fjack modules use floating data up to a depth of 5, so beware!
\ ** The Hfp387.fth module has a much worse performance as it does not optimize well, also there are severe problems with external libraries and
\ ** callbacks. fJACK does *\bold{not} support it.
\ *P A - for me completely new experience - was working on ARM single board systems. I got myself a ARTIK710 system, it
\ ** uses a comparably fast CPU running 8 Cortex A53 cores. The Vfx ARM Linux compiler generates good integer code but the
\ ** floating point performance is pretty bad because the code is almost not optimized.
\ *P What can we do to tune performance on these small boards?  The most effective tuning will be done
\ ** by decreasing the audio sampling rate!
\ *P Many ARM cpus have very limited ram cache, so it will be good
\ ** to decrease the history buffer size (see the config options on how to do that). FIR filters need a lot of FP maths,
\ ** use IIR filters instead until we have NEON optimised code or reduce the number of filter points.
\ ** Let's hope vfx works on the floating point compiler! 
\ *P There is an option *\{JACK-EMBEDDED?}, it will set the defaults to values more
\ ** appropriate on these boards.

\ *S What about MIDI?
\ *P fJACK supports the jack MIDI system - see the MIDI section for details.
\ *P Bruno Degazio wrote the Karplus-Pluck synthesizer for fJACK which can be used as a MIDI instrument.
\ ** Another hammond-like synthesizer is also available. Both use the fjack/instrument framework. This is
\ ** a new feature and the instrument API is not yet fixed, so only a few words
\ ** are in the FORTH vocabulary yet. Read the source and ask via email if you want to dig into this.

\ *S What about memory?
\ *P fJACK needs quite a lot of ram allocated from the system. As this is a realtime application, all memory *\bold{must}
\ ** be permanent and *\bold{MUST NOT} be swapped in/out. When using the default config parameters fJACK needs ~10MB of
\ ** allocated memory. The biggest 'memory-grabber' is the history mechanism.
\ *P See the 'What can be configured in fJACK?' chapter to find out more about changing such config options.

\ *S Contact, help & bug reports
\ *P Feel free to ask for help at hanno@schwalm-bremen.de - normally i will respond in a few days.
\ *P Whenever you plan to use fJACK in commercial applications read the Licence.txt and send me an email.
\ ** Using fJACK in private or educational settings is free of charge on all platforms.


\ *S Where are the fJACK sources?
\ *P They are all inside the *\bold{fjack} directory containing the client and extension module sources
\ ** plus the 'loader' *\bold{fjack.frt}. The fjack/sysfiles directory holds the connections
\ ** and some binary files. This subdirectory should be writable.
\ *P *\bold{All iForth} versions are supported. The fJACK sources are inside the *\bold{include} directory. *\forth{IN fjack} will do.
\ *P On *\bold{VFX Forth} systems the fJACK sources are in the *\bold{examples} directory. You should *\bold{cd} to this directory, then
\ ** *\forth{INCLUDE fjack.frt} will do. When you get an fjack update you could replace that version or have it somewhere
\ ** else.
\ *P The release history file is fjack/history.txt

\ *S Testing the client
\ *P Loading as above compiles the basic Forth interface to jack audio,
\ ** defines a jack-client with some ports, connects these ports and plays some tones via the karplus instrument.
\ *P There are some demos availabe, see fjack/examples/tests.frt or fjack/examples/board.frt.
\ *P When you *\bold{don't want the whole fJACK} stuff included you have some configure options, see below.  
\ *P iForth systems have a realtime capable graphics user interface, the gui/graftool system has some plugins that can be used
\ ** to control filter settings and more for the analyser plugin fjack/fft.frt and the visual frontend gui/showfft.frt
\ *P As fully working example applications we have gui/studio and the radio/swlapp applications.

\ *S Making fJACK Turnkey Applications
\ *P fJACK can be used in turnkey applications, all plugin object structs
\ ** and the client data are in permanent memory space to support this.
\ ** fJACK inside a turnkey applications will start automatically, you will just have to put your application 'xt' in the
\ ** boot chain.
\ *C ' application AtCold
\ *P Stopping the client and system housekeeping is done in the *\forth{AtExit}, *\forth{AtCold} and fJACKs internal chain.
\ *P vfx produces 'complete' binaries, iForth turnkeys using *\forth{SAVE-SYSTEM} are fully supported, see gui/studio.frt as an example.

\ *S Removing the fJACK client
\ *P On all fJACK systems we have *\forth{ -fjack ( -- )}. It works like a *\forth{MARKER}, it stops the client and background processing,
\ ** frees all allocated memory and removes the fJACK system from forth dictionary space. 


\ *T The fJACK Client User Manual

[DEFINED] IFORTH!  CONSTANT IS-IFORTH? IMMEDIATE
[DEFINED] VFXFORTH CONSTANT IS-VFX? IMMEDIATE

VOCABULARY JACKAUDIO
\ *G The *\forth{JACKAUDIO} vocabulary holds internals of the fJACK system, it's words could be changed
\ ** in next releases and should be used in end-user applications with care.

[UNDEFINED] SF!+	[IF]	: SF!+	DUP SF! 1 SFLOATS + ;	[THEN]
[UNDEFINED] SF@+	[IF]	: SF@+	DUP SF@ 1 SFLOATS + ;	[THEN]
[UNDEFINED] SF+!	[IF]	: SF+!	DUP SF@ F+ SF! ;	[THEN]
[UNDEFINED] F+!		[IF]	: F+!	DUP F@ F+ F! ;		[THEN]
[UNDEFINED] F-!		[IF]	: F-!	DUP F@ FSWAP F- F! ;	[THEN]
[UNDEFINED] FTRUNC	[IF]	: FTRUNC    FINT ;   	   	[THEN]
[UNDEFINED] F2*		[IF]	: F2*	FDUP F+ ;		[THEN]
[UNDEFINED] F<>		[IF]	: F<>	F= 0= ;			[THEN]
[UNDEFINED] SF+!+	[IF]	: SF+!+	DUP SF@ F+ SF!+ ;	[THEN]
[UNDEFINED] FTUCK	[IF]	: FTUCK	FSWAP FOVER ;		[THEN]
[UNDEFINED] F2DROP	[IF]	: F2DROP FDROP FDROP ;		[THEN]

ALSO JACKAUDIO ALSO DEFINITIONS
IS-IFORTH?
[IF]	INCLUDE fjack/I_osdeps.frt
[THEN]

IS-VFX?
[IF]	INCLUDE fjack/V_osdeps.frt
[THEN]

\ *S A short processing overview
\ *P fJACK does real-time processing on a user defined number of audio channels (ports) *\forth{JACK-CHANNELS}, it controls and is controlled by
\ ** a running jack server. You can connect fJACK with other jack-ready application via these ports. To allow audio data processing
\ ** you can define audio plugins and put them in any of the slots. So the basic overview is like shown in *\bold{fjack/io_slots.png}

\ *P Please recognize that every processing plugin has it's
\ ** own history buffer. Later you will find out, that *\bold{every} plugin can indeed read data from *\bold{any} history buffer.
\ ** This sheme allows much improved sound effects and mixing.
	INCLUDE fjack/jacklib.frt
	INCLUDE fjack/fftw3lib.frt
	INCLUDE fjack/config.frt
	INCLUDE fjack/filkern.frt
	INCLUDE fjack/types.frt
	INCLUDE fjack/client.frt

PREVIOUS PREVIOUS DEFINITIONS

