(*
 * LANGUAGE    : iForth
 * PROJECT     : Audio processing
 * DESCRIPTION : common MIDI commands
 * CATEGORY    : fjack audio processing addon
 * AUTHOR      : Hanno Schwalm
 * LAST CHANGE : 2012-08-11 22:20:44
 *)

[DEFINED] -fjack 0=		[IF] INCLUDE fjack.frt [THEN]

IS-IFORTH?
[IF]	REVISION -fjack/midi "--- MIDI commands       Version 0.08 ---"
[ELSE]	ANEW -fjack/midi
[THEN]

DECIMAL ALSO JACKAUDIO ALSO DEFINITIONS

\ *S JACK MIDI support
\ *P fJACK supports MIDI now! There is a basic wordset available that allows control of external instruments,
\ ** communication with other JACK MIDI clients and receiving MIDI-IN events for things like synthesizers.
\ ** Of course this is all done in realtime or made safe by fifos.
\ *N Receiving MIDI messages
\ YOUR-MIDI-IN-EVENT ( struct -- )
\ *G The fJACK midi ports receive messages from connected midi enabled jack clients.
\ ** You can define a word that is executed by the processing callback whenever a MIDI event is received.
\ ** All MIDI events valid at this time are served in a bunch *\bold{before} the audio processing is done.
\ ** *\forth{' YOUR-MIDI-IN-EVENT IS DO-PROCESS-MIDI-IN-EVENT} defines the predefined hook to do what you desire.
\ *P Please remember again:
\ ** this is all done in the processing callback! So all comments regarding realtime processing elsewhere in this manual are relevant
\ ** for the midi processing too.
\ *P The struct passed holds the pure 'MIDI-command' in the first bytes.
\ ** A simple MIDI interpreter/synthesizer could just interpret this information.
\ ** As JACK-MIDI is designed for sample-precise audio we need an exact-to-the-frame time for the command.
\ ** The struct has more information than just the command bytes. See *\forth{MIDI-IN-EVENT-DATA}
FORTH DEFINITIONS

: MIDI-IN-EVENT-DATA	( event-struct -- str len time )
\ *G Reads information from the midi-event struct. This event-struct is the given to you in the *\forth{DO-PROCESS-MIDI-IN-EVENT}.
\ ** This word will be used by a more advanced interpreter hooked into *\forth{DO-PROCESS-MIDI-IN-EVENT}. You
\ ** will probably also use *\forth{INTERPRET-MIDI-MESSAGE} that extracts the information passed.
	DUP 2 CELLS + @ OVER 3 CELLS + @ ;

: SEND-MIDI-MESSAGE	( p2? p1? p0 pars frame -- )
\ *G We also can send MIDI messages via the outport to other clients like a fully featured synthesizer.
\ ** This command is availble at the output port precise-to-the-frame - you can calculate the frame by
\ ** *\forth{JACK-POSITION} plus some offset you would like.
\ ** *\forth{SEND-MIDI-MESSAGE} is threading safe by semaphore protection. All sent messages are queued and
\ ** presented at the outport in the right order with precise timing.
	fJACK-ID 0= IF 2DROP 2DROP DROP EXIT THEN
	OVER 3 > ABORT" MIDI messages should only have a length of up to 3 bytes"
	<MIDI #midi_out >midi_out TUCK 3 CELLS + !	( p2 p1 p0 pars addr )
	0 OVER 1 CELLS + !			\ marks this as not sent already
	2DUP 2 CELLS + !
	SWAP BOUNDS DO I C! LOOP
	1 +TO #midi_out MIDI> ;

: MIDI-TO-MYSELF	( flag -- )
\ *G You can connect fJACKs MIDI output port to fJACKS MIDI input port or can disconnect it.
\ ** This is just a shortcut to let the sent MIDI messages appear at fJACK MIDI-inport.
	fJACK-ID 0= IF DROP EXIT THEN
	{: flag | dest[ $100 ] src[ $100 ] -- :}
	fJACK-ID (client_name)	src[ PLACE  S" :" src[ PLACE+
	src[ COUNT dest[ PLACE
	S" Midi_in" src[ PLACE+
	S" Midi_out" dest[ PLACE+
	fJACK-ID  dest[ COUNT  src[ COUNT  flag
	IF (jack_connect) ELSE (jack_disconnect) THEN DROP ;

: ACTIONS-MIDI-SAMPLE	( -- 0 | cmd0 len0  cmd1 len1 .. cmd_n-1 len_n-1  n )
\ *G All MIDI-IN events are presented to the *\forth{DO-PROCESS-MIDI-IN-EVENT} hook in a bunch.
\ ** In some situations you might want an io-processing plugin to know about at MIDI-IN event at the
\ ** current sample.
	0  #midi_in #midi_in_read
	?DO	I >midi_in DUP 3 CELLS + @ #FJACK-SAMPLE =
		IF DUP 2 CELLS + @ ROT 1+ ELSE DROP THEN
	LOOP ;

: INTERPRET-MIDI-MESSAGE	( message-string -- type channel val1 val2 )
\ *G A MIDI message string will be interpreted quite often, this returns the parameters to be
\ ** presented to locals or what you like ...
	DUP C@ DUP $F0 AND SWAP $0F AND ROT 1 CHARS + DUP C@ SWAP 1 CHARS + C@ ;

JACKAUDIO DEFINITIONS

0 VALUE midi-timing-reference

FORTH DEFINITIONS
\ *N Sending MIDI messages
\ *P The fJACK MIDI system allows MIDI messages to be sent/received to/from other MIDI handling clients with
\ ** sample-precise timing. This means: Sending a message to start a tone will be received at exactly the sample
\ ** you desired when sending the message.
\ ** You can either send a message at a time-from-now or at a time-from-reference-time.
\ *P When you want to start a bunch of tones all at exactly the same time you will probably define that time
\ ** by using *\forth{SET-TIMEREFERENCE} and later use the *\forth{TIMERREFERENCE>FRAME} calculation

: TIMEDIFF>FRAME	( -- #frame ) ( F: sec -- )
\ *G Calculates the absolute #frame from the time (seconds-from-now)
	JACK-TIME? JACK-PROCESS-TIME? D- D>F 1e-6 F* F-
	JACK-SAMPLE-F-RATE F* F>S 0 MAX JACK-POSITION + ;

: SET-TIMEREFERENCE	( -- ) ( F: seconds -- )
\ *G Sets the current time + seconds as the time reference to be used later by *\forth{TIMERREFERENCE>FRAME}.
	TIMEDIFF>FRAME TO midi-timing-reference ;
: TIMERREFERENCE>FRAME	( -- #frame ) ( F: seconds -- )
\ *G Calculate the absolute #frame from the time (seconds-from-timer-reference)
	JACK-SAMPLE-F-RATE F* F>S midi-timing-reference + ;

: NOTE-ON  ( velocity note channel -- )	( F: sec -- )
\ *G Sends a command to start a midi-note with given velocity and midi-channel. Also the time has to be given.
\ ** This commands sends at time-from-now.
	$90 OR 3 TIMEDIFF>FRAME SEND-MIDI-MESSAGE ;

: NOTE-OFF ( note channel -- )	( F: sec -- )
\ *G Sends a command to stop a midi-note with given midi-channel. Also the time has to be given.
\ ** This commands sends at time-from-now.
	0 -ROT $80 OR 3 TIMEDIFF>FRAME SEND-MIDI-MESSAGE ;

:NONAME	TRUE TO midi-action ; AT-START-JACK-CLIENT
:NONAME	FALSE TO midi-action ; AT-STOP-JACK-CLIENT

IS-IFORTH?
[IF]
	:NONAME	DROP FALSE TO midi-action ; IS-FORGET NOTE-OFF
	:ABOUT ??CR ." Sending MIDI messages at precise times" ;
[THEN]
IS-VFX?
[IF]	prune:	FALSE TO midi-action ;
[THEN]

PREVIOUS PREVIOUS DEFINITIONS

