\ CallDefWin64.fth - Callbacks using an EXTERN: style syntax

((
Copyright (c) 2023
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)

CallDef:  type NAME( int a, float b, ... ):
  Forth source
  ....
;
Execution of NAME returns the callback's structure address

create NAME
  entrypoint address ,			\ NAME is essentially a constant
  address holding call to Forth	,	\ where call ACTION goes
  0 ,
<entrycode ... entry code>
<exitcode ... exit code>
l: entrypoint
  call entrycode   call ACTION   call exitcode   ret
l: ACTION
  ...
;

To do
=====

Change history
==============
20230927 SFP002 Corrected float parameters.
		Added DEFCALLPROC:
20230420 SFP001 First working release.
))

only forth definitions
decimal


\ ===========
\ *> callback
\ *S Callbacks using a C prototype.
\ ===========
\ *P The enhanced mechanism was developed to support more C types
\ ** accurately, and to be faster than the previous mechanism.
\ ** Code generators are provided for obtaining arguments from
\ ** the O/S and for returning data. The type notation is the
\ ** same as that of the *\fo{EXTERN:} notation. Source code
\ ** is in the file *\i{VFXBase\CallDefWin64.fth}, which also
\ ** contains an about box example.

\ *P The callback is in two portions. The first is a C prototype
\ ** for the callback. The second portion is a nameless Forth
\ ** word which forms the action of the callback. Note that
\ ** referencing the name of a *\fo{CALLDEF:} returns the
\ ** address of a structure from which the entry point can be
\ ** found.

\ *E CallDef: uint * AboutVFXDialogProc(
\ **   HWND hDlg, UINT msg, WPARAM wparam1, LPARAM lparam1
\ ** ):
\ **   {: hdlg message wparam lparam -- ior :}
\ **   message case
\ **     WM_INITDIALOG of
\ **       ...
\ **     endof
\ **
\ **     WM_CLOSE of
\ **         hDlg WM_COMMAND IDOK 0 SendMessage drop 0
\ **     endof
\ **
\ **     drop 0
\ **   end-case
\ ** ;
\ **
\ ** AboutVFXDialogProc get-CallDefEntry  \ -- entrypoint


also extern.voc definitions

: parseCBtext	\ "<text>" -- ;  type NAME( int a, float b, ... ):
\ Parse the text up to the closing ')', leaving the type information
\ in the argument list.
  initArgData setFirstLine
  splitRet&Name
  isSingleName     			\ -- c-addr u ; leaves return type info
  setDefExt				\ set default declaraction type
  ExtractArgs				\ get data from (...)
;

: calcCSrcFrame	\ --
\ Calculate the C stack offsets. For a callback, the C stack is the
\ source stack on entry, whereas for EXTERNs it was the destination
\ stack.
  numArgs @ 0
  ?do  i ArgList[] CalcDest  loop
; doNotSin

: calcFDestOff	\ --
\ Calculate the offsets of data items on the Forth input stack.
\ For a callback, the Forth stack is the destination for the
\ parameter copies, whereas for EXTERNs it was the source
\ stack.
  numArgs @ if
    0 numArgs @ 1-
    do  i ArgList[] CalcSource  -1 +loop
  endif
; doNotSin

\ +p Note that the source offset starts at zero. When (as in all VFXes)
\ +* TOS is cached in a register, a source offset of 0 indicates the register.
\ +* Hence, the cell size should be subtracted when using the offset.

\ +P The Extern source code refers to the destination and source of the
\ +* copy. They refer to generating an Extern, in which the C frame stack
\ +* is thr destination. When generating a Callback, the C frame stack is
\ +* the source.

\ +P The offsets are described as being for the C frame (Coff, RSP),
\ +* Forth parameter stack (Poff, RBP) or the Forth float stack
\ +* (Foff, R13).

\ ============================================
\ Integer copies ; C Frame to Forth data stack
\ ============================================

: bytePSP,	\ offset -- ; destination is byte on parameter stack
  dup 0= if
    drop  [a  bl,  a]
  else
    8 - >r  [a  byte r> [rbp],  a]
  then
;

: wordPSP,	\ offset -- ; destination is 16 bits on parameter stack
  dup 0= if
    drop  [a  bx,  a]
  else
    8 - >r  [a  word r> [rbp],  a]
  then
;

: dwordPSP,	\ offset -- ; destination is 32 bits on parameter stack
  dup 0= if
    drop  [a  ebx,  a]
  else
    8 - >r  [a  dword r> [rbp],  a]
  then
;

: xwordPSP,	\ offset -- ; destination is 64 bits on parameter stack
  dup 0= if
    drop  [a  rbx,  a]
  else
    8 - >r  [a  r> [rbp],  a]
  then
;

[-short-branches
: CDcopyArgS8	{: ^ai Poff Coff -- :}
\ Copy a signed 8 bit item from the C frame to the Forth stack.
\   C [RSP] to Poff [RBP]
\ cr ." S8 Coff=" Coff .  ." Poff=" Poff .
  ^ai ai.inReg? @ if
    [a  movsx  a]  Poff bytePSP,  ^ai IntArgReg  [a]
  else
    [a  movsx   RAX, byte Coff #32 + [RSP]
        mov    a]  Poff xwordPSP,  [a  RAX a] [a]
  then
;

: CDcopyArgS16	{: ^ai Poff Coff -- :}
\ Copy a signed 16 bit item from the C frame to the Forth stack.
\   C [RSP] to Poff [RBP]
\ cr ." S16 Coff=" Coff .  ." Poff=" Poff .
  ^ai ai.inReg? @ if
    [a  movsx  a]  Poff wordPSP,  ^ai IntArgReg  [a]
  else
    [a  movsx   RAX, worda Coff #32 + [RSP]
        mov    a]  Poff xwordPSP,  [a  RAX  a] [a]
  then
;

: CDcopyArgS32	{: ^ai Poff Coff -- :}
\ Copy a signed 32 bit item from the C frame to the Forth stack.
\   C [RSP] to Poff [RBP]
\ cr ." S32 Coff=" Coff .  ." Poff=" Poff .
  ^ai ai.inReg? @ if
    [a  movsxd  RAX, a]  ^ai IntArgReg32  [a]
    [a  mov  a]  Poff xwordPSP,  [a  RAX  a] [a]
  else
    [a  movsxd  RAX, dword Coff #32 + [RSP]
        mov     a]  Poff xwordPSP,  [a RAX a]  [a]
  then
;

: CDcopyArgU8	{: ^ai Poff Coff -- :}
\ Copy an unsigned 8 bit item from the C frame to the Forth stack.
\   C [RSP] to Poff [RBP]
\ cr ." U8 Coff=" Coff .  ." Poff=" Poff .
  ^ai ai.inReg? @ if
    [a  movzx  a]  Poff bytePSP,  ^ai IntArgReg  [a]
  else
    [a  movzx   RAX, byte Coff #32 + [RSP]
        mov    a]  Poff xwordPSP,  [a  RAX a] [a]
  then
;

: CDcopyArgU16	{: ^ai Poff Coff -- :}
\ Copy a signed 16 bit item from the C frame to the Forth stack.
\   C [RSP] to Poff [RBP]
\ cr ." U16 Coff=" Coff .  ." Poff=" Poff .
  ^ai ai.inReg? @ if
    [a  movzx  a]  Poff wordPSP,  ^ai IntArgReg  [a]
  else
    [a  movzx   RAX, worda Coff #32 + [RSP]
        mov    a]  Poff xwordPSP,  [a  RAX  a] [a]
  then
;

: CDcopyArgU32	{: ^ai Poff Coff -- :}
\ Copy a signed 32 bit item from the C frame to the Forth stack.
\   C [RSP] to Poff [RBP]
\ cr ." U32 Coff=" Coff .  ." Poff=" Poff .
  ^ai ai.inReg? @ if
    [a  mov      EAX,  a]  ^ai IntArgReg32  [a]
  else
    [a  mov      EAX,  Coff #32 + [RSP]  a] [a]
  then
  [a  mov  a]  Poff dwordPSP,  [a  RAX  a] [a]
;

: CDcopyArgX64	{: ^ai Poff Coff -- :}
\ Copy a 64 bit item from the C frame to the Forth stack.
\   C [RSP] to Poff [RBP]
\ cr ." X64 Coff=" Coff .  ." Poff=" Poff .
  ^ai ai.inReg? @ if
    [a  mov  a]  Poff xwordPSP,  ^ai IntArgReg  [a]
  else
    [a  mov     RAX, Coff #32 + [RSP]  a] [a]
    [a  mov  a]  Poff xwordPSP,  [a  RAX  a] [a]
  then
;
short-branches]


\ ============================================
\ Float copies ; C Frame to Forth Float stack
\ ============================================

[-short-branches
: CDcopyFParg32	{: ^ai Foff Coff -- :}
\ Copy a 32 bit float from the C frame to the Forth FP stack.
\   Coff [RSP] to Foff [R13]
\ For callbacks the source items are always SSE items from the OS.
\ The destination items are in the FP format selected by *\fo{FPsystem}
  FPsystem SSE64system = if
    ^ai ai.inReg? @ if
      Foff 0= if
        [a  cvtss2sd   XMM8,  a]  ^ai FPArgReg  [a]
      else
        [a  cvtss2sd   XMM9,  a]  ^ai FPArgReg  [a]
        [a  movsd      Foff 8 - [r13], XMM9  a] [a]
      then
    else
      Foff 0= if
        [a  cvtss2sd   XMM8, Coff #32 + [RSP]  a] [a]
      else
        [a  cvtss2sd   XMM9, Coff #32 + [RSP]  a] [a]
        [a  movsd      Foff 8 - [r13], XMM9  a] [a]
      then
    then
  then
  FpSystem HFP387System = if 		\ FTOS in deep FP stack
    ^ai ai.inReg? @ if
      Foff 0= if
        [a  cvtss2sd   $-10 [r13],  a]  ^ai FPArgReg  [a]
        [a  fld        double $-10 [r13]  a] [a]
      else
        [a  cvtss2sd   $-10 [r13],  a]  ^ai FPArgReg  [a]
        [a  fld        double $-10 [r13]  a] [a]
        [a  fstp       tbyte Foff #16 - [r13]  a] [a]
      then
    else
      Foff 0= if
        [a  fld        float Coff #32 + [RSP]  a] [a]
      else
        [a  fld        float Coff #32 + [RSP]  a] [a]
        [a  fstp       tbyte Foff #16 - [r13]  a] [a]
      then
    then
  then
  FpSystem NDP387System = if 		\ NDP internal  stack
    ^ai ai.inReg? @ if
      [a  cvtss2sd   $-10 [r13],  a]  ^ai FPArgReg  [a]
      [a  fld        double $-10 [r13]  a] [a]
    else
      [a  fld        float Coff #32 + [RSP]  a] [a]
    then
  then
; doNotSin

: CDcopyFParg64	{: ^ai Foff Coff -- :}
\ Copy a 64 bit float from the C frame to the Forth FP stack.
\   Coff [RSP] to Foff [R13]
\ For callbacks the source items are always SSE items from the OS.
\ The destination items are in the FP format selected by *\fo{FPsystem}
  FPsystem SSE64system = if
    ^ai ai.inReg? @ if
      Foff 0= if
        [a  movsd      XMM8,  a]  ^ai FPArgReg  [a]
      else
        [a  movsd      Foff 8 - [r13],  a]  ^ai FPArgReg  [a]
      then
    else
      Foff 0= if
        [a  movsd      XMM8, Coff #32 + [RSP]  a] [a]
      else
        [a  movsd      XMM9, Coff #32 + [RSP]  a] [a]
        [a  movsd      Foff 8 - [r13], XMM9  a] [a]
      then
    then
  then
  FpSystem HFP387System = if 		\ FTOS in deep FP stack
    ^ai ai.inReg? @ if
      Foff 0= if
        [a  movsd      $-10 [r13],  a]  ^ai FPArgReg  [a]
        [a  fld        double $-10 [r13]  a] [a]
      else
        [a  movsd      $-10 [r13],  a]  ^ai FPArgReg  [a]
        [a  fld        double $-10 [r13]  a] [a]
        [a  fstp       tbyte Foff #16 - [r13]  a] [a]
      then
    else
      Foff 0= if
        [a  fld        double Coff #32 + [RSP]  a] [a]
      else
        [a  fld        double Coff #32 + [RSP]  a] [a]
        [a  fstp       tbyte Foff #16 - [r13]  a] [a]
      then
    then
  then
  FpSystem NDP387System = if 		\ NDP internal  stack
    ^ai ai.inReg? @ if
      [a  movsd      $-10 [r13],  a]  ^ai FPArgReg  [a]
      [a  fld        double $-10 [r13]  a] [a]
    else
      [a  fld        double Coff #32 + [RSP]  a] [a]
    then
  then
; doNotSin
short-branches]


\ =======================
\ Copy callback arguments
\ =======================

: CDoffsets	\ ^ai -- destoff srcoff
\ Extract the base offsets from the structure. The source offset
\ is assumed to be on the C/system stack. The dest offset is on
\ the PSP/RBP stack.
  dup ai.PSPoff @  swap ai.CFrameOff @
;

: CDFPoffsets	\ ^ai -- destoff srcoff
\ Extract the base offsets from the structure. The source offset
\ is assumed to be on the FSP/R13 stack and is adjusted to be based
\ from the FTOS. The dest offset is on the C/system stack.
  dup ai.FSPoff @  swap ai.CFrameOff @
;

: CDcopySINT	\ ^ai --
\ Compile code to copy signed ints from the frame.
\ with any required promotion.
  case  dup ai.size @
    1 of  dup CDoffsets CDcopyArgS8  endof
    2 of  dup CDoffsets CDcopyArgS16  endof
    4 of  dup CDoffsets CDcopyArgS32  endof
    8 of  dup CDoffsets CDcopyArgX64  endof
     err_CGsize throw
  endcase
;

: CDcopyUINT	\ ^ai --
\ Compile code to copy unsigned ints from the frame.
\ with any required promotion.
  case  dup ai.size @
    1 of  dup CDoffsets CDcopyArgU8  endof
    2 of  dup CDoffsets CDcopyArgU16  endof
    4 of  dup CDoffsets CDcopyArgS32  endof
    8 of  dup CDoffsets CDcopyArgX64  endof
     err_CGsize throw
  endcase
;

: CDcopyBOOL	\ ^ai --
\ Compile code to copy a BOOL from the frame.
\ with any required promotion/demotion.
  dup CDoffsets CDcopyArgU8
;

: CDcopyFLOAT	\ ^ai --
\ Compile code to copy FTOS from the frame as a float or double
  ?noFP
  case  dup ai.size @
    4 of  dup CDFPoffsets CDcopyFPArg32  endof
    8 of  dup CDFPoffsets CDcopyFPArg64  endof
       err_CGsize throw
  endcase
; doNotSin

: CDcopyArg	\ ^ai --
\ Compile code to copy an item from the calling frame to Forth
\ with any required promotion.
  case  dup ai.type @
    UINTtype of  CDcopyUINT  endof
    SINTtype of  CDcopySINT  endof
    BOOLtype of  CDcopyBOOL  endof
    FLOATtype of  CDcopyFloat  endof
      err_ImpType throw
  endcase
;

: CDcopyFrame	\ --
\ Copy the arguments from the Forth stack to the O/S stack with
\ any required promotion/demotion.
  L>R? REQUIRE_NREV xor 0= if
    numArgs @ if
      0 numArgs @ 1-
      do  i ArgList[] CDcopyArg  -1 +loop
    endif
  else
    numArgs @ 0
    ?do  i ArgList[] CDcopyArg  loop
  endif
;


\ ========================
\ Return data - Forth to C
\ ========================

[-short-branches
: CDretInt	\ ^ai -- ; return an integer (max 64 bits)
\ Forth integers just need copying into RAX
  drop					\ discard type
  [a  mov    rax, rbx  a] [a]
;

: CDretFloat	\ ^ai -- ; return a Float/Double
  ?NoFP
  FPsystem SSE64System = if
    ai.size @ 4 = if
      [a  cvtsd2ss   XMM0, XMM8  a] [a]
    else
      [a  movsd      XMM0, XMM8  a] [a]
    then
    exit
  then
  FPsystem HFP387system =  FPsystem NDP387system = or  if
    [a  fstp       $-10 [r13]  a] [a]	\ save in shadow float space
    ai.size @ 4 = if
      [a  cvtsd2ss   XMM0, $-10 [r13]  a] [a]
    else
      [a  movsd      XMM0, $-10 [r13]  a] [a]
    then
    exit
  then
;

: CDretData	\ -- ; compile exit code
\ Compile code to return an item from Forth to the calling frame
\ with any required promotion.
  case  ReturnRes  dup ai.type @
    UINTtype of  CDretInt  endof
    SINTtype of  CDretInt  endof
    BOOLtype of  CDretInt  endof
    FLOATtype of  CDretFloat  endof
      err_ImpType throw
  endcase
;
short-branches]


\ ===========================
\ CallDef entry and exit code
\ ===========================

12 cells constant #CDF_pars	\ start of parameters (0..n)
11 cells constant #CDF_SYSRET	\ overall callback return address
10 cells constant #CDF_RA	\ callback-entry/exit return address
 9 cells constant #CDF_VM	\ setCallDefVM return address
\ ============= for #in <= 4 all params in registers
\ first/left: %rcx, %rdx, %r8, %r9 :last/right
\ preserve callee save registers
\ 8 cells equ CBF_R12
\ 7 cells equ CBF_R13
\ 6 cells equ CBF_R14
\ 5 cells equ CBF_R15
\ 4 cells equ CBF_RDI
\ 3 cells equ CBF_RSI
\ 2 cells equ CBF_RBX
\ 1 cells equ CBF_RBP
\ 0 cells equ CBF_GS:0

Proc setCallDefVM	\ --
\ +G Called from within the entry routine to generate the Forth VM regs.
\ +* The CPU stack is dropped, and new stacks and frames are built for
\ +* Forth VM.
\ Parameters 4..n
\ ---
\ 32 byte buffer space (params 0..3)
\ --- Start of Forth callback frame ---
\ CALLBACK RETURN ADDRESS	\ 3 return addresses	  CDF_SYSRET
\ CALLDEF-ENTRY RETURN ADDRESS                            CDF_RA
\ SETCALLDEFVM return address                             CDF_VM
	push	r12			\ 9 regs here
	push	r13
	push	r14
	push	r15
	push	rdi
	push	rsi
	push	rbx
	push	rbp
	xor	rax, rax		\ zero
	gs: mov	r10, 0 [rax]		\ exception pointer
	push	r10
\ --- End of Forth callback frame ---

\ drop stack, space for VM, RSP at bottom
	lea	rsp, #CBdropped negate [rsp]
\ User area
	lea	rsi, #rsp_USER [rsp]
\ TIB
	lea	rdi, #rsp_TIB [rsp]		\ EDI = TIB
	mov	$0B cells [rsi], rdi		\ set 'TIB
\ Float stack, prevent crash into TIB
	lea	r13, #rsp_FSPtop /FSPguard - [rsp]  \ float stack guard space
	mov	$14 cells [rsi], r13		\ FS0
	mov	$15 cells [rsi], r13		\ FSP
\ Data stack, below Float Stack
	lea	rbp, #rsp_PSPtop /PSPguard - [rsp] \ data stack guard space
	mov	0 [rsi], rbp			\ set S0
\ Return stack, no protection as below Parameter/Data Stack
	mov	8 [rsi], rsp			\ set R0

\ Set USER variables
	mov	rax, # $0A
	mov	6 cells [rsi], rax		\ BASE is decimal
	xor	rax, rax
	mov	$18 cells [rsi], rax		\ ThreadExit?
	mov	$19 cells [rsi], rax		\ ThreadTCB
	mov	$1A cells [rsi], rax		\ ThreadSync

	lea	rax, USER-SIZE /ExceptStack - /Syspad - [rsi]		\ Syspadstart ; SFP007
	mov	$1F cells [rsi], rax		\ ^SysPad

	lea	rax, USER-SIZE 4 cells - [rsi]	\ top of exception stack = UP+$0FF0
	mov	USER-SIZE 3 cells - [rsi], rax	\ EXC0
	mov	USER-SIZE 2 cells - [rsi], rax	\ EXCP

\ Set the default I/O pointers. RIP relative addressing means that
\ it works with ATTACH/DETACH in a DLL before relocation.
	mov     rax, [] ip-default
	mov	$0F cells [rsi], rax		\ IP-HANDLE
	mov     rax, [] op-default
	mov	$0E cells [rsi], rax		\ OP-HANDLE

\ Get the return address
	jmp	#CBdropped #CDF_VM + [rsp] \ return address from SETCALLDEFVM
	ret				\ using ret keeps the disassembler happy
end-code

\ CALLBACK RETURN ADDRESS		\ 3 return addresses	  CDF_SYSRET
\ CALLBACK-ENTRY RETURN ADDRESS - MUST be left for CALLBACK-EXIT  CDF_RA
\ CLEARCALLDEFVM return address                                   CDF_VM
\   call calldef-entry  |  call action-xt  |  call calldef-exit  |  ret
\    call setcalldefvm                        call clrcalldefvm

Proc clrCallDefVM	\ restore Forth VM to C VM
	pop	r8			\ save return address from dropped stack
\ Parameters 4..n			\ on dropped stack, so pop ok
\ ---
\ 32 byte buffer space (params 0..3)
\ --- Start of Forth callback frame ---
\ CALLBACK RETURN ADDRESS		\ 3 return addresses	  CDF_SYSRET
\ CALLBACK-ENTRY RETURN ADDRESS - MUST be left for CALLBACK-EXIT  CDF_RA
\ CLEARCALLDEFVM return address                                   CDF_VM
	lea	rsp, #CBdropped [rsp]	\ restore RSP to before frame creation
\ restore saved regs
 	pop	r11			\ saved exception pointer
 	xor	r10, r10		\ zero
 	gs: mov	0 [r10], r11		\ put back
 	pop	rbp
 	pop	rbx
 	pop	rsi
 	pop	rdi

 	pop	r15
 	pop	r14
 	pop	r13
 	pop	r12

\ RSP points at CDF_VM - the stack is no longer dropped
	lea	rsp, $10 [rsp]		\ allow for two returns

	jmp	r8			\ return into CALLDEF-EXIT
	ret				\ following is final RET
end-code

: [asm-frag	\ -- start ; start an assembler fragment
  code-align  !csp  here
;

also asm-core
: asm-frag]	\ -- ; finish an assembler fragment
  unres-refs? #-257 ?throw  ?csp
;
previous

struct /calldef	\ -- len ; structure used by CALLDEF:
  int cd.EntryPoint			\ entry point called by OS
  int cd.magic				\ magic number here
  int cd.cdEntryCode			\ entry code subroutine
  int cd.cdAction			\ Forth xt
  int cd.cdExitCode			\ exit code subroutine
end-struct

$0123456789ABCDEF constant #cdMagic	\ magic number

: genCallDef,	\ struct --
\ Given a structure, lay a CALLDEF: code sequence and fill in the
\ entrypoint in he structure
  >r  align
  here r@ cd.EntryPoint ! 		\ entry here
  r@ cd.cdEntryCode @ doLayAbsCall	\ call entrycode  +0
  r@ cd.cdAction @ doLayAbsCall		\ call xt         +5
  r> cd.cdExitCode @ doLayAbsCall	\ call exitcode   +10
  $C3 c,				\ ret		  +15
;					\                 +16


\ =================
\ *N User interface
\ =================

only forth definitions
also extern.voc

: set-CallDef	\ xt struct --
\ *G Given an xt and a struct, place CALL XT at offset 5 in the code
\ ** sequence.
  2dup cd.cdAction !			\ -- xt struct ; mark xt
  cd.EntryPoint @ skipCall !call	\ set call as second call
;

: get-CallDefEntry	\ struct -- entrypoint
\ *G Given a calldef structure, return the entrypoint address that
\ ** is passed to Windows.
  cd.EntryPoint @
;

[-short-branches
: CALLDEF:	\ "<text>" -- ; -- struct
\ *G Parses a C_style prototype, and use the results to create
\ ** a callback from Win64 to Forth. The following example is
\ ** for a Windows *\b{winproc} routine.
\ *E calldef: int WinProc2(
\ **   HWND myhandle,  UINT message,  WPARAM wparam1,  LPARAM lparam
\ ** ):
\ ** \ hwnd msg wparam lparam -- int
\ **   {: hwnd msg wparam lparam -- int :}	\ can use local variables
\ **   case  msg
\ **     ...
\ **   endcase
\ ** ;
\ *P When the *\fo{CALLDEF:}'s name is exexuted, it returns the
\ ** address of the calldef structure. This form is easier to
\ ** debug than *\fo{DefCallProc:} below.
  parseCBtext				\ parse arguments
  initoffsets CalcCSrcFrame  calcFDestOff

  ShowExtData? if
    cr ." CallDef: " ExternName count type
    cr ." #Args:" Numargs ?  ."  #Ints:" NumInts ?  ."  #Floats:" NumFloats ?
    .ArgList
  then

  [asm-frag >r	\ -- exit ; start of exit code
    [a  pop   r9  a] [a]		\ on dropped stack
    CDretData				\ lay data return code
    [a  call  clrCallDefVM  a] [a]
    [a  jmp   r9  a] [a]		\ return
    [a  ret  a] [a]			\ keeps disassembler happy
  asm-frag]

  [asm-frag >r	\ -- entry ; start of entry code
    [a  call  setCallDefVM  a] [a]
    CDcopyFrame
    [a  jmp  #CBdropped #CDF_RA + [rsp]  a] [a] \ return address from entry code
    [a  ret  a] [a]			\ keeps disassembler happy
  asm-frag]

  ExternName count ($create)		\ will return structure
  here  /CallDef allot			\ structure ; -- struct
  \ entry point as yet unknown
  #cdMagic over cd.magic !		\ magic number
  r> over cd.cdEntryCode !		\ entry subroutine
  ['] noop over cd.cdAction !		\ action
  r> over cd.cdExitCode !		\ exit subroutine ; -- struct
  dup genCallDef,			\ lay CALLDEF sequance
  >r  :noname  r> set-CallDEF		\ :NONAME ... ;
;
short-branches]

: dis-cd	\ struct --
\ *G Disassemble the sequence forming the callback.
\ ** Use only with children of *\fo{CallDef:}.
  cd.EntryPoint @ disasm/f  ;
: dis-cdEntry	\ struct --
\ *G Disassemble the sequence forming the callback entry code
\ ** Use only with children of *\fo{CallDef:}.
  cd.cdEntryCode @ disasm/f  ;
: dis-cdAction	\ struct --
\ *G Disassemble the sequence forming the callback action code
\ ** Use only with children of *\fo{CallDef:}.
  cd.cdAction @ disasm/f  ;
: dis-cdExit	\ struct --
\ *G Disassemble the sequence forming the callback exit code
\ ** Use only with children of *\fo{CallDef:}.
  cd.cdExitCode @ disasm/f  ;

: DefCallProc:	\ "<text>" -- ; -- entrypoint
\ *G Parses a C_style prototype, and use the results to create
\ ** a callback from Win64 to Forth. The following example is
\ ** for a Windows *\b{winproc} routine.
\ *E calldef: int WinProc2(
\ **   HWND myhandle,  UINT message,  WPARAM wparam1,  LPARAM lparam
\ ** ):
\ ** \ hwnd msg wparam lparam -- int
\ **   {: hwnd msg wparam lparam -- int :}	\ can use local variables
\ **   case  msg
\ **     ...
\ **   endcase
\ ** ;
\ *P When a child of *\fo{DefCallProc:}'s name is executed, it returns the
\ ** address of the callback's entrypoint, which makes it more
\ ** suitable for everyday programming.
  CallDef: does> get-CallDefEntry  ;

previous


\ ============
\ Example code - not an automated test
\ ============
((
calldef: int WinProc2(
  HWND myhandle,  UINT message,  WPARAM wparam1,  LPARAM lparam
):
\ hwnd msg wparam lparam -- int
  .s
;

calldef: int64 Proc2(
  int64 a,  double b,  int64 c,  double d,  int64 e,  double f
):
\ a c e -- long ; F: -- b d f
 1000
;
))

0 [if]
\ *[
\ Text output device
textbuff: abouttextdev

\ About boxes
NextID: IDD_ABOUT_VFX

\ Bitmaps
NextID: IDB_MPE_LOGO
IDB_MPE_LOGO BITMAP "%LOAD_PATH%\Mpelogo.bmp"

\ Control IDs
NextID: IDD_ABOUT_VFX
NextID: IDC_LOGO1
NextID: IDC_STATIC1

IDD_ABOUT_VFX DIALOG DISCARDABLE  0, 0, 272, 135
STYLE DS_MODALFRAME | DS_3DLOOK | DS_CENTER | WS_POPUP | WS_CAPTION |
    WS_SYSMENU
CAPTION "About VFX Forth"
FONT 8, 100, 0, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,                    215,116,   50,14
    CONTROL         "",IDC_LOGO1,"Static",SS_BITMAP | SS_CENTERIMAGE |
                    SS_REALSIZEIMAGE | SS_SUNKEN, 5,5,       40,35
    CTEXT           "VFX Forth 64 for Windows",
                    IDC_STATIC1,                  50,5,      215,106,SS_SUNKEN
END

CallDef: uint * AboutVFXDialogProc(
  HWND hDlg, UINT msg, WPARAM wparam1, LPARAM lparam1
):
  {: hdlg message wparam lparam -- ior :}
  message case
    WM_INITDIALOG of
        lparam if
          hDlg
          lparam HIWORD
          STM_SETIMAGE
          IMAGE_BITMAP
          lparam LOWORD RESOURCE::CreateBitmap
          SendDlgItemMessage drop
        then

        abouttextdev dup ip-handle ! op-handle !
        NULL 4096 0 open-gen nip if
          drop
        else
          .cold
          hdlg IDC_STATIC1 WM_SETTEXT 0
          abouttextdev gen-handle @
          dup 4096 bounds do
	    i c@ 0= if  bl i c!  then
	  loop
          SendDlgItemMessage drop
          close-gen drop
        then
        1
    endof

    WM_COMMAND of
        wparam LOWORD IDOK = if  hdlg DlgDone  endif
        wparam LOWORD IDCANCEL = if  hdlg DlgDone  endif
	0
    endof

    WM_CLOSE of
        hDlg WM_COMMAND IDOK 0 SendMessage drop 0
    endof

    drop 0
  end-case
;

: AboutVFXDialog	\ --
  NULL GetModuleHandle                  	\ -- inst
  IDD_ABOUT_VFX RESOURCE::GetDialogTemplate	\ -- inst template
  console@					\ -- inst template hparent
  AboutVFXDialogProc get-CallDefEntry		\ -- inst template hparent proc
  IDC_LOGO1 IDB_MPE_LOGO swap MAKELONG		\ -- inst template hparent proc lparam
  DialogBoxIndirectParam drop
;
\ *]
[then]


\ ======
\ *> ###
\ ======

decimal
