\ ExtProtectWin64.fth

((
Copyright (c) 2022, 2023
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal


\ **************************
\ *> extern
\ *S Using the Windows hooks
\ **************************
\ *P In VFX64 for Windows, the protected extern code is *\b{not} 
\ ** compiled by default. The code is provided in the file
\ ** *\i{VFXBase\ExtProtectWin64.fth}.

\ *P The hooks *\fo{preExtCall} and *\fo{postExtCall} are
\ ** *\fo{DEFER}red words into which you can plug actions that
\ ** will be run before and after any external call. They are
\ ** principally used:
\ *(
\ *B To save and restore the NDP state when using screen and printer
\ ** drivers that do not obey all the Windows rules.
\ *B To save and restore the NDP state and you want the NDP state
\ ** preserved regardless of any consequences. Although this is safe,
\ ** the system overhead is greater than that of preserving your
\ ** floats in variables or locals as required.
\ *B Installing error handlers that work with nested callbacks.
\ *)

\ *P The hooks *\fo{preFPExtCall} and *\fo{postFPExtCall} are
\ ** compiled into calls with floating point parameters or return
\ ** values. They do not affect the NDP state.

\ *P The examples below illustrate both actions.

only forth definitions  decimal
also extern.voc definitions


\ ===============================
\ *N Deferred words and variables
\ ===============================

also forth definitions
(( for DocGen
defer preExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls. It is executed at the start of the external
\ ** call before any parameter processing.
  assign noop to-do preExtCall
defer postExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls. It is executed at the end of the external
\ ** call after return data processing.
  assign noop to-do postExtCall
defer preFPExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls with floating point parameters or return
\ ** items. It is executed at the start of the external
\ ** call before any parameter processing.
  assign noop to-do preFPExtCall
defer postFPExtCall	\ --
\ *G *\b{Windows only}. A hook provided for debugging and extending
\ ** external calls with floating point parameters or return
\ ** items. It is executed at the end of the external
\ ** call after return data processing.
))
variable XcallSaveNDP?	\ -- addr
\ *G Set true when imports must save and restore the NDP state.
\ ** Windows only. For Windows 64 bit onwards, the default behaviour
\ ** for Windows does not include saving and restoring the NDP state.
\ ** This can be inhibited by clearing *\fo{XcallSaveNDP?}
\ ** before execution.
  0 XcallSaveNDP? !
variable abort-code	\ -- addr
\ *G Holds error code for higher level routines, especially
\ ** *\fo{RECOVERY} below. Windows versions only.
  abort-code off
variable aborting?	\ -- addr
\ *G Holds a flag to indicate whether error recovery should be
\ ** performed by a calling routine.
  aborting? off
defer xcall-fault	\ -- ; handles errors in winprocs
\ *G Used by application code in the *\fo{DEFER}red words
\ ** *\fo{preExtCall} and *\fo{postExtCall} above to install
\ ** user-defined actions.
  assign noop to-do xcall-fault
previous definitions


\ ===================
\ *N Default versions
\ ===================

also forth definitions

\ *[
code PreExtern	\ -- ; R: -- sys
\ Clears the abort code and saves the NDP state if XcallSaveNDP?
\ is set.
  mov   qword ptr abort-code , # 0	\ no previous abort code
  cmp	  [] XcallSaveNDP? , # 0	\ Win: required
  nz, if,
    pop     rax
    lea     rsp, -/fsave [rsp]
    fsave   0 [rsp]
    push    rax
  endif,
  ret
end-code
assign preExtern to-do preExtCall

code PostExtern	\ -- ; R: sys --
\ Restore the NDP state if XcallSaveNDP? is set and test the
\ abort code.
  cmp   [] XcallSaveNDP? , # 0		\ required
  nz, if,
    pop     rax
    frstor  0 [rsp]
    lea     rsp, /fsave [rsp]
    push    rax
  endif,
  \ Detecting faults in nested callbacks.
  cmp   qword ptr abort-code , # 0	\ test previous aborting code
  nz, if,
    call  [] ' xcall-fault 5 +		\ execute xcall-fault if set
  endif,
  ret
end-code
assign postExtern to-do postExtCall

code PreFPExtern	\ -- ; R: -- sys ; SFP006
\ Clears the abort code.
  mov   qword ptr abort-code , # 0	\ no previous abort code
  ret
end-code
assign preFPExtern to-do preFPExtCall

code PostFPExtern	\ -- ; R: sys -- ; SFP006
\ Test the abort code.
  cmp   qword ptr abort-code , # 0	\ test previous aborting code
  nz, if,
    call  [] ' xcall-fault 5 +		\ execute xcall-fault if set
  endif,
  ret
end-code
assign postFPExtern to-do postFPExtCall

\ *]

: DefaultExterns	\ --
\ *G Set the default PRE and POST EXTERN handlers.
  assign preExtern to-do preExtCall
  assign postExtern to-do postExtCall
  assign preFPExtern to-do preFPExtCall
  assign postFPExtern to-do postFPExtCall
;
previous definitions


\ ====================
\ *N Protected EXTERNs
\ ====================
\ *P Protected EXTERNs allow VFX Forth to recover when a crash
\ ** occurs inside a Windows call and the Forth registers have
\ ** been corrupted. For example
\ *C   255 0 GetCurrentDirectory
\ *P will crash because an address of zero is invalid. Protected
\ ** EXTERNs save the Forth registers befor making the call so
\ ** that exception handlers can restore VFX Forth to a known
\ ** state.

also system definitions

[undefined] /SCBdata [if]

struct /SCBdata	\ -- len
\ Length of data saved for a protected callback or Extern.
  int scb.xt		\ xt of callback handler
  int scb.RSP		\ ESP on entry RSP
  int scb.RBP		\ EBP on entry PSP
  int scb.RSI		\ ESI on entry
  int scb.RDI		\ EDI on entry
  int scb.R13		\ R13 on entry FSP
end-struct

/SCBdata buffer: XcallBuffer	\ -- addr
\ Saved data for protected EXTERNs.

[then]

also forth definitions

\ *[
code PreProtExtern	\ -- ; R: -- sys
\ Clears the abort code and saves the NDP state if XcallSaveNDP?
\ is set.
  mov    rdx, # XcallBuffer		\ where the saved data goes
  mov    rax, 0 [rsp]			\ return address
  sub    rax, # 6			\ xt of EXTERN (call [] prexx)
  mov    0 scb.xt [rdx], rax		\ save it
  lea    rax, 8 [rsp]			\ RSP on entry
  mov    0 scb.Rsp [rdx], rax
  mov    0 scb.Rbp [rdx], rbp
  mov    0 scb.Rsi [rdx], rsi
  mov    0 scb.Rdi [rdx], rdi
  mov    0 scb.R13 [rdx], r13

  mov    qword ptr abort-code , # 0	\ no previous abort code
  cmp	 [] XcallSaveNDP? , # 0	\ Win: required
  nz, if,
    pop     rax				\ return address
    lea     rsp, -/fsave [rsp]
    fsave   0 [rsp]
    push    rax
  endif,
  ret
end-code

code PostProtExtern	\ -- ; R: sys --
\ Restore the NDP state if XcallSaveNDP? is set and test the
\ abort code.
  mov   qword ptr XcallBuffer scb.xt , # 0  \ reset Extern in progress
  cmp   [] XcallSaveNDP? , # 0		\ required
  nz, if,
    pop     rax				\ return address
    frstor  0 [rsp]
    lea     rsp, /fsave [rsp]
    push rax
  endif,
  \ Detecting faults in nested callbacks.
  cmp   qword ptr abort-code , # 0	\ test previous aborting code
  nz, if,
    call  [] ' xcall-fault 5 +		\ execute xcall-fault if set
  endif,
  ret
end-code

code PreProtFPExtern	\ -- ; R: -- sys ; SFP006
\ Clears the abort code.
  mov    rdx, # XcallBuffer		\ where the saved data goes
  mov    rax, 0 [rsp]			\ return address
  sub    rax, # 6			\ xt of EXTERN (call [] prexx)
  mov    0 scb.xt [rdx], rax		\ save it
  lea    rax, 8 [rsp]			\ RSP on entry
  mov    0 scb.Rsp [rdx], rax
  mov    0 scb.Rbp [rdx], rbp
  mov    0 scb.Rsi [rdx], rsi
  mov    0 scb.Rdi [rdx], rdi
  mov    0 scb.R13 [rdx], r13

  mov    qword ptr abort-code , # 0	\ no previous abort code
  ret
end-code

code PostProtFPExtern	\ -- ; R: sys -- ; SFP006
\ Test the abort code.
  mov    qword ptr XcallBuffer scb.xt , # 0	\ reset Extern in progress
  cmp    qword ptr abort-code , # 0	\ test previous aborting code
  nz, if,
    call   [] ' xcall-fault 5 +		\ execute xcall-fault if set
  endif,
  ret
end-code
\ *]

: ProtectedExterns	\ --
\ *G Set the protected PRE and POST EXTERN handlers.
  assign preProtExtern to-do preExtCall
  assign postProtExtern to-do postExtCall
  assign preProtFPExtern to-do preFPExtCall
  assign postProtFPExtern to-do postFPExtCall
;
previous previous definitions

only forth definitions


\ ======
\ *> ###
\ ======

decimal

0 [if]
+DebugExterns
Extern: void mpetest0-0( void );
Extern: int mpetest0-1i0f( void );
Extern: float mpetest0-0i1f( void );
Extern: void mpetest1-0( int );
Extern: void mpetest2-0( int, void * );
Extern: void mpetest3-0( int, void *, int );
Extern: void mpetest4-0( int, void *, int, void * );
Extern: uint mpetest1-1( int );
Extern: longlong mpetest4-1( longlong a, longlong b, longlong c, longlong d );
Extern: longlong mpetest6-1( longlong a, longlong b, longlong c, longlong d, longlong e, longlong f );
-DebugExterns
[then]

