\ MiscLinux.fth - miscellaneous Linux tools

((
Copyright (c) 2007, 2008, 2009, 2018, 2020
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20200708 SFP005 Corrected MS@ and NS@ and /TIMEVAL.
20180629 SFP004 Added more directory handling.
20120306 SFP003 Reduced CPU usage in MS.
		Added COPY-FILE.
20090829 SFP002 Added some time functions.
20080506 SFP001 Moved TICKS and friends to the kernel.
))

only forth definitions
decimal

\ ===========
\ *> lintools
\ ===========

\ ******************
\ *S Error variables
\ ******************
\ *P VFX Forth for Linux uses many functions from the *\b{libc}
\ ** shared library. The thread local error variables are exposed.

AliasedExtern: errno int * __errno_location( void );
\ *G *\fo{errno} is the well known *\i{errno} C thread local variable
\ ** used by libraries and system calls.
\ ** Can be read by *\fo{l@} and written by *\fo{l!}

AliasedExtern: h_errno int * __h_errno_location( void );
\ *G *\fo{h_errno} is the *\i{h_errno} C thread local variable.
\ ** Can be read by *\fo{l@} and written by *\fo{l!}


\ *************************
\ *S Environment variables
\ *************************

: ReadEnv	\ naddr1 nlen -- vaddr vlen
\ *G Read the environment variable whose name is given by
\ ** *\i{naddr/nlen} and return the string. If there is no
\ ** such variable *\i{vaddr} is *\fo{zNull} and *\i{vlen} is
\ ** zero.
  { | temp[ #256 ] -- }
  #255 umin temp[ zplace
  temp[ getenv dup
  if  zcount  else  drop zNull 0  endif
;

: WriteEnv	\ vaddr vlen naddr nlen --
\ *G Write the string value *\i{vaddr/vlen} to the environment
\ ** variable named by *\i{vaddr/vlen}.
  { | ntemp[ #256 ] vtemp[ #256 ] -- }
  #255 umin ntemp[ zplace
  #255 umin vtemp[ zplace
  ntemp[ vtemp[ 1 setenv drop
;

: DelEnv	\ naddr nlen --
\ *G Delete the environment variable *\i{naddr/nlen}.
  { | temp[ #256 ] -- }
  #255 umin temp[ zplace
  temp[ unsetenv
;

: EnvMacro:	\ naddr nlen "<var>" -- ; -- caddr
\ *G Create a text macro called *\i{<var>} that queries the
\ ** environment variable named by *\i{naddr/nlen} the returned
\ ** string is a counted string. Use in the form:
\ *C   s" HOME" EnvMacro: $home
\ *P By convention, environment macro names start with a '$'.
   also substitutions definitions
   create
     here 0 , 0 ,			\ name pointer, buffer pointer
     here over ! -rot  $,		\ lay name string
     align here swap cell + !  0 c, #255 allot	\ value buffer
     previous definitions
   does>
     dup @ count ReadEnv #255 min	\ -- pfa caddr len
     rot cell + @ dup >r place  r>	\ -- vaddr
;

s" HOME" EnvMacro: $home
\ *G Text macro for the home directory.


\ ********************
\ *S Critical sections
\ ********************
\ *P Critical sections are implemented using the standard Linux
\ ** semaphore structures and calls.

#32 constant /sem_t	\ -- len
\ *G Size of a Linux x64 *\b{sem_t} structure. You can treat this
\ ** as an opaque type that you do not have to deal with directly.
\ ** All you have to do is to reserve memory for it, e.g.
\ *C   /sem_t buffer: MyCritSec

: InitCritSec	\ sem --
\ *G Initialise the critical section. This *\b{must} be done
\ ** before using it.
  0 1 sem_init drop
;

: TermCritSec	\ sem --
\ *G Delete the critical section associated with the smaphore.
\ ** This releases internal Linux data, the *\fo{/sem_t} structure
\ ** is still available but needs to be initialised again before
\ ** reuse. Nothing should be waiting on the seamphore before
\ ** calling *\fo{TermCritSec}.
  sem_destroy drop
;

: [CritSec	\ sem --
\ *G Wait until the section is available and lock it.
\ ** Does not call *\fo{PAUSE}.
  sem_wait drop
;

: CritSec]	\ sem --
\ *G Unlock the section.
  sem_post drop
;

: CritSec?	\ sem -- u
\ *G Returns the section's counter, where non-zero indicates
\ ** that it is available, or zero when it is locked. Returns
\ ** zero on error.
  { | temp[ cell ] -- }
  temp[ off
  temp[ sem_getvalue drop
  temp[ @
;

\ *P The critical section words use Linux semaphores, which are
\ ** counted semaphores. Thus when using critical sections you
\ ** must be careful to match the use of *\fo{[CritSec} and
\ *\fo{CritSec]}.


\ ********************
\ *S Millisecond timer
\ ********************
\ *P The Linux ticker frequency varies between implementations.
\ ** The code in this section provides simple tools to return
\ ** and handle a millisecond ticker.

Extern: long sysconf( int name );
Extern: int times( void * buf );

2 constant CLK_TCK	\ -- u

struct /tms	\ -- len
\ Time structure used by times above.
  int tms.utime  /* user time */
  int tms.stime  /* system time */
  int tms_cutime /* user time of children */
  int tms_cstime /* system time of children */
end-struct

10 value /CLK_MS	\ -- ms
\ Default ms per clock tick

: (ticks)	\ -- ms ; return ticks in ms
\ *G Return the system ticker in milliseconds. Treat this as a 32 bit
\ ** unsigned value that wraps around on overflow.
  { | tmp[ /tms ] -- }
  tmp[ times /CLK_MS *
;

: SetTicks	\ --
\ *G Calibrate the Linux ticker and install it as the action of
\ ** *\fo{TICKS}. Performed at start up.
  CLK_TCK sysconf 1 max
  #1000 swap / -> /CLK_MS
  ['] (ticks) to-do ticks
;
' SetTicks AtCold  SetTicks

also system
5 value tickStepMs	\ -- ms
\ *G Minimum interval and granularity used by *\fo{tick-ms}
\ ** below.

: tick-ms	\ ms --
\ *G Waits for at least *\i{ms} milliseconds. Uses *\fo{PAUSE}
\ ** every *\fo{tickStepMs}. This is the default action of *\fo{MS},
\ ** which is *\fo{DEFER}red.
  later  begin
    tickStepMs (OS_ms)
    dup expired
  until
  drop
; assign tick-ms to-do ms
previous


\ ****************
\ *S Time handling
\ ****************

struct /timeval \ -- size
\ *S A structure to mimic the *\b{timeval} structure for libc
\ *[
  8 field tv_sec
  8 field tv_nsec
end-struct
\ *]

struct /tm	\ -- size
\ *S A structure to mimic the *\b{tm} structure for libc
  4 field tm_sec       // 0 to 59 (or 60 for occasional rare leap-seconds)
  4 field tm_min       // 0 to 59
  4 field tm_hour      // 0 to 23
  4 field tm_mday      // 1 to 31
  4 field tm_mon       // 0 to 11, stupidly 0=January, 11=December
  4 field tm_year      // year-1900, so 79 means 1979, 103 means 2003
  4 field tm_wday      // 0 to 6, 0=Sunday, 1=Monday, ..., 6=Saturday
  4 field tm_yday      // 0 to 365, 0=1st January
  4 field tm_isdst     // 0 to 1, 1=DST is in effect, 0=it isn't
  aligned
  8 field tm_gmtoff    // Seconds east of UTC; EST=-18000, WET=3600
  8 field *tm_zone     // Timezone abbreviation, e.g. "PDT", "EST".
end-struct

: td>epoch	\ seconds mins hours day month year -- epoch
\ *G Returns the seconds since the start of the epoch. The input
\ ** time is treated as GMT/UTC.
  { | tm[ /tm ] -- }
  tm[ /tm erase
  tm[
  swap 1900 - over tm_year l!  swap 1- over tm_mon l!  tuck tm_mday l!
  tuck tm_hour l!  tuck tm_min l!  tuck tm_sec l!
  z" GMT" over *tm_zone !
  mktime
;

: epoch>td	\ epoch -- seconds mins hours day month year
\ *G Converts an epochal second into a GMT/UTC time and date.
  { | epoch[ cell ]  tm[ /tm ] -- }
  epoch[ !  tm[ /tm erase
  epoch[ tm[ gmtime_r
  dup tm_sec l@ swap  dup tm_min l@ swap  dup tm_hour l@ swap
  dup tm_mday l@ swap  dup tm_mon l@ 1+ swap  tm_year l@ 1900 +
;


\ ****************
\ *S Time and date
\ ****************
\ *P These functions rely on the ANS Forth word
\ ** *\fo{TIME&DATE ( -- s m h dd mm yyyy )} and the non-standard
\ ** *\fo{DOW ( -- dow, 0=Sun)} to get the day of the week.

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

create months	\ -- addr
\ *G String containing 3 character text for the months.
  z", JanFebMarAprMayJunJulAugSepOctNovDec"

[-sin

: .dow		\ dow --
\ *G Display day of week.
  3 * days$ 1+ + 3 type
;

: .2r		\ n --
\ *G Display n as a two digit number with leading zeros.
  s>d <# # # #> type
;

: .4r		\ n --
\ *G Display n as a four digit number with leading zeros.
  s>d <# # # # # #> type
;

: .Time&Date	\ s m h dd mm yy --
\ *G Display the system time
\ ** The format is:
\ *C   hh:mm:ss dd Mmm yyyy
  base @ >r  decimal
  dup #100 <				\ correct two digit year clocks
  if  #2000 +  endif
  rot .2r space				\ day - generate date
  swap 1- 3 * months + 3 type space	\ month
  .4r space				\ year
  .2r ." :"			 	\ hours - generate time
  .2r ." :"				\ minutes
  .2r					\ seconds
  r> base !
; doNotSin

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT (system time) is displayed, otherwise local time
\ ** is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  if
    time&date				\ -- s m h dd mm yy
    dow .dow ." , " .time&date ."      "
  else
    systime&date			\ -- s m h dd mm yy
    sysdow .dow ." , " .time&date ."  GMT "
  endif
;

sin]


\ ***************************
\ *S Hardware I/O port access
\ ***************************
\ *P The following notes are for developers working under x86-32
\ ** versions of Linux. Under normal use, direct access to I/O
\ ** ports is forbidden. However, if you are running with root
\ ** privilege, you can use the glibc functions *\b{ioperm()}.
\ ** Since Linux 2.6.8 *\b{ioperm()} works for all 65536 i/o
\ ** ports.

code pc@        \ port -- b ; read port
\ *G Read a byte from the hardware control port supplied.
  mov   edx, ebx                \ port address must be in DX
  in    al, dx                  \ read port
  movzx rbx, al                 \ clear RBX high
  next,                         \ back to Forth
end-code

code pc!        \ b port -- ; write port
\ *G Write the supplied byte to the selected hardware control port.
  mov   edx, ebx                \ port address must be in DX
  mov   eax, 0 [rbp]            \ get NOS (b)
  out   dx, al                  \ write port
  mov   rbx, cell [rbp]         \ restore TOS
  add   rbp, # 2 cells
  next,                         \ back to Forth
end-code

code pw@        \ port -- w ; read port
\ *G Read a 16 bit word from the hardware control port supplied.
  mov   edx, ebx                \ port address must be in DX
  in    ax, dx                  \ read port
  movzx rbx, ax                 \ clear EBX
  next,                         \ back to Forth
end-code

code pw!        \ w port -- ; write port
\ *G Write the supplied 16 bit word to the selected hardware control port.
  mov   edx, ebx                \ port address must be in DX
  mov   eax, 0 [rbp]            \ get NOS (b)
  out   dx, ax                  \ write port
  mov   ebx, cell [rbp]         \ restore TOS
  add   rbp, # 2 cells
  next,                         \ back to Forth
end-code

code pl@        \ port -- x ; read port
\ *G Read 32 bits from the hardware control port supplied.
  mov   edx, ebx                \ port address must be in DX
  in    eax, dx                 \ read port
  mov   ebx, eax		\ EAX -> TOS
  next,                         \ back to Forth
end-code

code pl!        \ l port -- ; write port
\ *G Write the supplied 32 bits to the selected hardware control port.
  mov   edx, ebx                \ port address must be in DX
  mov   eax, 0 [rbp]            \ get NOS (b)
  out   dx, eax                 \ write port
  mov   ebx, cell [rbp]         \ restore TOS
  add   ebp, # 2 cells
  next,                         \ back to Forth
end-code

: +Ports	\ port #ports -- ior
\ *G Enable access to a range of ports starting at *\i{port}.
\ ** Return 0 on success. Port numbers must be in the range
\ ** 0..$FFFF. You must have root permissions.
  1 ioperm
;

: -Ports	\ port #ports -- ior
\ *G Disable access to a range of ports starting at *\i{port}.
\ ** Return 0 on success. Port numbers must be in the range
\ ** 0..$FFFF. You must have root permissions.
  0 ioperm
;

: PlayNote      \ hertz ms --
\ *G Play a note on the internal PC speaker.
\ ** Ports $42, $43 and $61 must be enabled first.
  1193180 rot /
  $61 pc@ $3 or $61 pc!         \ Start Speaker Going
  $B6 $43 pc!                   \ set timer freq.
  dup $42 pc!
  8 rshift $42 pc!
  ms                         	\ Sleep for duration of note
  $61 pc@ $FC and $61 pc!       \ Switch off speaker
;

: pio-test      \ --
\ *G A test routine for hardware access. Enables ports $40..$6F
\ ** and confirms access. If you hear a familiar tune, all is well!
   $40 $30 +Ports abort" Can't enable ports"
   897   500 PlayNote
  1007   500 PlayNote
   800   500 PlayNote
   400   500 PlayNote
   599  1000 PlayNote
   $40 $30 -Ports drop
;


\ ************************
\ *S Program launch status
\ ************************
\ *P Programs can be launched in several ways.
\ *(
\ *B From a shell in foreground mode: it has a terminal.
\ *B From a shell or script in background mode, e.g. *\f{vfxlin &}:
\ ** it shares a terminal which is normally only useful for output.
\ *B From the init process or detached by another process. There
\ ** is no terminal at all.
\ *)
\ *P This code allows you to determine how the program was
\ ** launched.

0 value AppPPID		\ -- x
\ *G This application's parent's process ID.
0 value AppPGRP		\ -- x
\ *G This application's process group.
0 value ctPGRP		\ -- x
\ *G The controlling terminal's process group.
0 value AppLaunch	\ -- x
\ *G How the application was launched:
\ *(
\ *B -1 - detached
\ *B 0 - backgrounded
\ *B 1 - foreground
\ *)

: TestLaunch	\ --
\ *G Set the data above to determine how the application was
\ ** launched. Run at program launch.
  0 -> ctPGRP  -1 -> AppLaunch
  getppid -> AppPPID  getpgrp -> AppPGRP
  z" /dev/tty" 0 0 open dup 0<		\ if no terminal, detached
  if  drop  exit  endif
  dup tcgetpgrp dup -> ctPGRP AppPGRP =
  if 1 else 0 then
  -> AppLaunch
  close drop
;  ' TestLaunch AtCold


\ ********************
\ *S Folders and Files
\ ********************

: (changedir)	\ caddr len -- ior ; 0=success
\ *G Expand text macros and change the current working directory.
\ ** Return 0 on success.
  {: | cl[ 1024 chars ] -- ior :}
  cl[ 1023 substituteZ  cl[ chdir
;

: changeDir	\ caddr len --
\ *G Expand text macros and change the current working directory.
\ ** There is no return code.
  (changeDir) drop
;

: workingdir	\ -- caddr len
\ *G Return the text for the current working directory.
  {: | cd[ #1024 ] -- :}
  0 cd[ c!  cd[ #1023 getcwd
  if  cd[   else  znull  endif
  zcount >syspad
;

: dirExists?	\ caddr len -- flag
\ *G Return true if a directory exists. Macros are expanded.
  { | name[ max_path 2+ ] -- }
  expand max_path min name[ zplace
  name[ opendir dup if
    closedir drop  -1
  endif
;

: create-dir	\ caddr len -- ior
\ *G Create a directory, returning zero on success. Macros are expanded.
\ ** Default permissions are used.
  { | name[ max_path 2+ ] -- }
  expand max_path min name[ zplace
  name[ #493 mkdir			\ Decimal 493 = Octal 755
;

: forceDir	\ caddr len -- ior
\ *G Create the directory if it does not exist. Macros are expanded.
  2dup dirExists? if
    2drop 0 exit
  endif
  create-dir
;

\ SFP004...

: +dirSep	\ c$ --
\ *G Add a directory separator to the end of a counted string.
  dir1-char c@ swap addChar
; doNotSin

: prepFileName	\ caddr len --
\ *G Convert '\' and '/' characters in-place as required by the
\ ** operating system.
  bounds ?do
   i c@ dirchar? if  dir1-char @ i c!  then
  loop
; doNotSin

: prepDirName	\ caddr --
\ *G Force the counted string at caddr to end with a '\' character.
  dup count + 1- c@ dir1-char @ <>	\ -- caddr flag ; no terminator
  if  +dirSep  else  drop  then		\ add terminator
; doNotSin

[-short-branches
: makeDirLevels	\ caddr1 len1 caddr2 len -- ior
\ *G The string *\i{caddr1/len1} represents a directory that
\ ** must already exist. *\i{Caddr2/len2} represents additional
\ ** directory levels that may or may not already exist. The
\ ** additional levels are created if they do not exist.
\ ** Successful operation returns 0. For example, to create the
\ ** directory *\f{/Users/stephen/jim/foo/bar}, you could use
\ *C   s" /Users/stephen"  s" jim/foo/bar"  makedirlevels
  {: | exists[ #256 ] maybe[ #256 ] force[ #256 ] -- :}
  #255 umin maybe[ ExpandMacro prepFileName	\ expand macros, force separators
  #255 umin exists[ ExpandMacro prepFileName
  exists[ prepDirName			\ make strings end in dir separator
  maybe[ prepDirName
  maybe[ count  over >r			\ mark start of string
  begin					\ -- caddr len ; R: start --
    dup					\ if string not empty
   while
    dir1-char @ scan dup		\ find next separator
   while				\ -- caddr' len' ; R: start --
    over r@ dup rot swap -		\ -- caddr' len' start lendir
    exists[ count force[ place		\ first part of name
    force[ append			\ add second part
    force[ count forceDir		\ make directory
    if  r> drop 2drop #-425 exit  then	\ exit on error
    1 /string				\ step over '\'
  repeat then
  r> drop 2drop 0
;
short-branches]
\ : tt  s" D:\"  s" jim\foo\bar" makedirlevels  ;

\ ...SFP004

: copy-clean	\ hs hd buff --
\ clean up after the copy
  free drop				\ buff
  dup -1 =				\ hd
  if  drop  else  close-file drop  endif
  dup -1 =				\ hs
  if  drop  else  close-file drop  endif
; doNotSin

[-short-branches
: copy-file	\ src srclen dest destlen nooverwrite -- ior
\ *G Copy the file. If *\i{nooverwrite} is non-zero and the destination
\ ** exists, an error is returned.
  {: src slen dest dlen nov | hs hd nrd buff fsize[ 2 cells ] -- ior :}
  nov if
    dest dlen FileExist?
    if  -401 exit  endif
  endif
  -1 to hs  -1 to hd  0 to buff
  $1000 allocate
  if  drop  -402 exit  endif
  to buff
  src slen r/o open-file
  if  drop -403  hs hd buff copy-clean  exit  endif
  to hs
  dest dlen r/w create-file
  if  drop -404  hs hd buff copy-clean  exit  endif
  to hd
  hs file-size drop fsize[ 2!
  begin
    fsize[ 2@ d0<>
   while
    buff $1000 hs read-file
    if  drop -405  hs hd buff copy-clean  exit  endif
    to nrd
    buff nrd hd write-file
    if  -406  hs hd buff copy-clean  exit  endif
    fsize[ 2@ nrd 0 d- fsize[ 2!
  repeat
  0
  hs hd buff copy-clean
;
short-branches]

((  MacOS/Linux version 
char * realpath(
  const char *restrict file_name, char *restrict resolved_name)
;
PATH_MAX = 1024
))

: fullPathName	\ caddr1 len1 -- caddr2 len2
\ *G Convert a file name to its (sort of) canonical form. For
\ ** example, relative path names are converted to absolute path
\ ** names. Macro names are expanded before conversion. The output
\ ** name is zero terminated; len does not include the terminator.
\ ** The output is in the *\fo{SYSPAD} area.
\ ** On error, *\i{caddr2/len2} is the portion of the file name
\ ** that caused the error, usually ".". The word *\fo{fullPathName}
\ ** uses the MacOS function *\f{realpath()} which is unreliable
\ ** with relative file names.
  {: | ipath[ 1024 1+ ] opath[ 1024 1+ ] -- :}
  expand ipath[ zplace  0 opath[ !
  ipath[ opath[ realpath drop
  opath[ zcount >SysPad
;

variable pathFound
2variable inPath
2variable sfPath

also system
: samePathName	     \ #thread nfa -- flag
  nip name> execute SFSTRUCT.szPathName	\ -- szname ; long name of file
  zcount fullPathName sfpath 2!		\ stash name
  inpath 2@ sfpath 2@ istr=
  dup pathFound !  0=			\ save found status, exit if found
;
previous

: (required2)	\ caddr u --
\ *G If the file specified by *\i{c-addr/u} has already been
\ ** *\fo{INCLUDED}, discard *\i{c-addr/u}; otherwise, perform
\ ** the function of *\fo{INCLUDED}. You must provide the
\ ** source file's extension.
  {: | buff[ #1024 ] -- :}
  pathFound off
  2dup fullPathName dup >r buff[ zplace	\ SYSPAD is transient
  buff[ r> inPath 2!			\ save buff/len
  ['] samePathName ['] SourceFiles >body @ WalkWordList
  pathFound @
  if  2drop  else  included  then
;
assign (required2) to-do required


\ ======
\ *> ###
\ ======

decimal
