/*
 *	Experiments with getting signals to return to a different place.
 *
 *	R.J.Dunlop    3 June 2007
 *
 *	WARNING:  Non-portable ia32 linux code for experiment only
 */

#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <sys/ucontext.h>

#include "hd.h"


/*
 *	Stack structure seen by siginfo signal handler

	|             |
	+-------------+
	| Return code |		8 bytes of code to return to kernel context at
	|             |		end of the handler run via a special system call
	|             |		(now a dummy left as a signature for debuggers
	|             |		as many systems don't allow execution from the
	|             |		stack)
	+-------------+
	| Floating    |		FPU state if FPU has been used by this process
	| point state |
	+-------------+
	| User        |		POSIX extension user context for the signal
	| Context     |         handler.  Includes register contents etc, many
     +->|             |		of which are modifiable by the signal handler
     |  +-------------+
     |  | Siginfo     |		Traditional signal information structure.
     |  | structure   |<-+	Duplicates/simplifies some of User Context
     |  +-------------+  |
     +--| void *      |	 |	Pointer to POSIX user context
	+-------------+  |
	| siginfo_t * |--+	Pointer to Siginfo structure
	+-------------+
	| Signum      |		Actual signal number generated
	+-------------+
	| Return      |		Originally this pointed at the stack based
	| address     |		return code above.  Now points directly at the
	|             |		sigreturn syscall gate in the vDSO
	+-------------+
 */



static int zero_dividend = 0;	/* Used to generate a divide by zero signal */

				/* Record the traps context here.  Currently
				 * for testing I record the entire user context
				 * but could possibly get away with just the
				 * stack and instruction pointers
				 */
static struct ucontext div_zero_trap_uc;

				/* This will be the user context that sprung
				 * the trap
				 */
static struct ucontext div_zero_sprung_uc;
static int div_zero_trap_sprung = 0;


/*
 *	Called via a know div by zero error to record the user context for
 *	the div by zero trap
 */
static void
sigrecordtrap ( int signum, siginfo_t *siginfo, void *vp )
{
	struct ucontext *uc = vp;

	printf ("sigrecordtrap( %d, %p, %p )\n", signum, siginfo, vp );
/*	hexdump_buf ("Siginfo", siginfo, sizeof ( siginfo_t )); */
/*	hexdump_buf ("Ucontext", uc, sizeof ( struct ucontext )); */

	/* Save the traps user context */
	div_zero_trap_uc = *uc;

	/* Escape from div by 0 loop */
	/* This only works because we know the cause of the error (in fact I
	 * don't think it's guaranteed anyway but depends on the code the
	 * compiler generated).  A proper escape would require decoding the
	 * faulty instruction to work out what register to poke.
	 */
	zero_dividend = 1;
}


/*
 *	Modify the user context to return to the previous trap point
 */
static void
sigspringtrap ( int signum, siginfo_t *siginfo, void *vp )
{
	struct ucontext *uc = vp;

	printf ("sigspringtrap( %d, %p, %p )\n", signum, siginfo, vp );
/*	hexdump_buf ("Siginfo", siginfo, sizeof ( siginfo_t )); */
/*	hexdump_buf ("Ucontext", uc, sizeof ( struct ucontext )); */

	/* Record the springing context for user diagnostics */
	div_zero_sprung_uc = *uc;
	div_zero_trap_sprung = 1;

	/* Restore previous context */
	*uc = div_zero_trap_uc;
}


/*
 *	Dummy user code with a fatal error
 */
static void
div_zero ()
{
	int x = 0;

	printf ("User divide by zero\n");
	x = 99 / x;
	printf ("After user divide by zero\n");
}

static void
user_func ()
{
	div_zero ();
}


/*
 *	Setup a divide by zero trap (actually any FPE) and then test it
 */ 
int
main ( int argc, char *argv[] )
{
	int x;
	struct sigaction act;	/* Set signal handler/options here */
	int err;


	/* Setup a signal handler to record the user context for a trap */
	act.sa_sigaction = sigrecordtrap;
	sigemptyset ( &act.sa_mask );
	act.sa_flags = SA_SIGINFO;
	err = sigaction ( SIGFPE, &act, NULL );
	if ( err != 0 )
	{
	    fprintf ( stderr,"Setting recording trap. Err %d, errno %d\n", err,
				errno );
	}

	/* Divide by zero to generate a signal */
	printf ("Before trap setting divide by 0\n");
	x = 99 / zero_dividend;

	if ( ! div_zero_trap_sprung )
	{
	    printf ("After trap setting divide by 0\n");
	}
	else
	{
	    /* A user divide by zero will return here */

	    printf ("Divide by zero returned to trap point\n");
	    hexdump_buf ("User context", &div_zero_sprung_uc,
						sizeof ( struct ucontext ));
	    exit ( 0 );
	}


	/* Now set the signal routine to return to the trap point */
	act.sa_sigaction = sigspringtrap;
	sigemptyset ( &act.sa_mask );
	act.sa_flags = SA_SIGINFO;

	err = sigaction ( SIGFPE, &act, NULL );
	if ( err != 0 )
	{
	    fprintf ( stderr,"Setting signal trap. Err %d, errno %d\n", err,
					    errno );
	}

	
	/* Call the user code, which deliberately springs the trap */
	user_func ();

	exit ( 0 );
}

