\ SIGNAL.FTH - Linux signal handling

((
Copyright (c) 2007, 2009, 2020, 2023
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20230529 SFP004 Changed SETSIGNAL to return nothing.
                Previous version is now called (SETSIGNAL.)
20230525 GEW003 Added SIGSHOW? and register dump check.
20200710 SFP002 Linux 64 bit conversion.
20090416 SFP001 Added Exception frame dump.
))

only forth definitions
decimal


\ ===========
\ *> lintools
\ *S Linux signal handling
\ ===========

\ *************
\ *N Structures
\ *************
\ *P Signal numbers
\ ** See /usr/src/linux-2.6.8-24/include/asm-i386/sigcontext.h
\ *[
#define SIGHUP		 1
#define SIGINT		 2
#define SIGQUIT		 3
#define SIGILL		 4
#define SIGTRAP		 5
#define SIGABRT		 6
#define SIGIOT		 6
#define SIGBUS		 7
#define SIGFPE		 8
#define SIGKILL		 9
#define SIGUSR1		10
#define SIGSEGV		11
#define SIGUSR2		12
#define SIGPIPE		13
#define SIGALRM		14
#define SIGTERM		15
#define SIGSTKFLT	16
#define SIGCHLD		17
#define SIGCONT		18
#define SIGSTOP		19
#define SIGTSTP		20
#define SIGTTIN		21
#define SIGTTOU		22
#define SIGURG		23
#define SIGXCPU		24
#define SIGXFSZ		25
#define SIGVTALRM	26
#define SIGPROF		27
#define SIGWINCH	28
#define SIGIO		29
#define SIGPOLL		SIGIO
#define SIGPWR		30
#define SIGSYS		31
#define	SIGUNUSED	31
\ *]

struct /fpstate	\ -- len
\ *G Regular FPU environment.
\ ** See *\i{/usr/src/linux-2.6.8-24/include/asm-i386/sigcontext.h}.
  int fp.cw
  int fp.sw
  int fp.tag;
  int fp.ipoff
  int fp.cssel
  int fp.dataoff
  int .fp.datasel
  #10 8 * field fp.st[]
  2 field fp.status
  2 field fp.magic			/* 0xffff = regular FPU data only */
\ FXSR FPU environment
  cell 6 * field fp.fxsr_env[]		/* FXSR FPU env is ignored */
  int fp.mxcsr
  cell +
  #16 8 * field fp.fxsr_st[]		/* FXSR FPU reg data is ignored */
  #16 8 * field fp.xmm[]
  4 #56 * field fp.padding[]
end-struct

struct /fpstate64	\ -- len
\ *G The FPU environment.
\ ** See *\i{/usr/include/sigcontext.h}.
  2 field fp.cwd
  2 field fp.swd
  2 field fp.ftw;
  2 field fp.fop
  int fp.rip
  int fp.rdp
  4 field fp.mxcsr
  4 field fp.mxcr_mask
  #16 8 * field fp.st[]			\ ST0..7
  #16 #16 * field fp.xmmx[]		\ XMM0..15
  #48 field fp.reserved2
  #48 field fp.reserved3
end-struct

struct /lfpstate64	\ -- len
\ *G The libc FPU environment.
\ ** See *\i{/usr/include/ucontext.h}.
  2 field lfp.cwd
  2 field lfp.swd
  2 field lfp.ftw;
  2 field lfp.fop
  int lfp.rip
  int lfp.rdp
  4 field lfp.mxcsr
  4 field lfp.mxcr_mask
  #16 8 * field lfp.st[]		\ ST0..7
  #16 #16 * field fp.xmmx[]		\ XMM0..15
  96 field lfp.reserved
end-struct

8 constant /fpregset64	\ --len
\ *G The type *\b{fpregset_t} is a pointer to a *\b{_libc_fpstate}.

struct /gregset64	\ -- len
\ *G CPU *\b{gregset_t} structure.
\ ** See *\i{/usr/src/include/x86_64/ucontext.h}.
  int gr.r8
  int gr.r9
  int gr.r10
  int gr.r11
  int gr.r12
  int gr.r13
  int gr.r14
  int gr.r15
  int gr.rdi
  int gr.rsi
  int gr.rbp
  int gr.rbx
  int gr.rdx
  int gr.rax
  int gr.rcx
  int gr.rsp
  int gr.rip
  int gr.rflags
  2 field gr.cs		\ really a 16 bit item, use w@ and w!
  2 field gr.gs		\ really a 16 bit item, use w@ and w!
  2 field gr.fs		\ really a 16 bit item, use w@ and w!
  2 field gr.ss		\ really a 16 bit item, use w@ and w!
  int gr.err
  int gr.trapno
  int gr.oldmask
  int gr.cr2
end-struct

struct /sigcontext	\ -- len
\ *G CPU *\b{sigcontext} structure.
\ ** See *\i{sigcontext.h}.
\ ** Note that this is *\b{not} the same as a *\b{ucontext} structure.
  /gregset64 field sc.gregs
  int sc.*fpstate		\ pointer to an fpstate
  64 field sc.reserved1
end-struct

struct /stack_t		\ -- len
\ *S Stack description structure.
  int st.sp		\ pointer to ...
  int st.flags		\ ...
  int st.size		\ ..
end-struct

struct /mcontext64	\ -- len
\ *G System *\b{ucontext} structure returned by signal handlers.
  /gregset64 field mc.gregs
  int mc.*fpregset		\ pointer to _libc_fpstate64
  #64 field mc.reserved
end-struct

struct /ucontext64	\ -- len
\ *G System *\b{ucontext} structure returned by signal handlers.
\ ** See *\i{ucontext.h}
  int uc.flags
  int uc.*link
  /stack_t field uc.stack
  /mcontext64 field uc.mcontext64
  #128 field uc.sigmask		\ 1024 bits
  /lfpstate64 field uc.lfpmem	\ FPU environment
end-struct

struct /sigaction	\ -- len ; 64 bit
\ *G System *\b{sigaction} structure.
  int sa.action		\ action of signal, entrypoint
  #128 field sa.mask	\ 1024 bits
  int sa.flags		\ special flags
  int sa.restorer	\ restore handler
end-struct

Extern: int sigaction(int signum, void * act, void * oldact);
\ act can specify either
\   void sa_handler( int signum );
\ or
\   void sa_siginfo( int signum, siginfo_t * siginfo, ucontext_t * uc );
\ return 0 on success, -1 on error
Extern: int sigemptyset(void * set);
\ generate an empty signal set.
Extern: int sigaddset(void * set, int signum);
Extern: int sigdelset(void * set, int signum);
Extern: int raise(int sig);

4 constant SA_SIGINFO
0x0400:0000 constant SA_RESTORER


\ ******************
\ *N Signal handling
\ ******************
\ *P The following is the stack structure seen by the *\b{siginfo}
\ ** signal handler.
\ *E
\ **    |             |
\ **    +-------------+
\ **    | Return code |		8 bytes of code to return to kernel context at
\ **    |             |		end of the handler run via a special system call
\ **    |             |		(now a dummy left as a signature for debuggers
\ **    |             |		as many systems don't allow execution from the
\ **    |             |		stack)
\ **    +-------------+
\ **    | Floating    |		FPU state if FPU has been used by this process
\ **    | point state |
\ **    +-------------+
\ **    | User        |		POSIX extension user context for the signal
\ **    | Context     |         handler.  Includes register contents etc, many
\ ** +->|             |		of which are modifiable by the signal handler
\ ** |  +-------------+
\ ** |  | Siginfo     |		Traditional signal information structure.
\ ** |  | structure   |<-+	Duplicates/simplifies some of User Context
\ ** |  +-------------+  |
\ ** +--| void *      |	 |	Pointer to POSIX user context
\ **    +-------------+  |
\ **    | siginfo_t * |--+	Pointer to Siginfo structure
\ **    +-------------+
\ **    | Signum      |		Actual signal number generated
\ **    +-------------+
\ **    | Return      |		Originally this pointed at the stack based
\ **    | address     |		return code above.  Now points directly at the
\ **    |             |		sigreturn syscall gate in the vDSO
\ **    +-------------+
\ **
\ ** Not checked for 64 bit Linux.


create sigNames	\ -- addr
\ *G Holds the signal numbers and names as counted strings.
   0 , ", ???"	      1 , ", SIGHUP"    2 , ", SIGINT"     3 , ", SIGQUIT"
   4 , ", SIGILL"     5 , ", SIGTRAP"   6 , ", SIGIOT"     7 , ", SIGBUS"
   8 , ", SIGFPE"     9 , ", SIGKILL"  10 , ", SIGUSR1"   11 , ", SIGSEGV"
  12 , ", SIGUSR2"   13 , ", SIGPIPE"  14 , ", SIGALRM"   15 , ", SIGTERM"
  16 , ", SIGSTK"    17 , ", SIGCHLD"  18 , ", SIGCONT"   19 , ", SIGSTOP"
  20 , ", SIGTSTP"   21 , ", SIGTTIN"  22 , ", SIGTTOU"   23 , ", SIGURG"
  24 , ", SIGXCPU"   25 , ", SIGFSXZ"  26 , ", SIGVTALRM" 27 , ", SIGPROF"
  28 , ", SIGWINCH"  29 , ", SIGIO"    30 , ", SIGPWR"    31 , ", SIGSYS"

: .sigName	\ n --
\ *G Given a signal number, display its name.
  dup #31 u>
  if  drop 0  endif
  sigNames swap 0 ?do
    cell +  count + aligned
  loop
  cell + count type
;

: .RSitem	\ x --
\ *G Display an item retrieved from the faulting return stack.
  cr ."  $" dup .xword			\ display stack item as number
  dup ['] firstname u>			\ above start of dictionary
  over here u< and if			\ below end of dictionary
    space  ip>nfa .name			\ show word name
  else
    drop
  endif
; doNotSin

: .mcontext64	\ mc --
\ *G Display data from the mcontext structure.
  mc.gregs				\ step to gregs structure
  cr
  ."  CS=" dup gr.cs w@ .word
  ."  FS=" dup gr.fs w@ .word  ."  GS=" dup gr.gs w@ .word
  cr ."  RAX=" dup gr.rax @ .xword  ."  RBX=" dup gr.rbx @ .xword
  cr ."  RCX=" dup gr.rcx @ .xword  ."  RDX=" dup gr.rdx @ .xword
  cr ."  RSI=" dup gr.rsi @ .xword  ."  RDI=" dup gr.rdi @ .xword
  cr ."  RBP=" dup gr.rbp @ .xword  ."  RSP=" dup gr.rsp @ .xword
  cr ."  R8 =" dup gr.r8  @ .xword  ."  R9 =" dup gr.r9  @ .xword
  cr ."  R10=" dup gr.r10 @ .xword  ."  R11=" dup gr.r11 @ .xword
  cr ."  R12=" dup gr.r12 @ .xword  ."  R13=" dup gr.r13 @ .xword
  cr ."  R8 =" dup gr.rbp @ .xword  ."  R9 =" dup gr.rsp @ .xword
  cr ."  RIP=" dup gr.RIP @ .xword  ."  EFLAGS=" dup gr.rflags @ .dword

  cr ." --- RS top ---"
  dup gr.rsp @ 8 cells bounds
  do  i @ .RSitem  cell +loop
  drop
; doNotSin

: SigThrow	\ --
\ *G Runs the O/S THROW action.
  #-57005 throw
;

variable #SigNest	\ --
\ Counts Signal nesting level.

: .sigTrap	\ signum *siginfo *ucontext --
  nip					\ discard siginfo
  uc.mcontext64				\ point at CPU context
  dup .mcontext64			\ display context
  cr ." Signal number " swap .sigName	\ display signal number
  cr ." at address " mc.gregs gr.RIP @ dup .xword
     ." , probably in " ip>nfa .name
  cr
;

1 value -NestedSigs?	\ -- x
\ *G Set non-zero to cause an exit if a nested signal exception
\ ** occurs in *\fo{(SigGenTrap)} below.
1 value SigPause?	\ -- x
\ *G Set non-zero to cause a pause when a signal is processed
\ ** in *\fo{(SigGenTrap)} below.
1 value SigShow?
\ *G Set non-zero to show register dump when a signal is processed
\ ** in *\fo{(SigGenTrap)} below.

create restartData	\ -- addr
  up@ ,  r0 @ ,  s0 @ ,			\ UP, RSP, PSP

: setRestart	\ --
\ Set up the restart data area.
  restartData  up@ over !  r0 @ over cell + !  s0 @ swap 2 cells + !
;
' setRestart AtCold

: doRestart	\ --
\ Executed to perform a restart by running QUIT.
  restartData  dup @ up!  dup cell + @ rp!  2 cells + @ sp!
  ip-default @ ip-handle !  op-default @ op-handle !
  quit
;

: setRIP	\ xt *uc --
\ set address for signal return.
  uc.mcontext64 mc.gregs gr.RIP !
;

also system
: bailOut	\ exitcode --
\ Get out as gracefully as possible
  ExitCode !  cr ?UnprepTerm (bye)
;
previous

: SigNested	\ signum *siginfo *ucontext -- rip
\ Handle nested signal/fault.
  {: signum *siginfo *ucontext -- rip :}
  #SigNest off				\ clear nesting
  -NestedSigs? if			\ check for permission
    cr ." Exiting on nested signal" cr
    signum #128 + bailOut
  else
    ['] SigThrow
  endif
;

[-short-branches
: SigUnNested	\ signum *siginfo *ucontext -- rip
\ Handle unnested signal/fault.
  2 pick >r					\ preserve signum
  #SigNest incr
  SigShow? if  .sigTrap  else  2drop drop  then
  #SigNest decr	\ --
  SigPause? if
    flushKeys
    cr ." Press E to exit, R to restart, other to continue: "
    case  key upc  dup emit
      [char] E of  r> #128 + bailOut  endof
      [char] R of  ['] doRestart  endof
        ['] SigThrow swap
    endcase
  else
    ['] SigThrow
  endif
  r> drop
;
short-branches]

CallDef: void SigGenCall( int signum, void * siginfo, void *ucontext ):
\ *G Generic trap handler that causes a *\fo{-57005 THROW} on return.
\ ** Callbacks are documented in a separate section of the manual.
\ ** The action of *\fo{SigGenCall} is to display an error message. On
\ ** return to OS X, a Forth *\fo{THROW} will occur.
\ signum *siginfo *ucontext --
\  cr depth .
\ cr  2 pick ." #" .xword space  over ." S" .xword space  dup ." U" .xword
\ cr  key? drop
  dup >r  #SigNest @
  if  SigNested  else  SigUnNested  endif
  r> setRIP
;

SigGenCall get-CallDefEntry constant SigGenTrap

: (setSignal)	\ callback signum -- int
\ *G Set a signal handler to execute the given callback. The
\ ** callback must have the stack effect *\fo{( signum *siginfo *ucontext -- )}
  { | act[ /sigaction ] -- }
  act[ /sigaction erase
  act[ sa.mask sigemptyset drop		\ clean masks
  SA_SIGINFO  act[ sa.flags l!		\ which type of handler
  swap act[ sa.action !			\ what to run
  act[ NULL sigaction			\ set it
;

: setSignal	\ callback signum --
\ *G Set a signal handler to execute the given callback. The
\ ** callback must have the stack effect *\fo{( signum *siginfo *ucontext -- )}
  (setSignal) drop
;

: setSigTraps	\ --
\ *G Install the *\fo{SigGenTrap} signal handler for signals SIGILL,
\ ** SIGFPE and SIGSEGV. Performed at startup.
  #SigNest off
  SigGenTrap SIGILL setSignal
  SigGenTrap SIGFPE setSignal
  SigGenTrap SIGSEGV setSignal
; ' setSigTraps AtCold

\ Check for default I/O device
ip-default @ 0= [if]
  default-console  dup ip-default !  op-default !
[then]


\ ======
\ *> ###
\ ======

decimal

